/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#include "burtgooey.h"
#include "burtgooeyfuncdecls.h"
#include <unistd.h>

#define MATHOP1VIEWSNAPBUTTONNAME "math_op1_view_snap_button"
#define MATHOP2VIEWSNAPBUTTONNAME "math_op2_view_snap_button"

#define MATHOP1PRINTSNAPBUTTONNAME "math_op1_print_snap_button"
#define MATHOP2PRINTSNAPBUTTONNAME "math_op2_print_snap_button"

/* Externally visible functions */
void main_addsub_button_cb();
void cleanup_addsub();

/* Private local functions */
static void setup_math();
/* callbacks */
static void math_op1_snap_file_sel_button_cb();
static void math_op2_snap_file_sel_button_cb();
static void math_view_snap_button_cb();
static void math_print_snap_button_cb();
static void math_go_button1_cb();
static void math_view_log_button1_cb();
static void math_file_sel_ok_cb();
static void math_operation_toggle_cb();
static void math_sdds_toggle1_cb();

/* Local Global Variables */
static char *Math_labels_strings[] =
    {"Snapshot Filename:", "Comments:", "Keywords:"};
/* math operations */
#define ADDITION    0
#define SUBTRACTION 1
/* NOTE: Must be in the same order as the #defines */
static char *Math_ops_labels_strings[] = {"Addition", "Subtraction"};
static Widget Math_shell;
static Widget Math_log1_shell;
static Widget Math_op1_text;
static Widget Math_op1_view_snap_scroll_txt;
static Widget Math_op1_view_snap_shell;
static Widget Math_op2_text;
static Widget Math_op2_view_snap_scroll_txt;
static Widget Math_op2_view_snap_shell;
static Widget Math_log1_scroll_txt;
static Widget Math1_text[XtNumber(Math_labels_strings)];
static Widget Math_status1_label; 
static char Math_logfilename[MAXPATHLEN];
static char Math_operation;
static char Math_sdds_output_spec1;

/****************************************************************
*
*
****************************************************************/

void main_addsub_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

static char default_snapshot_filename[MAXPATHLEN];

    if (!Addsub_setup)
    {
	setup_math((Widget) client_data);
	Addsub_setup = TRUE;
    } /* endif */

    if (BurtCreateUniqueFilename((char *) NULL, ".snap", 
	default_snapshot_filename))
    {
	XmTextFieldSetString(Math1_text[0], default_snapshot_filename);
	XtMapWidget(Math_shell);
    }
    else
	popup_message("SYS FAULT: could not create default_snapshot_filename.");


} /* end main_math_button_cb() */

/****************************************************************
*
*
****************************************************************/

void cleanup_addsub()
{
    unlink(Math_logfilename);
} /* end cleanup_math() */

/****************************************************************
*
*
****************************************************************/

static void setup_math(shell)
Widget shell;
{

Arg args[NARGS];
int nargs;
XmString xs;
int i;
Widget math_form, math_rc;
Widget math_op1_snap_file_sel, math_op2_snap_file_sel;
Widget math_buttons1_rc;
Widget math_go1_button, math_view_log1_button, math_help1_button;
Widget math_done1_button, math_print_log1_button;
Widget math_buttons2_rc;
Widget math_go2_button, math_view_log2_button, math_done2_button;
Widget math_print_log2_button;
Widget math_ops_rc;
Widget math_toggle[XtNumber(Math_ops_labels_strings)];
Widget math_equal1_label;
Widget math_equal2_label;

Widget math_addition_subtraction_label;
Widget math_op1_rc;
Widget math_op1_snap_files_buttons_rc;
Widget math_op1_snap_file_sel_button, math_op1_view_snap_button;
Widget math_op1_print_snap_button;
Widget math_op1_view_snap_form, math_op1_view_snap_rc, math_op1_view_snap_done_button;

Widget math_op2_rc;
Widget math_op2_snap_files_buttons_rc;
Widget math_op2_snap_file_sel_button, math_op2_view_snap_button;
Widget math_op2_print_snap_button;
Widget math_op2_view_snap_form, math_op2_view_snap_rc, math_op2_view_snap_done_button;
Widget math1_labels[XtNumber(Math_labels_strings)];

Widget math_sdds1_rc, math_sdds1_toggle[XtNumber(Sdds_labels_strings)];
Widget math_log1_form, math_log1_rc, math_log_done1_button;

    if (!BurtCreateUniqueFilename(PUBLICWRITEDIRECTORY, ".addsub.log",
            Math_logfilename))
    {
        fprintf(Log_fp,
            "ERROR: unable to create a unique filename for addsub log\n");
        exit(1);
    } /* endif */

    /**************/
    /*            */
    /* math shell */
    /*            */
    /**************/

    Math_shell = XtVaCreateWidget("Addition/Subtraction of Snapshot Files", 
		    topLevelShellWidgetClass, shell,
		    XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
		    NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 500); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 800); nargs ++;
    math_form = XtCreateWidget("math_form", xmFormWidgetClass, Math_shell, 
		    args, nargs);

    /* math popup snapshot file selection boxes */
    xs = XmStringCreateSimple("Snapshot File Selector (Operand 1)");
    nargs = 0;
    XtSetArg(args[nargs], XmNdialogTitle, xs); nargs ++;
    math_op1_snap_file_sel = XmCreateFileSelectionDialog(math_form, 
				"math_op1_snap_file_sel", args, nargs);
    XtUnmanageChild(XmFileSelectionBoxGetChild(math_op1_snap_file_sel, 
	XmDIALOG_HELP_BUTTON)); /* removing HELP button */
    XmStringFree(xs);

    xs = XmStringCreateSimple("Snapshot File Selector (Operand 2)");
    nargs = 0;
    XtSetArg(args[nargs], XmNdialogTitle, xs); nargs ++;
    math_op2_snap_file_sel = XmCreateFileSelectionDialog(math_form, 
				"math_op2_snap_file_sel", args, nargs);
    XtUnmanageChild(XmFileSelectionBoxGetChild(math_op2_snap_file_sel, 
	XmDIALOG_HELP_BUTTON)); /* removing HELP button */
    XmStringFree(xs);

    /* main set rowcolumn */
    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    math_rc = XmCreateRowColumn(math_form, "math_rc", args, nargs);

	/* snapshot addition/subtraction label */
	math_addition_subtraction_label = XmCreateLabel(math_rc, 
	    "math_addition_subtraction_label", NULL, 0);
	xs = XmStringCreateSimple("Snapshot File Addition/Subtraction");
	XtVaSetValues(math_addition_subtraction_label,
	    XmNlabelString, xs,
	    XmNalignment, XmALIGNMENT_CENTER,
	    NULL);
	XmStringFree(xs);
	XtManageChild(math_addition_subtraction_label);

	/* op1 */
	math_op1_rc = XmCreateRowColumn(math_rc, "math_op1_rc", NULL, 0);
	    Math_op1_text = XmCreateTextField(math_op1_rc, 
			    "Math_op1_text", 
			    NULL, 0);
	    XtManageChild(Math_op1_text);

	    /* math op1 snapshot file buttons rowcolumn */
	    nargs = 0;
	    XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	    math_op1_snap_files_buttons_rc = XmCreateRowColumn(math_op1_rc, 
					"math_op1_snap_files_buttons_rc", 
					args, nargs);

		math_op1_snap_file_sel_button =  
		    XmCreatePushButton(math_op1_snap_files_buttons_rc, 
			"math_op1_snap_file_sel_button", 
			NULL, 0);
		XtManageChild(math_op1_snap_file_sel_button);

		math_op1_view_snap_button =  
		    XmCreatePushButton(math_op1_snap_files_buttons_rc, 
			MATHOP1VIEWSNAPBUTTONNAME, 
			NULL, 0);
		XtManageChild(math_op1_view_snap_button);

		math_op1_print_snap_button =  
		    XmCreatePushButton(math_op1_snap_files_buttons_rc, 
			MATHOP1PRINTSNAPBUTTONNAME, 
			NULL, 0);
		XtManageChild(math_op1_print_snap_button);

	    XtManageChild(math_op1_snap_files_buttons_rc);
	XtManageChild(math_op1_rc);

	/* math operation */
	/* from page 414 AT&T Motif Course Manual */
	nargs = 0;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	math_ops_rc = XmCreateRadioBox(math_rc, "math_ops_rc", 
			args, nargs);

	    for (i = 0; i < XtNumber(Math_ops_labels_strings); i ++)
		math_toggle[i] = XmCreateToggleButton(math_ops_rc, 
				    Math_ops_labels_strings[i], NULL, 0);

	    /* setting initial radio box toggle state */
	    XmToggleButtonSetState(math_toggle[ADDITION], True, False);
	    Math_operation = ADDITION;

	    XtManageChildren(math_toggle, XtNumber(Math_ops_labels_strings));

	XtManageChild(math_ops_rc);

	/* op2 */
	math_op2_rc = XmCreateRowColumn(math_rc, "math_op2_rc", NULL, 0);
	    Math_op2_text = XmCreateTextField(math_op2_rc, 
				"Math_op2_text", NULL, 0);
	    XtManageChild(Math_op2_text);

	    /* math op2 snapshot file buttons rowcolumn */
	    nargs = 0;
	    XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	    math_op2_snap_files_buttons_rc = XmCreateRowColumn(math_op2_rc, 
					"math_op2_snap_files_buttons_rc", 
					args, nargs);

		math_op2_snap_file_sel_button =  
		    XmCreatePushButton(math_op2_snap_files_buttons_rc, 
			"math_op2_snap_file_sel_button", 
			NULL, 0);
		XtManageChild(math_op2_snap_file_sel_button);

		math_op2_view_snap_button =  
		    XmCreatePushButton(math_op2_snap_files_buttons_rc, 
			MATHOP2VIEWSNAPBUTTONNAME, 
			NULL, 0);
		XtManageChild(math_op2_view_snap_button);

		math_op2_print_snap_button =  
		    XmCreatePushButton(math_op2_snap_files_buttons_rc, 
			MATHOP2PRINTSNAPBUTTONNAME, 
			NULL, 0);
		XtManageChild(math_op2_print_snap_button);

	    XtManageChild(math_op2_snap_files_buttons_rc);
	XtManageChild(math_op2_rc);

	/* = label */
	math_equal1_label = XmCreateLabel(math_rc, "math_equal1_label", 
		    NULL, 0);
	xs = XmStringCreateSimple("=");
	XtVaSetValues(math_equal1_label,
	    XmNlabelString, xs,
	    NULL);
	XmStringFree(xs);
	XtManageChild(math_equal1_label);

	for (i = 0; i < XtNumber(Math_labels_strings); i ++)
	{
	    math1_labels[i] = XmCreateLabel(math_rc, Math_labels_strings[i], 
				NULL, 0);
	    Math1_text[i] = XmCreateTextField(math_rc, "", NULL, 0);
	} /* endfor */
	XtManageChildren(math1_labels, XtNumber(Math_labels_strings));
	XtManageChildren(Math1_text, XtNumber(Math_labels_strings));

	/* math sdds spec */
	/* from page 414 AT&T Motif Course Manual */
	nargs = 0;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	math_sdds1_rc = XmCreateRadioBox(math_rc, "math_sdds1_rc", args, nargs);

	    for (i = 0; i < XtNumber(Sdds_labels_strings); i ++)
		math_sdds1_toggle[i] = XmCreateToggleButton(math_sdds1_rc, 
				    Sdds_labels_strings[i], NULL, 0);

	    /* setting initial radio box toggle state */
	    XmToggleButtonSetState(math_sdds1_toggle[NOREQ], True, False);
	    Math_sdds_output_spec1 = NOREQ;

	    XtManageChildren(math_sdds1_toggle, XtNumber(Sdds_labels_strings));

	XtManageChild(math_sdds1_rc);

	/* math buttons rowcolumn */
	nargs = 0;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	math_buttons1_rc = XmCreateRowColumn(math_rc, 
		"math_buttons1_rc", 
		args, nargs);

	    math_go1_button = XmCreatePushButton(math_buttons1_rc, 
				    "math_go1_button", NULL, 0);
	    XtManageChild(math_go1_button);

	    Math_status1_label = 
		XmCreateLabel(math_buttons1_rc, "Math_status1_label", NULL, 0);
	    set_ok(Math_status1_label);
	    XtManageChild(Math_status1_label);

	    math_view_log1_button = XmCreatePushButton(math_buttons1_rc, 
				    "math_view_log1_button", NULL, 0);
	    XtManageChild(math_view_log1_button);

	    math_print_log1_button = XmCreatePushButton(math_buttons1_rc, 
				    "math_print_log1_button", NULL, 0);
	    XtManageChild(math_print_log1_button);

	    math_help1_button = XmCreatePushButton(math_buttons1_rc, 
				"math_help1_button", NULL, 0);
	    XtManageChild(math_help1_button);

	    math_done1_button = XmCreatePushButton(math_buttons1_rc, 
				"math_done1_button", NULL, 0);
	    XtManageChild(math_done1_button);

	XtManageChild(math_buttons1_rc);

    XtManageChild(math_rc);

    /* math view op1 snap shell */
    Math_op1_view_snap_shell = 
	XtVaCreateWidget("Snapshot File Viewer (Operand 1)", 
		    topLevelShellWidgetClass, Math_shell,
		    XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
		    NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 350); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    math_op1_view_snap_form = XtCreateWidget("math_op1_view_snap_form", 
				xmFormWidgetClass, Math_op1_view_snap_shell,
				args, nargs);

    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    math_op1_view_snap_rc = XmCreateRowColumn(math_op1_view_snap_form, 
			    "math_op1_view_snap_rc", args, nargs);

	nargs = 0;
	XtSetArg(args[nargs], XmNeditMode, XmMULTI_LINE_EDIT); nargs ++;
	XtSetArg(args[nargs], XmNeditable, False); nargs ++;
	XtSetArg(args[nargs], XmNheight, 300); nargs ++;
	Math_op1_view_snap_scroll_txt = 
	    XmCreateScrolledText(math_op1_view_snap_rc, 
		"math_op1_view_math_scroll_txt", args, nargs);
	XtManageChild(Math_op1_view_snap_scroll_txt);

	math_op1_view_snap_done_button = 
	    XmCreatePushButton(math_op1_view_snap_rc, 
		"math_op1_view_snap_done_button", NULL, 0);
	XtManageChild(math_op1_view_snap_done_button);

    XtManageChild(math_op1_view_snap_rc);
    XtManageChild(math_op1_view_snap_form);
    XtManageChild(Math_op1_view_snap_shell);

    /* math view op2 snap shell */
    Math_op2_view_snap_shell = 
	XtVaCreateWidget("Snapshot File Viewer (Operand 2)", 
		    topLevelShellWidgetClass, Math_shell,
		    XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
		    NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 350); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    math_op2_view_snap_form = XtCreateWidget("math_op2_view_snap_form", 
				xmFormWidgetClass, Math_op2_view_snap_shell,
				args, nargs);

    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    math_op2_view_snap_rc = XmCreateRowColumn(math_op2_view_snap_form, 
			    "math_op2_view_snap_rc", args, nargs);

	nargs = 0;
	XtSetArg(args[nargs], XmNeditMode, XmMULTI_LINE_EDIT); nargs ++;
	XtSetArg(args[nargs], XmNeditable, False); nargs ++;
	XtSetArg(args[nargs], XmNheight, 300); nargs ++;
	Math_op2_view_snap_scroll_txt = 
	    XmCreateScrolledText(math_op2_view_snap_rc, 
		"math_op2_view_math_scroll_txt", args, nargs);
	XtManageChild(Math_op2_view_snap_scroll_txt);

	math_op2_view_snap_done_button = 
	    XmCreatePushButton(math_op2_view_snap_rc, 
		"math_op2_view_snap_done_button", NULL, 0);
	XtManageChild(math_op2_view_snap_done_button);

    XtManageChild(math_op2_view_snap_rc);
    XtManageChild(math_op2_view_snap_form);
    XtManageChild(Math_op2_view_snap_shell);

    /* math log1 shell */
    Math_log1_shell = XtVaCreateWidget("Addition/Subtraction Log Viewer",
			topLevelShellWidgetClass, Math_shell,
			XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
			NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 350); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    math_log1_form = XtCreateWidget("math_log1_form", xmFormWidgetClass, 
		    Math_log1_shell, 
		    args, nargs);

    /* main math log rowcolumn */
    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    math_log1_rc = XmCreateRowColumn(math_log1_form, "math_log1_rc", 
		    args, nargs);

	nargs = 0;
	XtSetArg(args[nargs], XmNeditMode, XmMULTI_LINE_EDIT); nargs ++;
	XtSetArg(args[nargs], XmNeditable, False); nargs ++;
	XtSetArg(args[nargs], XmNheight, 300); nargs ++;
	Math_log1_scroll_txt = XmCreateScrolledText(math_log1_rc, 
				"Math_log1_scroll_txt", args, nargs);
	XtManageChild(Math_log1_scroll_txt);

	math_log_done1_button = XmCreatePushButton(math_log1_rc, 
				"math_log_done1_button", 
				NULL, 0);
	XtManageChild(math_log_done1_button);

    XtManageChild(math_log1_rc);
    XtManageChild(math_log1_form);
    XtManageChild(Math_log1_shell);

    XtManageChild(math_form);
    XtManageChild(Math_shell);

    /*************/
    /*           */
    /* Callbacks */
    /*           */
    /*************/

    XtAddCallback(math_go1_button, XmNactivateCallback, math_go_button1_cb, 
	NULL);
    XtAddCallback(math_view_log1_button, XmNactivateCallback, 
	math_view_log_button1_cb, NULL);
    XtAddCallback(math_print_log1_button, XmNactivateCallback, 
	shared_print_log_button_cb, (char *) Math_logfilename);
    XtAddCallback(math_help1_button, XmNactivateCallback, 
	shared_temp_help_button_cb, NULL);
    XtAddCallback(math_done1_button, XmNactivateCallback, 
	shared_done_button_cb, (Widget) Math_shell);
    XtAddCallback(math_op1_snap_file_sel_button, XmNactivateCallback, 
	math_op1_snap_file_sel_button_cb, (Widget) math_op1_snap_file_sel);
    XtAddCallback(math_op2_snap_file_sel_button, XmNactivateCallback, 
	math_op2_snap_file_sel_button_cb, (Widget) math_op2_snap_file_sel);
    XtAddCallback(math_op1_view_snap_button, XmNactivateCallback, 
	math_view_snap_button_cb, NULL);
    XtAddCallback(math_op2_view_snap_button, XmNactivateCallback, 
	math_view_snap_button_cb, NULL);
    XtAddCallback(math_op1_print_snap_button, XmNactivateCallback, 
	math_print_snap_button_cb, NULL);
    XtAddCallback(math_op2_print_snap_button, XmNactivateCallback, 
	math_print_snap_button_cb, NULL);
    XtAddCallback(math_op1_snap_file_sel, XmNcancelCallback, 
	XtUnmanageChild, NULL);
    XtAddCallback(math_op2_snap_file_sel, XmNcancelCallback, 
	XtUnmanageChild, NULL);
    XtAddCallback(math_op1_snap_file_sel, XmNokCallback, math_file_sel_ok_cb, 
	(Widget) Math_op1_text);
    XtAddCallback(math_op2_snap_file_sel, XmNokCallback, math_file_sel_ok_cb, 
	(Widget) Math_op2_text);
    XtAddCallback(math_op1_view_snap_done_button, XmNactivateCallback, 
	shared_done_button_cb, (Widget) Math_op1_view_snap_shell);
    XtAddCallback(math_op2_view_snap_done_button, XmNactivateCallback, 
	shared_done_button_cb, (Widget) Math_op2_view_snap_shell);
    for (i = 0; i < XtNumber(Math_ops_labels_strings); i ++)
	XtAddCallback(math_toggle[i], XmNvalueChangedCallback,
	    math_operation_toggle_cb, (int) i);
    /* math log1 shell */
    XtAddCallback(math_log_done1_button, XmNactivateCallback, 
	shared_done_button_cb, (Widget) Math_log1_shell);
    for (i = 0; i < XtNumber(Sdds_labels_strings); i ++)
	XtAddCallback(math_sdds1_toggle[i], XmNvalueChangedCallback,
	    math_sdds_toggle1_cb, (int) i);

} /* end setup_math() */

/************************/
/*                      */
/* math shell callbacks */
/*                      */
/************************/

/****************************************************************
*
*
****************************************************************/

static void math_op1_snap_file_sel_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

    XtManageChild((Widget) client_data);
    XtPopup(XtParent((Widget) client_data), XtGrabNone);

} /* end math_op1_snap_file_sel_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void math_op2_snap_file_sel_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

    XtManageChild((Widget) client_data);
    XtPopup(XtParent((Widget) client_data), XtGrabNone);

} /* end math_op2_snap_file_sel_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void math_view_snap_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

XmStringTable choices;
int n_choices;
int i, j;
int total_bytes;
char *total_buff;
int fd;
char mbuff[1000];
struct filerep fbuff;
char *cp;
char *fname;
struct stat stat_buff;
BOOL all_ascii;

    if (!(strcmp(XtName(w), MATHOP1VIEWSNAPBUTTONNAME)))
    {
	/* op1 */
	XtVaSetValues(Math_op1_view_snap_scroll_txt,
	    XmNvalue, "",
	    NULL);
	fname = XmTextFieldGetString(Math_op1_text);
    }
    else 
    {
	/* op2 */
	XtVaSetValues(Math_op2_view_snap_scroll_txt,
	    XmNvalue, "",
	    NULL);
	fname = XmTextFieldGetString(Math_op2_text);
    } /* endif */

    if (*fname)
    {
	if ((fd = open(fname, O_RDONLY)) != -1)
	{
	    if (!fstat(fd, &stat_buff))
	    {
		if (S_ISREG(stat_buff.st_mode))
		{
		    if (fbuff.buff = (char *) calloc (stat_buff.st_size,1))
		    {
			if (read(fd, fbuff.buff, stat_buff.st_size) != -1)
			{
			    /* loop to check file is printable */
			    for (j = 0, all_ascii = TRUE; 
				all_ascii && j < stat_buff.st_size; 
				    j ++)
					all_ascii = isascii(fbuff.buff[j]);

			    if (all_ascii)
			    {
				fbuff.nbytes = stat_buff.st_size;
				fbuff.fname = strdup(fname);

				/* now filling scrolled text */
				total_bytes = 
			(fbuff.nbytes+strlen(fbuff.fname)+1+strlen("\n\n"));
				/* for null char at end */
				total_bytes ++;

				if (total_bytes > 1)
				{
				    /* there are files to view */
				    if (total_buff = 
					    (char *) calloc(total_bytes, 1))
				    {
					cp = total_buff;

					/* appending (via copy) file to */
					/* total_buff and then free mem */
					/* from fbuff */
					memcpy(cp, fbuff.fname, 
					    strlen(fbuff.fname));
					cp = cp + strlen(fbuff.fname);

					*cp = '\n';
					cp ++;

					memcpy(cp, fbuff.buff, fbuff.nbytes);
					cp = cp + fbuff.nbytes;

					memcpy(cp, "\n\n", strlen("\n\n"));
					cp = cp + strlen("\n\n");

					free(fbuff.buff);
					free(fbuff.fname);

					/* although buffer was allocated with */
					/* calloc which presumably allocates */
					/* a block with all zeros, still     */ 
					/* need to explicitly place the null */
					/* char at the end  so as not to get */
					/* extra garbage characters in the   */
					/* scroll text ... can't explain why */
					/* this works ... just happy that it */
					/* does.                             */

					*cp = '\0';

					if (strcmp(XtName(w), 
					    MATHOP2VIEWSNAPBUTTONNAME))
				    XtVaSetValues(Math_op1_view_snap_scroll_txt,
						XmNvalue, total_buff,
						NULL);
					else
				    XtVaSetValues(Math_op2_view_snap_scroll_txt,
						XmNvalue, total_buff,
						NULL);

					free(total_buff);
				    }
				    else
				    {
					sprintf(mbuff, 
					"SYS FAULT: failed calloc() %d bytes.", 
					    total_bytes);
					popup_message(mbuff);
				    } /* endif */
				} /* endif */
			    }
			    else
			    {
				free(fbuff.buff);

				sprintf(mbuff, "ERROR: file %s not viewable.",
				    fname);
				popup_error(mbuff);
			    } /* endif */
			}
			else
			{
			    free(fbuff.buff);

			    sprintf(mbuff, 
				"SYS FAULT: could not read() file %s.", fname);
			    popup_message(mbuff);
			} /* endif */

		    }
		    else
		    {
			sprintf(mbuff, "SYS FAULT: failed calloc() %d bytes.", 
			    stat_buff.st_size);
			popup_message(mbuff);
		    } /* endif */
		}
		else
		{
		    sprintf(mbuff, 
		"NOTE: cannot view file %s because it is not a regular file.", 
			fname);
		popup_message(mbuff);
		} /* endif */
	    }
	    else
		popup_message("SYS FAULT: error in fstat().");

	    close(fd);
	}
	else
	{
	    sprintf(mbuff, "SYS FAULT: cannot open() file %s.", fname);
	    popup_message(mbuff);
	} /* endif */
    } /* endif */

    XtFree(fname);

    if (strcmp(XtName(w), MATHOP2VIEWSNAPBUTTONNAME))
	XtMapWidget(Math_op1_view_snap_shell);
    else
	XtMapWidget(Math_op2_view_snap_shell);

} /* end math_view_snap_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void math_print_snap_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

char *fname;

    if (!(strcmp(XtName(w), MATHOP1PRINTSNAPBUTTONNAME)))
	/* op2 */
	fname = XmTextFieldGetString(Math_op1_text);
    else
	/* op2 */
	fname = XmTextFieldGetString(Math_op2_text);

    print_file(fname);
    XtFree(fname);

} /* end math_print_snap_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void math_go_button1_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

int rc;
char *cp;
char command[CMDSIZE];
char *op1_fname, *op2_fname, *result_fname;
BOOL ok;

    ok = TRUE;

    op1_fname = XmTextFieldGetString(Math_op1_text);
    op2_fname = XmTextFieldGetString(Math_op2_text);
    result_fname = XmTextFieldGetString(Math1_text[0]);

    if (*op1_fname && *op2_fname && *result_fname)
    {
	if (strlen("burtmath") < CMDSIZE)
	    strcpy(command, "burtmath");
	else
	    ok = FALSE;

	if (ok &&
	    (((strlen(command) + 
		1+strlen(op1_fname) + 1+strlen(op2_fname)
		+strlen(" -o")+ 1+strlen(result_fname)
		+strlen(" -v")
		+5 /* for -add or -sub */
		+strlen(" -l" + 1+strlen(Math_logfilename)))+1) < CMDSIZE))
	{
	    strcat(command, " ");
	    strcat(command, op1_fname);
	    strcat(command, " ");
	    strcat(command, op2_fname);

	    switch (Math_operation)
	    {
		case ADDITION: strcat(command, " -add"); break;
		case SUBTRACTION: strcat(command, " -sub"); break;
		default:
		    /* unrecognizable command ... should never get here */
		popup_message("SYS FAULT: Unrecognizable operator specified.");
		    ok = FALSE;
		    break;
	    } /* end switch() */

	    strcat(command, " -o");
	    strcat(command, " ");
	    strcat(command, result_fname);

	    strcat(command, " -v");

	    unlink(Math_logfilename);
	    strcat(command, " -l");
	    strcat(command, " ");
	    strcat(command, Math_logfilename);

	} 
	else
	    ok = FALSE;

        cp = XmTextFieldGetString(Math1_text[1]);
        if (*cp)
        {
            if (ok && (strlen(command) + 4+strlen(cp)) < CMDSIZE)
            {
                strcat(command, " -c\"");
                strcat(command, cp);
                strcat(command, "\"");
            }
            else
                ok = FALSE;
        } /* endif */
        XtFree(cp);

        cp = XmTextFieldGetString(Math1_text[2]);
        if (*cp)
        {
            if (ok && (strlen(command) + 4+strlen(cp)) < CMDSIZE)
            {
                strcat(command, " -k \"");
                strcat(command, cp);
                strcat(command, "\"");
            }
            else
                ok = FALSE;
        } /* endif */
        XtFree(cp);


        if (ok)
        {
            if (Math_sdds_output_spec1 == REQSDDS)
            {
                if (((strlen(command) + strlen(" -sdds"))+1) < CMDSIZE)
                    strcat(command, " -sdds");
                else
                    ok = FALSE;
            }
            else if (Math_sdds_output_spec1 == REQNOSDDS)
            {
                if (((strlen(command) + strlen(" -nosdds"))+1) < CMDSIZE)
                    strcat(command, " -nosdds");
                else
                    ok = FALSE;
            } /* endif */
        } /* endif */

	if (ok)
	{
	    set_running(Math_status1_label);
	    clear_scrolled_text(Math_log1_scroll_txt);
	    
printf(">%s<\n", command);
	    /* calling burtmath */
	    rc = system(command);

	    /* dumping logfile into scrolled text */
	    file_to_scrolled_text(Math_logfilename, Math_log1_scroll_txt);

	    if (WIFEXITED(rc))
		if (WEXITSTATUS(rc))
		    set_not_ok(Math_status1_label);
		else
		    set_ok(Math_status1_label);
	    else
	    popup_message("SYS FAULT: system() returned but cmd not exited.");

	}
	else
popup_message("SYS FAULT: could not construct command ... not long enough.");
    } 
    else
    {
	if (!(*op1_fname) || !(*op2_fname))
	    popup_error("ERROR: must specify two snapshot files as operands.");

	if (!(*result_fname))
	    popup_error("ERROR: must specify a result file.");
    } /* endif */

    XtFree(op1_fname);
    XtFree(op2_fname);
    XtFree(result_fname);

} /* end math_go_button1_cb() */

/****************************************************************
*
*
****************************************************************/

static void math_view_log_button1_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    XtMapWidget(Math_log1_shell);
} /* end math_view_log_button1_cb() */

/****************************************************************
*
* called for BOTH operands of the math operation.
*
****************************************************************/

static void math_file_sel_ok_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

char *cp;
Widget text_field;

    text_field = (Widget) client_data;

    cp = XmTextFieldGetString(XmFileSelectionBoxGetChild(w, XmDIALOG_TEXT));

    XtVaSetValues(text_field,
	XmNvalue, cp,
	NULL);

    XtFree(cp);

} /* end math_file_sel_ok_cb() */

/****************************************************************
*
*
****************************************************************/

static void math_operation_toggle_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

    if (XmToggleButtonGetState(w))
	Math_operation = (int) client_data;

} /* end math_operation_toggle_cb() */

/****************************************************************
*
*
****************************************************************/

static void math_sdds_toggle1_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

    if (XmToggleButtonGetState(w))
	Math_sdds_output_spec1 = (int) client_data;

} /* end math_sdds_toggle1_cb() */

