/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#define JRW_ALLOW_BLANK_VALUES

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h> /* for stat() */
#include <sys/stat.h>  /* for stat() */
#include <ctype.h>     /* for isspace() */
#include <sys/types.h> /* for time() and geteuid() */
#include <sys/time.h>  /* for time() and wait3()*/
#include <pwd.h>       /* for getpwuid() */

/* for wait3() */
#include <sys/wait.h>
#include <sys/resource.h>

#include <sys/unistd.h>

#include "burtcommon.h"
#include "burtobjs.h"

#include "list.h"
/*#include "cursor.h"*/
#include "burtlist.h"

#include "SDDS.h"

#ifndef NODEVICES
#include "libdev.h"
#endif 

/***************************************/
/*                                     */
/* Externally Visibile Local Functions */
/*                                     */
/***************************************/

void BurtCommonInit(const BOOL);
BOOL BurtCreateUniqueFilename(const char *const, const char *const,char *const);
void BurtExit(const int);
int BurtLongestLineLength(const char *const);
void BurtPerformArithmetic(const Burtobjlistlist *const, Burtobjlist *const, 
    const char, const double);
void BurtPipeWait(int *const, int *const);
BOOL BurtPreprocessFile(const char *const, const char **const, 
    const char **const, char *const);
void BurtProcessRequestFile(char *const, Burtobjlist *const);
Burtobjlist *BurtProcessSnapshotFile(char *const);
Burtobjlist *BurtProcessSnapshotFile(char *const, Burtobjlist *const);
BOOL BurtSDDSColExists(SDDS_TABLE *const, char *const, const long);
BOOL BurtSDDSParmExists(SDDS_TABLE *const, char *const);
BOOL BurtSDDSFile(const char *const);

extern "C" {
/* Channel Access Interface Functions */
void BurtInitializeChannelAccess();
BOOL BurtCaArrayPut(Channel *);
void BurtCaFlushIo();
BOOL BurtQueueCaArrayGetCallback(Channel *);
void BurtQueueCaPendEvent(float);
BOOL BurtQueueCaPendIO();
void BurtQueueCaSearchAndConnect(Channel *);
} /* end extern c thing */

/***************************/
/*                         */
/* Private Local Functions */
/*                         */
/***************************/

#ifndef NODEVICES
static int dev_count(const char *const, devnamearr *const);
static void dev_info(const char *const, devstructarr *const);
#endif
static BOOL file_exists(const char *const);
static BOOL get_through_header(FILE *const, Burtobjlist *const, 
    char *const, const int);
static BOOL nextchar(FILE *const, const char);
static char *prepare_val(char *const, const int);
#ifndef NODEVICES
static void process_device(Burtobjlist *const, const char *const, 
    const char *const, SDDS_TABLE *const, const int, const BOOL, const BOOL, 
    const BOOL, const BOOL);
#endif
static BOOL process_optional_request_and_end(FILE *const, Burtobjlist *const, 
    char *const, const int);
static BOOL sdds_get_parm(SDDS_TABLE *const, char *const, void *const);
static void sdds_get_value(SDDS_TABLE *const, char *const, const long, 
    void *const);

/* Callbacks */
extern "C" void my_connection_callback(struct connection_handler_args);
extern "C" void my_get_callback(struct event_handler_args);

/****************************************************************
*
*
****************************************************************/

void BurtCommonInit(const BOOL need_channel_access)
{

    Outfilename = (char *) NULL;

    Sh_path = "/bin/sh";
    Sh_arg1 = "-c";

    Exec_args[0] = Sh_path;
    Exec_args[1] = Sh_arg1;
    Exec_args[2] = (char *) NULL;
    Exec_args[3] = (char *) NULL;

    if (need_channel_access)
	BurtInitializeChannelAccess();

} /* end BurtCommonInit() */

/****************************************************************
*
* it is assumed that filename is a character string at 
* least MAXPATHLEN (from <sys/param.h> big.
*
****************************************************************/

BOOL BurtCreateUniqueFilename(const char *const prefix,
    const char *const suffix, char *const filename)
{
struct passwd *pp;
pid_t mpid;
BOOL rc;

    if (pp = getpwuid(geteuid()))
    {
        if(mpid = getpid())
        {
            sprintf(filename, "%s%s_%d%s",
                (prefix ? prefix : ""), pp->pw_name, mpid,(suffix ? suffix : ""));
            //printf("PID: %d tempname: %s \n",mpid, filename);
            rc = 1;
        }
        else
        {
            fprintf(Log_fp,
                "ERROR: BurtCreateUniqueFilename() could not get pid\n");
            fflush(Log_fp);
            BurtExit(1);
        }
    }
    else
    {
        fprintf(Log_fp,
             "ERROR: BurtCreateUniqueFilename() could not id user\n");
        fflush(Log_fp);
        BurtExit(1);
    }
    return rc;

} /* end BurtCreateUniqueFilename() */


/****************************************************************
*
*
****************************************************************/

void BurtExit(const int exit_status)
{

    if (Verbose)
    {
	fprintf(Log_fp, "\n**** END BURT LOGFILE\n");
	fflush(Log_fp);
    } /* endif */
    ca_task_exit();
    exit(exit_status);

} /* end BurtExit() */

/****************************************************************
*
*
****************************************************************/

int BurtLongestLineLength(const char *const fname)
{

FILE *fp;
char c;
int this_line;
int rc;

    if (file_exists(fname))
    {
        if (fp = fopen(fname, "r"))
        {
            rc = 0;
            this_line = 0;

            while ((c = getc(fp)) != EOF)
            {
                this_line ++;

                if (c == '\n')
                {
                    /* end of line */
                    if (rc < this_line)
                        rc = this_line;
                    this_line = 0;
                } /* endif */
            } /* endwhile */

            if (rc < this_line)
                rc = this_line;

            fclose(fp);
        }
        else
            rc = UNKNOWN;
    }
    else
        rc = UNKNOWN;

   return rc;

} /* end BurtLongestLineLength() */

/****************************************************************
*
* Produces a resulting snapshot where all Values are DBR_STRING
* that are spaced dbr_size_n(DBR_STRING, 1) (about 40 bytes) apart.
*
****************************************************************/

void BurtPerformArithmetic(const Burtobjlistlist *const snapshots, 
    Burtobjlist *const result_snapshot, const char operation, 
    const double mult_fact)
{

double op1, op2;
int nelem, i;
Cursor<Burtobj> *op1cp;
Burtobj *op1p, *op2p;
Burtpvar *pp;
Burtatomdev *dp;
char *databuff, *tempbuff;
char *op1_cp, *op2_cp, *ret_cp;
short *op1_sp, *op2_sp;
unsigned short *op1_usp, *op2_usp;
unsigned char *op1_ucp, *op2_ucp;
long *op1_lp, *op2_lp;
double *op1_dp, *op2_dp;
float *op1_fp, *op2_fp;
Cursor<Filename> *fcp;
Filename *fp;
BOOL ok;

    if (!snapshots || !result_snapshot)
    {
	fprintf(Log_fp, 
	"ERROR: BurtPerformArithmetic() got snapshots %x result_snapshot %x\n", 
	    snapshots, result_snapshot);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    /* need an empty result_snapshot because setting relative/absolute */
    if (!(result_snapshot->empty()))
    {
	fprintf(Log_fp, 
	    "ERROR: BurtPerformArithmetic() found non-empty result_snapshot\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    switch (operation)
    {
	case ADDITION:
	case SUBTRACTION:
	    if (snapshots->nelem() == 2)
	    {
		(result_snapshot->get_parms())->set_relative();

		/* loop to process through first list */
		op1cp = new Cursor<Burtobj>((*snapshots)[0]);
		for (op1p = op1cp->getfirst(); op1p; op1p = op1cp->getnext())
		{
		    if ((op1p->get_val())->defined() 
			&& (op2p = ((*snapshots)[1])->find(op1p))
			&& (op2p->get_val())->defined())
		    {
			/* found op1 (lhs) in op2 (rhs) list */

			nelem = (op1p->get_val())->get_nelem();

			if (nelem == (op2p->get_val())->get_nelem())
			{
			    /* they both have same nelem */

			    /********************************************/
			    /*                                          */
			    /* setting inital pointers into op1 and op2 */
			    /*                                          */
			    /********************************************/

			    /* op1 */
			    switch ((op1p->get_val())->get_dbr_type())
			    {
				case DBR_STRING:
				    op1_cp = 
					(char *) (op1p->get_val())->get_pval();
				    break;
				/* case DBR_INT: */
				case DBR_SHORT:
				    op1_sp = 
					(short *) (op1p->get_val())->get_pval();
				    break;
				case DBR_FLOAT:
				    op1_fp = 
					(float *) (op1p->get_val())->get_pval();
				    break;
				case DBR_ENUM:
				    op1_usp = 
			    (unsigned short *) (op1p->get_val())->get_pval();
				    break;
				case DBR_CHAR:
				    op1_ucp = 
				(unsigned char *) (op1p->get_val())->get_pval();
				    break;
				case DBR_LONG:
				    op1_lp = 
					(long *) (op1p->get_val())->get_pval();
				    break;
				case DBR_DOUBLE:
				    op1_dp = 
				    (double *) (op1p->get_val())->get_pval();
				    break;
				default:
				    fprintf(Log_fp, 
	    "ERROR: BurtPerformArithmetic() unrecognizable dbr type %d for op1\n",
					(op1p->get_val())->get_dbr_type());
				    fflush(Log_fp);
				    BurtExit(1);
				    break;
			    } /* end switch() */

			    /* op2 */
			    switch ((op2p->get_val())->get_dbr_type())
			    {
				case DBR_STRING:
				    op2_cp = 
					(char *) (op2p->get_val())->get_pval();
				    break;
				/* case DBR_INT: */
				case DBR_SHORT:
				    op2_sp = 
					(short *) (op2p->get_val())->get_pval();
				    break;
				case DBR_FLOAT:
				    op2_fp = 
					(float *) (op2p->get_val())->get_pval();
				    break;
				case DBR_ENUM:
				    op2_usp = 
			    (unsigned short *) (op2p->get_val())->get_pval();
				    break;
				case DBR_CHAR:
				    op2_ucp = 
				(unsigned char *) (op2p->get_val())->get_pval();
				    break;
				case DBR_LONG:
				    op2_lp = 
					(long *) (op2p->get_val())->get_pval();
				    break;
				case DBR_DOUBLE:
				    op2_dp = 
				    (double *) (op2p->get_val())->get_pval();
				    break;
				default:
				    fprintf(Log_fp, 
	    "ERROR: BurtPerformArithmetic() unrecognizable dbr type %d for op2\n",
					(op2p->get_val())->get_dbr_type());
				    fflush(Log_fp);
				    BurtExit(1);
				    break;
			    } /* end switch() */

			    /* using calloc assures that the block zero'd */
			    if (!(databuff = 
			    (char *) calloc(dbr_size_n(DBR_STRING, nelem), 1)))
			    {
				fprintf(Log_fp, 
		"ERROR: BurtPerformArithmetic() unable to calloc() %d bytes\n",
				    dbr_size_n(DBR_STRING, nelem));
				fflush(Log_fp);
				BurtExit(1);
			    } /* endif */
			    *databuff = '\0';

			    /* loop to process through all the elements */
			    /* and place result into char databuff      */
			    for (i = 0, ok = TRUE, tempbuff = databuff; 
				ok && i < nelem; 
				i ++, tempbuff += dbr_size_n(DBR_STRING,1))
			    {

				/**************************************/
				/*                                    */
				/* getting op1 and op2 as double vals */
				/*                                    */
				/**************************************/

				/* op1 */
				switch ((op1p->get_val())->get_dbr_type())
				{
				    case DBR_STRING:
					op1 = strtod(op1_cp, &ret_cp);
					if (op1_cp != ret_cp)
					{
					    /* conversion OK */
					    op1_cp += dbr_size_n(DBR_STRING,1);
					}
					else
					{
					    /* converison failed */

					    /* DO NOT EXIT, may be a str val */
					    /* in which case we just proceed */

					    ok = FALSE;

					    fprintf(Log_fp, 
"WARNING: op1 (lhs) >%s< could not convert item %d of %d val >%s< (start %x idx %x) ... not processed\n",
						op1p->get_name(), i+1, 
						(op1p->get_val())->get_nelem(),
						(op1p->get_val())->get_pval(), 
						(op1p->get_val())->get_pval(), 
						op1_cp);
					    fflush(Log_fp);
					} /* endif */
					break;
				    /* case DBR_INT: */
				    case DBR_SHORT:
					op1 = (double) (*op1_sp);
					op1_sp ++;
					break;
				    case DBR_FLOAT:
					op1 = (double) (*op1_fp);
					op1_fp ++;
					break;
				    case DBR_ENUM:
					op1 = (double) (*op1_usp);
					op1_usp ++;
					break;
				    case DBR_CHAR:
					op1 = (double) (*op1_ucp);
					op1_ucp ++;
					break;
				    case DBR_LONG: 
					op1 = (double) (*op1_lp); 
					op1_lp ++;
					break;
				    case DBR_DOUBLE: 
					op1 = (*op1_dp); 
					op1_dp ++;
					break;
				    default:
					fprintf(Log_fp, 
	"ERROR: BurtPerformArithmetic() unrecognizable dbr type %d for op1\n",
					    (op1p->get_val())->get_dbr_type());
					fflush(Log_fp);
					BurtExit(1);
					break;
				} /* end switch() */

				/* op2 */
				switch ((op2p->get_val())->get_dbr_type())
				{
				    case DBR_STRING:
					op2 = strtod(op2_cp, &ret_cp);
					if (op2_cp != ret_cp)
					{
					    /* conversion OK */
					    op2_cp += dbr_size_n(DBR_STRING,1);
					}
					else
					{
					    /* converison failed */

					    /* DO NOT EXIT, may be a str val */
					    /* in which case we just proceed */

					    ok = FALSE;

					    fprintf(Log_fp, 
"WARNING: op2 (rhs) >%s< could not convert item %d of %d val >%s< (start %x idx %x) ... not processed\n",
						op2p->get_name(), i+1, 
						(op2p->get_val())->get_nelem(),
						(op2p->get_val())->get_pval(), 
						(op2p->get_val())->get_pval(), 
						op2_cp);
					    fflush(Log_fp);
					} /* endif */
					break;
				    /* case DBR_INT: */
				    case DBR_SHORT:
					op2 = (double) (*op2_sp);
					op2_sp ++;
					break;
				    case DBR_FLOAT:
					op2 = (double) (*op2_fp);
					op2_fp ++;
					break;
				    case DBR_ENUM:
					op2 = (double) (*op2_usp);
					op2_usp ++;
					break;
				    case DBR_CHAR:
					op2 = (double) (*op2_ucp);
					op2_ucp ++;
					break;
				    case DBR_LONG: 
					op2 = (double) (*op2_lp); 
					op2_lp ++;
					break;
				    case DBR_DOUBLE: 
					op2 = (*op2_dp); 
					op2_dp ++;
					break;
				    default:
					fprintf(Log_fp, 
	"ERROR: BurtPerformArithmetic() unrecognizable dbr type %d for op2\n",
					    (op2p->get_val())->get_dbr_type());
					fflush(Log_fp);
					BurtExit(1);
					break;
				} /* end switch() */

				/* now have two operands in op1 and op2 */
				if (ok)
				{
				    if (operation == ADDITION)
					sprintf(tempbuff, DOUBLEFORMATSTRING, 
					    op1+op2);
				    else if (operation == SUBTRACTION)
					sprintf(tempbuff, DOUBLEFORMATSTRING, 
					    op1-op2);
				    else
				    {
					fprintf(Log_fp, 
	"ERROR: BurtPerformArithmetic() found unrecognizable operation %d\n", 
					    operation);
					fflush(Log_fp);
					BurtExit(1);
				    } /* endif */
				} /* endif */
			    } /* endfor */

			    if (ok)
			    {
				if (op1p->isa() == PVAR)
				{
				    pp = new Burtpvar;
				    (pp->get_channel())->
					set_val(pp->get_val());

				    (pp->get_channel())->
					set_name(op1p->get_name());

				    (pp->get_val())->
					set(DBR_STRING, nelem, databuff);

				    /* preserving read only */

				    if (op1p->read_only_notify() 
					    || op2p->read_only_notify())
				    {
					pp->set_read_only_notify();
					pp->set_read_only();
				    }
				    else
				    {
					pp->clear_read_only();
					
					if (op1p->read_only() 
						|| op2p->read_only())
					    pp->set_read_only();
					else
					    pp->clear_read_only();
				    } /* endif */

				    result_snapshot->append((Burtobj *) pp);

				}
				else
				{
				    dp = new Burtatomdev;
				    (dp->get_readchannel())->
					set_val(dp->get_val());
				    (dp->get_writechannel())->
					set_val(dp->get_val());

				    dp->set_name(op1p->get_name());
				    dp->set_lineage(op1p->get_lineage());

				    dp->set_readmsg(op1p->get_readmsg());
				    (dp->get_readchannel())->
			set_name((op1p->get_readchannel())->get_name());

				    dp->set_writemsg(op1p->get_writemsg());
				    (dp->get_writechannel())->
			set_name((op1p->get_writechannel())->get_name());

				    (dp->get_val())->
					set(DBR_STRING, nelem, databuff);

				    /* preserving read only */
				    if (op1p->read_only_notify() 
					    || op2p->read_only_notify())
				    {
					dp->set_read_only_notify();
					dp->set_read_only();
				    }
				    else
				    {
					dp->clear_read_only_notify();

					if (op1p->read_only() 
						|| op2p->read_only())
					    dp->set_read_only();
					else
					    dp->clear_read_only();
				    } /* endif */

				    result_snapshot->append((Burtobj *) dp);

				} /* endif */

				free(databuff);
			    } /* endif */
			} 
			else
			{
			    fprintf(Log_fp, 
"WARNING: >%s< lhs has %d elements rhs has %d elements ... not processed\n",
				op1p->get_name(),
				(op1p->get_val())->get_nelem(),
				(op2p->get_val())->get_nelem());
			    fflush(Log_fp);
			} /* endif */
		    } /* endif */
		} /* endfor */
		delete op1cp;

	    } 
	    else
	    {
		fprintf(Log_fp, 
	"ERROR: BurtPerformArithmetic() has operation >%s< with %d operands\n", 
		    (operation == ADDITION ? "addition" 
		    : (operation == SUBTRACTION ? "subtraction" : "unknown")), 
		    snapshots->nelem());
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */
	    break;
	case MULTIPLICATION:
	    if (snapshots->nelem() == 1)
	    {
		/* preserving snapshot type */

		if ((((*snapshots)[0])->get_parms())->is_absolute())
		    (result_snapshot->get_parms())->set_absolute();
		else if ((((*snapshots)[0])->get_parms())->is_relative())
		    (result_snapshot->get_parms())->set_relative();
		else if ((((*snapshots)[0])->get_parms())->is_nowrite())
		    (result_snapshot->get_parms())->set_nowrite();
		else
		{
		    fprintf(Log_fp, 
"ERROR: BurtPerformArithmetic() encountered unrecognizable snapshot type\n");
		    fflush(Log_fp);
		    BurtExit(1);
		} /* endif */

		/* copying request filenames */
		fcp = new Cursor<Filename>(((*snapshots)[0])->
			look_at_request_filenames());
		for (fp = fcp->getfirst(); fp; fp = fcp->getnext())
		    (result_snapshot->get_request_filenames())->
			append(new Filename(fp->get_dir(), fp->get_fname()));
		delete fcp;

		/* loop to process through snapshot */
		op1cp = new Cursor<Burtobj>((*snapshots)[0]);
		for (op1p = op1cp->getfirst(); op1p; op1p = op1cp->getnext())
		{
		    if ((op1p->get_val())->defined())
		    {
			nelem = (op1p->get_val())->get_nelem();

			/***********************************/
			/*                                 */
			/* setting inital pointer into op1 */
			/*                                 */
			/***********************************/

			/* op1 */
			switch ((op1p->get_val())->get_dbr_type())
			{
			    case DBR_STRING:
				op1_cp = (char *) (op1p->get_val())->get_pval();
				break;
			    /* case DBR_INT: */
			    case DBR_SHORT:
				op1_sp = 
				    (short *) (op1p->get_val())->get_pval();
				break;
			    case DBR_FLOAT:
				op1_fp = 
				    (float *) (op1p->get_val())->get_pval();
				break;
			    case DBR_ENUM:
				op1_usp = 
			    (unsigned short *) (op1p->get_val())->get_pval();
				break;
			    case DBR_CHAR:
				op1_ucp = 
			    (unsigned char *) (op1p->get_val())->get_pval();
				break;
			    case DBR_LONG:
				op1_lp = (long *) (op1p->get_val())->get_pval();
				break;
			    case DBR_DOUBLE:
				op1_dp = 
				    (double *) (op1p->get_val())->get_pval();
				break;
			    default:
				fprintf(Log_fp, 
	"ERROR: BurtPerformArithmetic() unrecognizable dbr type %d for op1\n",
				    (op1p->get_val())->get_dbr_type());
				fflush(Log_fp);
				BurtExit(1);
				break;
			} /* end switch() */

			/* using calloc to assure that the block zero'd */
			if (!(databuff = 
			    (char *) calloc(dbr_size_n(DBR_STRING, nelem), 1)))
			{
			    fprintf(Log_fp, 
		"ERROR: BurtPerformArithmetic() unable to calloc() %d bytes\n",
				dbr_size_n(DBR_STRING, nelem));
			    fflush(Log_fp);
			    BurtExit(1);
			} /* endif */

			/* loop to process through all the elements */
			/* and place result into char databuff      */
			for (i = 0, ok = TRUE, tempbuff = databuff; 
			    ok && i < nelem; 
				i ++, tempbuff += dbr_size_n(DBR_STRING,1))
			{

			    /*****************************/
			    /*                           */
			    /* getting op1 as double val */
			    /*                           */
			    /*****************************/

			    /* op1 */
			    switch ((op1p->get_val())->get_dbr_type())
			    {
				case DBR_STRING:
				    op1 = strtod(op1_cp, &ret_cp);
				    if (op1_cp != ret_cp)
				    {
					/* conversion OK */
					op1_cp += dbr_size_n(DBR_STRING,1);
				    }
				    else
				    {
					/* converison failed */

					/* DO NOT EXIT, may be a str val */
					/* in which case we just proceed */

					ok = FALSE;

					fprintf(Log_fp, 
"WARNING: op1 (lhs) >%s< could not convert item %d of %d val >%s< (start %x idx %x) ... not processed\n",
					    op1p->get_name(), i+1, 
					    (op1p->get_val())->get_nelem(),
					    (op1p->get_val())->get_pval(), 
					    (op1p->get_val())->get_pval(), 
					    op1_cp);
					fflush(Log_fp);
				    } /* endif */
				    break;
				/* case DBR_INT: */
				case DBR_SHORT:
				    op1 = (double) (*op1_sp);
				    op1_sp ++;
				    break;
				case DBR_FLOAT:
				    op1 = (double) (*op1_fp);
				    op1_fp ++;
				    break;
				case DBR_ENUM:
				    op1 = (double) (*op1_usp);
				    op1_usp ++;
				    break;
				case DBR_CHAR:
				    op1 = (double) (*op1_ucp);
				    op1_ucp ++;
				    break;
				case DBR_LONG: 
				    op1 = (double) (*op1_lp); 
				    op1_lp ++;
				    break;
				case DBR_DOUBLE: 
				    op1 = (*op1_dp); 
				    op1_dp ++;
				    break;
				default:
				    fprintf(Log_fp, 
	"ERROR: BurtPerformArithmetic() unrecognizable dbr type %d for op1\n",
					(op1p->get_val())->get_dbr_type());
				    fflush(Log_fp);
				    BurtExit(1);
				    break;
			    } /* end switch() */

			    /* got op1 */
			    if (ok)
				sprintf(tempbuff, DOUBLEFORMATSTRING, 
					    op1*mult_fact);
			} /* endfor */

			if (ok)
			{
			    if (op1p->isa() == PVAR)
			    {
				pp = new Burtpvar;
				(pp->get_channel())-> set_val(pp->get_val());

				(pp->get_channel())->
				    set_name(op1p->get_name());

				(pp->get_val())->
				    set(DBR_STRING, nelem, databuff);

				/* preserving read only */
				if (op1p->read_only_notify())
				{
				    pp->set_read_only_notify();
				    pp->set_read_only();
				}
				else
				{
				    pp->clear_read_only_notify();

				    if (op1p->read_only())
					pp->set_read_only();
				    else
					pp->clear_read_only();
				} /* endif */

				result_snapshot->append((Burtobj *) pp);

			    }
			    else
			    {
				dp = new Burtatomdev;
				(dp->get_readchannel())->
				    set_val(dp->get_val());
				(dp->get_writechannel())->
				    set_val(dp->get_val());

				dp->set_name(op1p->get_name());
				dp->set_lineage(op1p->get_lineage());

				dp->set_readmsg(op1p->get_readmsg());
				(dp->get_readchannel())->
			    set_name((op1p->get_readchannel())->get_name());

				dp->set_writemsg(op1p->get_writemsg());
				(dp->get_writechannel())->
			    set_name((op1p->get_writechannel())->get_name());

				(dp->get_val())->
				    set(DBR_STRING, nelem, databuff);

				/* preserving read only */
				if (op1p->read_only_notify())
				{
				    dp->set_read_only_notify();
				    dp->set_read_only();
				}
				else
				{
				    dp->clear_read_only_notify();

				    if (op1p->read_only())
					dp->set_read_only();
				    else
					dp->clear_read_only();
				} /* endif */


				result_snapshot->append((Burtobj *) dp);

			    } /* endif */

			    free(databuff);
			} /* endif */
		    } /* endif */
		} /* endfor */
		delete op1cp;

	    }
	    else
	    {
		fprintf(Log_fp, 
"ERROR: BurtPerformArithmetic() has operation >multiplication< with %d operands\n",
		    snapshots->nelem());
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */

	    break;
	default:
	    fprintf(Log_fp, 
	    "ERROR: BurtPerformArithmetic() found unrecognizable operator %d\n", 
		operation);
	    fflush(Log_fp);
	    BurtExit(1);
	    break;
    } /* end switch() */

} /* end BurtPerformArithmetic() */

/****************************************************************
*
* Wrapper around wait(), to wait for a child process to signal this process.
* Fill in pid, exit_status, and signal with the appropriate values.
* If block==1, then make it a blocking wait, otherwise make it a nonblocking
* wait.
*
****************************************************************/

void BurtPipeWait(int *const exit_status, int *const signal)
{

int status;
int status_int;
int status_byte;
int pid;

    /* a blocking wait */
#if 0
    pid = (int) wait3((union wait *) &status, 0, 0);
#else
	pid = (int) waitpid((pid_t)-1, &status, 0);
#endif
    /* pid = (int) wait3(&status, 0, 0); */

    if (pid <= 0)
    {
	*exit_status = -1;
	*signal = -1;
    }
    else
    {
	status_int = *((int *) &status);
	status_byte = status_int & 0xFF;
	if (status_byte == 0)
	{
	    /* child process exited due to exit call -- get the exit status */
	    *exit_status = (int) ((status_int & 0xFF00) >> 8);
	    *signal = -1;
	}
	else if (status_byte == 0177)
	{
	    /* child process has stopped */
	    *exit_status = -1;
	    *signal = -1;
	}
	else
	{
	    /* child process terminated due to a signal -- get the signal */
	    *exit_status = -1;
	    *signal = (int) (status_int & 0x7F);
	} /* endif */
    } /* endif */

} /* end BurtPipeWait() */

/******************************************************
*
* it is assumed that newfilename is a character string at 
* least MAXPATHLEN (from <sys/param.h> big.
*
******************************************************/

BOOL BurtPreprocessFile(const char *const filename, const char **const defines,
    const char **const includes, char *const newfilename)
{

	char	*system_command;
	int	nbytes;
	long	pid;
	int	exit_status, signal;
	int	status;
	int	i;
	BOOL	rc;
	char	*BurtCppEnv;

    if (!filename)
    {
	fprintf(Log_fp, "ERROR: BurtPreprocessFile() got NULL filename\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!defines)
    {
	fprintf(Log_fp, "ERROR: BurtPreprocessFile() got NULL defines\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!includes)
    {
	fprintf(Log_fp, "ERROR: BurtPreprocessFile() got NULL includes\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!newfilename)
    {
	fprintf(Log_fp, "ERROR: BurtPreprocessFile() got NULL newfile\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!BurtCreateUniqueFilename(PUBLICWRITEDIRECTORY, 
	(char *) NULL, newfilename))
    {
	fprintf(Log_fp, 
	"ERROR: BurtPreprocessFile() unable to BurtCreateUniqueFilename()\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (file_exists(filename))
    {
	if (Verbose)
	{
	    fprintf(Log_fp, "preprocessing ... ");
	    fflush(Log_fp);
	} /* endif */

/* This is stupid but it will always work - jrw */
#ifdef BURT_PREPROCESSOR
	nbytes = strlen(BURT_PREPROCESSOR) + strlen(" > ") + strlen(filename) + strlen(newfilename);
#else
	nbytes = strlen("cc -C -E > ")+ strlen(filename) + strlen(newfilename);
#endif
	nbytes += 2;
	for (i = 0; defines[i]; i ++)
	    nbytes += (strlen(defines[i]) + 1);
	for (i = 0; includes[i]; i ++)
	    nbytes += (strlen(includes[i]) + 1);

	BurtCppEnv = getenv(BURT_CPP_ENV_NAME);
	if (BurtCppEnv != NULL)
		nbytes += strlen(BurtCppEnv);

	if (system_command = (char *) malloc(nbytes))
	{
#ifdef BURT_PREPROCESSOR
		sprintf(system_command, BURT_PREPROCESSOR);
#else
	    sprintf(system_command, "cc -C -E ");
#endif
	    for (i = 0; defines[i]; i ++)
	    {
		strcat(system_command, defines[i]);
		strcat(system_command, " ");
	    } /* endfor */
	    for (i = 0; includes[i]; i ++)
	    {
		strcat(system_command, includes[i]);
		strcat(system_command, " ");
	    } /* endfor */
	    if (BurtCppEnv != NULL)
	    {
	    	strcat(system_command, BurtCppEnv);
		strcat(system_command, " ");
	    }
	    strcat(system_command, filename);
	    strcat(system_command, " > ");
	    strcat(system_command, newfilename);

	    Exec_args[2] = system_command;

	    pid = fork();

	    if (pid < 0)
	    {
		    rc = FALSE;

		    unlink(newfilename);

		fprintf(Log_fp, 
		"ERROR: BurtPreprocessFile() >%s< illegal pid %d from fork()\n",
			filename, pid);
		fflush(Log_fp);
	    } 
	    else if (pid > 0)
	    {
		/* I am parent */
		BurtPipeWait(&exit_status, &signal);

		if (exit_status)
		{
		    rc = FALSE;

		    unlink(newfilename);

		    fprintf(Log_fp, 
	    "ERROR: BurtPreprocessFile() cc abended exit_status %d signal %d\n",
			exit_status, signal);
		    fflush(Log_fp);
		}
		else
		{
		    rc = TRUE;

		    if (Verbose)
		    {
			fprintf(Log_fp, "done\n");
			fflush(Log_fp);
		    } /* endif */
		} /* endif */
	    } 
	    else
	    {
		/* I am child */
		status = execv(Sh_path, Exec_args);
		/* get to this point iff execv() failed */
		fprintf(Log_fp, 
		"ERROR: BurtPreprocessFile() child process failed execv()\n");
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */
	    free(system_command);
	}
	else
	{
	    fprintf(Log_fp, 
"ERROR: BurtPreprocessFile() unable to allocate %d bytes for system_command\n",
		nbytes);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
    }
    else
    {
	rc = FALSE;
	fprintf(Log_fp, "file >%s< does not exist\n", filename);
	fflush(Log_fp);
    } /* endif */

    return rc;

} /* end BurtPreprocessFile() */

/******************************************************
*
*
******************************************************/

void BurtProcessRequestFile(char *const filename, Burtobjlist *const snapshot)
{

Burtpvar *pp;
char tempc;
int startidx;
char *buff;
int nbytes;
int i;
FILE *fp;

SDDS_TABLE SDDS_table;
long l;
char *cp;
int nreq;
int row;
long nrows;
BOOL readmsg_col_present, writemsg_col_present;
BOOL nelem_col_present, mode_col_present;
BOOL type_col_present;

    if (!filename)
    {
	fprintf(Log_fp, "ERROR: BurtProcessRequestFile() got NULL filename\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!snapshot)
    {
	fprintf(Log_fp, "ERROR: BurtProcessRequestFile() got NULL snapshot\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!file_exists(filename))
    {
	fprintf(Log_fp, 
	    "ERROR: BurtProcessRequestFile() file >%s< does not exist\n", 
	    filename);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (BurtSDDSFile(filename))
    {
	if (!SDDS_InitializeInput(&SDDS_table, filename))
	{
	    fprintf(Log_fp, 
		"SYS ERR: BurtProcessRequestFile() SDDS_InitializeInput()\n");
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	/* seeing if required columns are present */
	if (!BurtSDDSColExists(&SDDS_table, NAME_COL, SDDS_STRING))
	{
	    fprintf(Log_fp, 
    "ERROR: BurtProcessRequestFile() string column >%s< not in file >%s<\n", 
		NAME_COL, filename);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	/* seeing if optional columns are present */
	readmsg_col_present = BurtSDDSColExists(&SDDS_table, READMSG_COL, 
				SDDS_STRING);
	writemsg_col_present = BurtSDDSColExists(&SDDS_table, WRITEMSG_COL, 
				SDDS_STRING);
	nelem_col_present = BurtSDDSColExists(&SDDS_table, NELEM_COL, 
				SDDS_LONG);
	mode_col_present = BurtSDDSColExists(&SDDS_table, MODE_COL, 
				SDDS_STRING);
        type_col_present = BurtSDDSColExists(&SDDS_table, TYPE_COL, SDDS_STRING);

	/* reading the data */
	while (SDDS_ReadTable(&SDDS_table) > 0)
	{
	    /* set all rows and columns to initially be "of interest" */
	    SDDS_SetColumnFlags(&SDDS_table, TRUE);
	    SDDS_SetRowFlags(&SDDS_table, TRUE);

	    /* getting nrows from this table */
	    if ((nrows = SDDS_CountRowsOfInterest(&SDDS_table)) == -1)
	    {
		fprintf(Log_fp, 
    "ERROR: BurtProcessRequestFile() could not SDDS_CountRowsOfInterest()\n");
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */

	    /* access tabular data values */
	    for (row = 0; row < nrows; row ++)
	    {
              if (!type_col_present)
                SDDS_CopyString(&cp, "pv");
              else
		sdds_get_value(&SDDS_table, TYPE_COL, (long) row, 
                               (void *) &cp);

		if (!(strcmp(cp, PVSTRING)))
		{
		    free(cp);

		    /* a process variable */
		    pp = new Burtpvar;
		    (pp->get_channel())->set_val(pp->get_val());

		    /* getting pvname */
		    sdds_get_value(&SDDS_table, NAME_COL, (long) row, 
			(void *) &cp);
		    (pp->get_channel())->set_name(cp);
		    free(cp);

		    /* establishing readonly */
		    if (mode_col_present)
		    {
			sdds_get_value(&SDDS_table, MODE_COL, (long) row, 
			    (void *) &cp);

			if (strcmp(cp, READONLYNOTIFYSTRING))
			{
			    /* not read only notify */
			    pp->clear_read_only_notify();

			    if (strcmp(cp, READONLYSTRING))
				pp->clear_read_only();
			    else
				pp->set_read_only();
			}
			else
			{
			    /* read only notify */
			    pp->set_read_only_notify();
			    pp->set_read_only();
			} /* endif */
			    
			free(cp);
		    }
		    else
		    {
			pp->clear_read_only_notify();
			pp->clear_read_only();
		    } /* endif */

		    /* getting nelem */
		    if (nelem_col_present)
		    {
			sdds_get_value(&SDDS_table, NELEM_COL, (long) row, 
			    (void *) &l);
			nreq = ((int) l ? (int) l : UNKNOWN);
		    } 
		    else
			nreq = UNKNOWN;

		    (pp->get_channel())->set_nreq(nreq);
		    
		    snapshot->append((Burtobj *) pp);

		    if (Verbose)
		    {
			fprintf(Log_fp, 
			    "pv >%s< nreq=%d\n", 
			    pp->get_name(), 
			    (pp->get_channel())->get_nreq());
			fflush(Log_fp);
		    } /*endif */
		}
		else if (!(strcmp(cp, DEVSTRING)))
		{
#ifdef NODEVICES
		    fprintf(Log_fp, 
"ERROR: attempted to find information about a device on a version of BURT that cannot process devices\n");
		    fflush(Log_fp);
		    BurtExit(1);
#else
		    free(cp);

		    /* a device */

		    /* getting dev name ... could be composite device */
		    sdds_get_value(&SDDS_table, NAME_COL, (long) row, 
			(void *) &cp);

		    /* process this device recursively */
		    process_device(snapshot, cp, DEFAULTSTRING, &SDDS_table, 
			row, mode_col_present, nelem_col_present, 
			readmsg_col_present, writemsg_col_present);

		    free(cp);
#endif
		}
		else
		{
		    fprintf(Log_fp, 
"ERROR: BurtProcessRequestFile() found unrecognizable type >%s< in file >%s<\n",
			cp, filename);
		    fflush(Log_fp);
		    BurtExit(1);
		} /* endif */
	    } /* endfor */
	} /* endwhile */
    }
    else
    {
	/* non-SDDS request file */

	if ((nbytes = BurtLongestLineLength(filename)) == UNKNOWN)
	{
	    fprintf(Log_fp,
    "ERROR: BurtProcessRequestFile() unable to BurtLongestLineLength() >%s<\n", 
		    filename);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	/* for \0 */
	nbytes ++; 

	if (!(buff = (char *) malloc(nbytes)))
	{
	    fprintf(Log_fp,
		"ERROR: BurtProcessRequestFile() unable to malloc %d bytes\n", 
		nbytes);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	if ((fp = fopen(filename, "r")) == NULL)
	{
	    fprintf(Log_fp,
		"ERROR: BurtProcessRequestFile() unable to open file >%s<\n", 
		    filename);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	while (fgets(buff, nbytes, fp))
	{
	    if (*buff != '\n' && *buff != '#' && *buff != '%')
	    {
		/* got a {RO, RON} pvar {nelements} */
		pp = new Burtpvar;
		(pp->get_channel())->set_val(pp->get_val());

		/* setting readonly */
		if (!(strncmp(buff, READONLYNOTIFYSTRING, 
				strlen(READONLYNOTIFYSTRING))) 
		    && buff[strlen(READONLYNOTIFYSTRING)] == ' ')
		{
		    /* we have a read only notify */
		    pp->set_read_only_notify();
		    pp->set_read_only();
		    startidx = strlen(READONLYNOTIFYSTRING);
		}
		else
		{
		    pp->clear_read_only_notify();

		    if (!(strncmp(buff, READONLYSTRING, strlen(READONLYSTRING)))
			    && buff[strlen(READONLYSTRING)] == ' ')
		    {
			/* we have a read only */
			pp->set_read_only();
			startidx = strlen(READONLYSTRING);
		    }
		    else
		    {
			pp->clear_read_only();
			startidx = 0;
		    } /* endif */

		} /* endif */

		/* loop to pass over starting blanks */
		while (buff[startidx] == ' ')
		    startidx ++;

		for (i = startidx; 
		    i < nbytes && buff[i] != ' ' && buff[i] != '\n'; 
			i ++)
		;

		if (buff[i] != ' ' && buff[i] != '\n')
		{
		    fprintf(Log_fp, 
"ERROR: BurtProcessRequestFile() encountered invalid pvar name in line >%s<", 
			buff);
		    fflush(Log_fp);
		    BurtExit(1);
		} /* endif */

		/* setting pvname */
		tempc = buff[i];
		buff[i] = '\0';
		(pp->get_channel())->set_name(buff+startidx);
		buff[i] = tempc;

		/* setting nreq */
		nreq = atoi(buff+i);
		if (nreq <= 0)
		    nreq = UNKNOWN;

		(pp->get_channel())->set_nreq(nreq);
		
		snapshot->append((Burtobj *) pp);

		if (Verbose)
		{
		    fprintf(Log_fp, 
			"pv >%s< nreq=%d\n", 
			pp->get_name(), 
			(pp->get_channel())->get_nreq());
		    fflush(Log_fp);
		} /*endif */
	    } /* endif */
	} /* endwhile */

	fclose(fp);
	free(buff);
    } /* endif */

} /* end BurtProcessRequestFile() */

/****************************************************************
*
*
****************************************************************/

Burtobjlist *BurtProcessSnapshotFile(char *const filename)
{

Burtobjlist *rc;

    rc = BurtProcessSnapshotFile(filename, (Burtobjlist *const) NULL);

    return rc;

} /* end BurtProcessSnapshotFile() */

/******************************************************
*
* if the snapshot file specified by filename is an SDDS file and there
* are any devices specified, it is assumed that they are atomic devices.
*
* the readonlynotify is an optional pointer.  if it is non-NULL, 
* it is assumed to be a pointer to a list that is only a view.
* RON entries that are found in this snapshot file are appended to
* readonlynotify.
*
******************************************************/

Burtobjlist *BurtProcessSnapshotFile(char *const filename, 
    Burtobjlist *const readonlynotify)
{

Burtpvar *pp;
char tempc;
char *buff;
int nbytes;
int i, startidx, nelem;
FILE *fp;
Burtobjlist *rc;

SDDS_TABLE SDDS_table;
long l;
char *cp;
int row;
long nrows;
BOOL readmsg_col_present, writemsg_col_present, mode_col_present ;
BOOL type_col_present;
BOOL gotfixedparms;
char *t1p, *t2p, *t3p;
char *valbuff;
char **associate_filenames;
char *pathname;
int32_t nnames;

#ifndef NODEVICES
devstructarr myinfo;
Burtatomdev *dp;
int j;
#endif

    if (!filename)
    {
	fprintf(Log_fp, "ERROR: BurtProcessSnapshotFile() got NULL filename\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!file_exists(filename))
    {
	fprintf(Log_fp, 
	    "ERROR: BurtProcessSnapshotFile() file >%s< does not exist\n", 
	    filename);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (Verbose)
    {
	fprintf(Log_fp, "file >%s<\n", filename);
	fflush(Log_fp);
    } /* endif */

    rc = new Burtobjlist;
    rc->set_owner();

    if (BurtSDDSFile(filename))
    {
	if (!SDDS_InitializeInput(&SDDS_table, filename))
	{
	    fprintf(Log_fp, 
		"SYS ERR: BurtProcessSnapshotFile() SDDS_InitializeInput()\n");
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

        /* seeing if required columns are present */
        if (!BurtSDDSColExists(&SDDS_table, NAME_COL, SDDS_STRING))
        {
            fprintf(Log_fp, 
    "ERROR: BurtProcessSnapshotFile() string column >%s< not in file >%s<\n", 
                NAME_COL, filename);
            fflush(Log_fp);
            BurtExit(1);
        } /* endif */

        if (!BurtSDDSColExists(&SDDS_table, LINEAGE_COL, SDDS_STRING))
        {
            fprintf(Log_fp, 
    "ERROR: BurtProcessSnapshotFile() string column >%s< not in file >%s<\n", 
                LINEAGE_COL, filename);
            fflush(Log_fp);
            BurtExit(1);
        } /* endif */
        if (!BurtSDDSColExists(&SDDS_table, NELEM_COL, SDDS_LONG))
        {
            fprintf(Log_fp, 
	"ERROR: BurtProcessSnapshotFile() long column >%s< not in file >%s<\n", 
                NELEM_COL, filename);
            fflush(Log_fp);
            BurtExit(1);
        } /* endif */
        if (!BurtSDDSColExists(&SDDS_table, VAL_COL, SDDS_STRING))
        {
            fprintf(Log_fp, 
    "ERROR: BurtProcessSnapshotFile() string column >%s< not in file >%s<\n", 
                VAL_COL, filename);
            fflush(Log_fp);
            BurtExit(1);
        } /* endif */

        /* seeing if optional columns are present */
        readmsg_col_present = BurtSDDSColExists(&SDDS_table, READMSG_COL, 
				SDDS_STRING);
        writemsg_col_present = BurtSDDSColExists(&SDDS_table, WRITEMSG_COL, 
				   SDDS_STRING);
        mode_col_present = BurtSDDSColExists(&SDDS_table, MODE_COL, 
				SDDS_STRING);
        type_col_present = BurtSDDSColExists(&SDDS_table, TYPE_COL, SDDS_STRING);

	/***********************/
	/*                     */
        /* loop to process the */
	/* tables in this file */
	/*                     */
	/***********************/

	gotfixedparms = FALSE;

        while (SDDS_ReadTable(&SDDS_table) > 0)
        {
            /* set all rows and columns in this    */
	    /* table to initially be "of interest" */
            SDDS_SetColumnFlags(&SDDS_table, TRUE);
            SDDS_SetRowFlags(&SDDS_table, TRUE);

	    if (!gotfixedparms)
	    {
		gotfixedparms = TRUE;

		/****************************/
		/*                          */
		/* reading fixed parameters */
		/*                          */
		/****************************/

		/* Time */
		if (sdds_get_parm(&SDDS_table, TIMEHEADERSTRING, &cp))
		{
		    (rc->get_parms())->set_currtime(cp);
		    free(cp);
		} /* endif */

		/* Login Id */
		if (sdds_get_parm(&SDDS_table, LOGINHEADERSTRING, &cp))
		{
		    if (!(t1p = strchr(cp, ' ')))
		    {
			fprintf(Log_fp,
		"ERROR: BurtProcessSnapshotFile() found illegal login id\n");
			fflush(Log_fp);

			BurtExit(1);
		    } /* endif */
		    *t1p = '\0';
		    (rc->get_parms())->set_loginid(cp);
		    *t1p = ' ';

		    if (!(t2p = strchr(t1p, '(')))
		    {
			fprintf(Log_fp,
		"ERROR: BurtProcessSnapshotFile() found illegal login id\n");
			fflush(Log_fp);
			BurtExit(1);
		    } /* endif */

		    if (!(t3p = strchr(t2p, ')')))
		    {
			fprintf(Log_fp,
		"ERROR: BurtProcessSnapshotFile() found illegal login id\n");
			fflush(Log_fp);
			BurtExit(1);
		    } /* endif */

		    *t3p = '\0';
		    (rc->get_parms())->set_real_world_name(t2p+1);
		    *t3p = ')';

		    free(cp);
		}
		else
		{
		    /* parm not in SDDS file */
		    (rc->get_parms())->set_loginid(UNKNOWNLOGINID);
		    (rc->get_parms())->
			set_real_world_name(UNKNOWNREALWORLDNAME);
		} /* endif */

		/* Eff Id */
		if (sdds_get_parm(&SDDS_table, EFFUIDHEADERSTRING, &cp))
		{
		    (rc->get_parms())->set_effective_uid((uid_t) atoi(cp));
		    free(cp);
		}
		else
		{
		    /* parm not in SDDS file */
		    (rc->get_parms())->set_effective_uid((uid_t) UNKNOWN);
		} /* endif */

		/* Group Id */
		if (sdds_get_parm(&SDDS_table, GROUPIDHEADERSTRING, &cp))
		{
		    (rc->get_parms())->set_group_id((gid_t) atoi(cp));
		    free(cp);
		}
		else
		{
		    /* parm not in SDDS file */
		    (rc->get_parms())->set_group_id((gid_t) UNKNOWN);
		} /* endif */

		/* Keywords */
		if (sdds_get_parm(&SDDS_table, KEYWORDSHEADERSTRING, &cp))
		{
		    (rc->get_parms())->set_keywords(cp);
		    free(cp);
		} /* endif */

		/* Comments */
		if (sdds_get_parm(&SDDS_table, COMMENTSHEADERSTRING, &cp))
		{
		    (rc->get_parms())->set_comments(cp);
		    free(cp);
		} /* endif */

		/* Type */
		if (sdds_get_parm(&SDDS_table, TYPEHEADERSTRING, &cp))
		{
		    if (!strcmp(cp, ABSOLUTESTRING))
			(rc->get_parms())->set_absolute();
		    else if (!strcmp(cp, RELATIVESTRING))
			(rc->get_parms())->set_relative();
		    else if (!strcmp(cp, NOWRITESTRING))
			(rc->get_parms())->set_nowrite();
		    else
		    {
			fprintf(Log_fp, 
"SYS ERR: BurtProcessSnapshotFile() file >%s< found unrecognizable snapshot type >%s<\n", 
			    filename, cp);
			fflush(Log_fp);
			BurtExit(1);
		    } /* endif */
		    free(cp);
		}
		else
		{
		    /* parm not in SDDS file */

		    /* assuming absolute */
		    (rc->get_parms())->set_absolute();

		    fprintf(Log_fp, 
"WARNING: snapshot file >%s< parm %s (Type of snapshot file) not specified ... assuming %s = %s\n",
			filename, TYPEHEADERSTRING, TYPEHEADERSTRING,
			ABSOLUTESTRING);
		    fflush(Log_fp);
		} /* endif */

		/* Getting any associate names */
		if (associate_filenames 
		    = SDDS_GetAssociateNames(&SDDS_table, &nnames))
		{
		    for (i = 0; i < nnames; i ++)
		    {
			for (cp = associate_filenames[i]; 
			    cp && *cp && isspace(*cp); 
				cp ++)
			;

			if (cp && *cp == '/')
			    /* absolute pathname */
			    (rc->get_request_filenames())->
				append(new Filename(cp));
			else
			{

			    /* relative pathname */
			    if (SDDS_GetAssociateInformation(&SDDS_table,"path",
			    &pathname, SDDS_GET_BY_NAME, associate_filenames[i])
				    == SDDS_STRING)
			    {
				(rc->get_request_filenames())->
				    append(new Filename(pathname, 
					associate_filenames[i]));
				free(pathname);
			    }
			    else
			    {
				fprintf(Log_fp, 
"ERROR: BurtProcessSnapshotFile() could not SDDS_GetAssociateInformation() for associate file %d >%s<\n", 
				    i+1, associate_filenames[i]);
				fflush(Log_fp);
				BurtExit(1);
			    } /* endif */
			} /* endif */
			free(associate_filenames[i]);
		    } /* endfor */
		}
		else
		{
		    fprintf(Log_fp, 
    "ERROR: BurtProcessSnapshotFile() could not SDDS_GetAssociateNames()\n");
		    fflush(Log_fp);
		    BurtExit(1);
		} /* endif */
	    } /* endif */

            /* getting nrows from this table */
            if ((nrows = SDDS_CountRowsOfInterest(&SDDS_table)) == -1)
            {
                fprintf(Log_fp, 
    "ERROR: BurtProcessSnapshotFile() could not SDDS_CountRowsOfInterest()\n");
                fflush(Log_fp);
                BurtExit(1);
            } /* endif */

            /* loop to process each row of this table */
            for (row = 0; row < nrows; row ++)
            {
              if (!type_col_present) 
                SDDS_CopyString(&cp, "pv");
              else
		sdds_get_value(&SDDS_table, TYPE_COL, (long) row, 
                               (void *) &cp);

                if (!(strcmp(cp, PVSTRING)))
                {
                    free(cp);

#ifdef JRW_ALLOW_BLANK_VALUES
		    sdds_get_value(&SDDS_table, VAL_COL, (long) row, 
			(void *) &cp);
		    if ((cp != NULL) && (cp[0] != '\0'))
		    {
#endif
		    /* a process variable */
		    pp = new Burtpvar;
		    (pp->get_channel())->set_val(pp->get_val());

                    /* getting pvar name */
		    sdds_get_value(&SDDS_table, NAME_COL, (long) row, 
			(void *) &cp);
		    (pp->get_channel())->set_name(cp);
		    free(cp);

		    /* establishing readonly */
		    if (mode_col_present)
		    {
			sdds_get_value(&SDDS_table, MODE_COL, (long) row, 
			    (void *) &cp);
			
			if (strcmp(cp, READONLYNOTIFYSTRING))
			{
			    pp->clear_read_only_notify();

			    if (strcmp(cp, READONLYSTRING))
				pp->clear_read_only();
			    else
				pp->set_read_only();
			}
			else
			{
			    pp->set_read_only_notify();
			    pp->set_read_only();
			} /* endif */

			free(cp);
		    }
		    else
		    {
			pp->clear_read_only_notify();
			pp->clear_read_only();
		    } /* endif */

		    /* setting val */
		    sdds_get_value(&SDDS_table, NELEM_COL, (long) row, 
			(void *) &l);
		    sdds_get_value(&SDDS_table, VAL_COL, (long) row, 
			(void *) &cp);
		    if (valbuff = prepare_val(cp, (int) l))
		    {
			(pp->get_val())->set(DBR_STRING, (int) l, valbuff);
			free(valbuff);
		    }
		    else
		    {
			fprintf(Log_fp, 
"ERROR: BurtProcessSnapshotFile() could not prepare_val() for nelem %d >%s<\n",
			    (int) l, cp);
			fflush(Log_fp);
			BurtExit(1);
		    } /* endif */
		    free(cp);

		    rc->append((Burtobj *) pp);

		    /* if readonlynotify list supplied */
		    if (readonlynotify && pp->read_only_notify())
			readonlynotify->append((Burtobj *) pp);

		    if (Verbose)
		    {
			if (pp->read_only_notify())
			    fprintf(Log_fp, 
				"pv >%s< nelem %d <read only notify>\n", 
				pp->get_name(), 
				(pp->get_val())->get_nelem());
			else if (pp->read_only())
			    fprintf(Log_fp, "pv >%s< nelem %d <read only>\n", 
				pp->get_name(), 
				(pp->get_val())->get_nelem());
			else
			    fprintf(Log_fp, "pv >%s< nelem %d\n", 
				pp->get_name(), 
				(pp->get_val())->get_nelem());

			fflush(Log_fp);

		    } /* endif */
#ifdef JRW_ALLOW_BLANK_VALUES
		    }
#endif
		}
                else if (!(strcmp(cp, DEVSTRING)))
                {
#ifdef NODEVICES
		    fprintf(Log_fp, 
"ERROR: attempted to find information about a device on a version of BURT that cannot process devices\n");
		    fflush(Log_fp);
		    BurtExit(1);
#else
                    free(cp);

		    /* a device ... known to be an atomic device */
		    dp = new Burtatomdev;
		    (dp->get_readchannel())->set_val(dp->get_val());
		    (dp->get_writechannel())->set_val(dp->get_val());

                    /* getting pvar name */
		    sdds_get_value(&SDDS_table, NAME_COL, row, &cp);
                    dp->set_name(cp);
                    free(cp);

                    /* getting lineage */
		    sdds_get_value(&SDDS_table, LINEAGE_COL, row, &cp);
                    dp->set_lineage(cp);
                    free(cp);

		    /* finding out about this device, for read/write pvars */
		    dev_info(dp->get_name(), &myinfo);

		    /* establishing readonly */
		    if (mode_col_present)
		    {
			sdds_get_value(&SDDS_table, MODE_COL, row, &cp);

			if (strcmp(cp, READONLYNOTIFYSTRING))
			{
			    dp->clear_read_only_notify();

			    if (strcmp(cp, READONLYSTRING))
				dp->clear_read_only();
			    else
				dp->set_read_only();
			}
			else
			{
			    dp->set_read_only_notify();
			    dp->set_read_only();
			} /* endif */
			    
			free(cp);
		    }
		    else
		    {
			dp->clear_read_only_notify();
			dp->clear_read_only();
		    } /* endif */

		    /*************************************/
		    /*                                   */
		    /* establishing read msg and pvar    */
		    /* done in case writedirective = add */
		    /*                                   */
		    /*************************************/

		    /* read msg */
		    if (readmsg_col_present)
			sdds_get_value(&SDDS_table, READMSG_COL, row, &cp);
		    else
			cp = DEFAULTREADMSG;

		    dp->set_readmsg(cp);

		    /* searching for restore msg in this device */
		    for (j = 0;
			myinfo[j] && strcmp(myinfo[j]->msg, cp);
			    j ++)
		    ;

		    if (myinfo[j])
			/* got it */
			(dp->get_readchannel())->set_name(myinfo[j]->pvname);
		    else
		    {
			/* not there */
			fprintf(Log_fp, 
"ERROR: BurtProcessSnapshotFile() file >%s< devname >%s< did not recognize msg >%s<\n", 
			    filename, dp->get_name(), cp);
			fflush(Log_fp);
			BurtExit(1);
		    } /* endif */

		    if (readmsg_col_present)
			free(cp);

		    /***********************************/
		    /*                                 */
		    /* establishing write msg and pvar */
		    /*                                 */
		    /***********************************/

		    /* write msg */
		    if (writemsg_col_present)
			sdds_get_value(&SDDS_table, WRITEMSG_COL, row, &cp);
		    else
			cp = DEFAULTWRITEMSG;

		    dp->set_writemsg(cp);

		    /* searching for restore msg in this device */
		    for (j = 0;
			myinfo[j] && strcmp(myinfo[j]->msg, cp);
			    j ++)
		    ;

		    if (myinfo[j])
			/* got it */
			(dp->get_writechannel())->set_name(myinfo[j]->pvname);
		    else
		    {
			/* not there */
			fprintf(Log_fp, 
"ERROR: BurtProcessSnapshotFile() file >%s< devname >%s< did not recognize msg >%s<\n", 
			    filename, dp->get_name(), cp);
			fflush(Log_fp);
			BurtExit(1);
		    } /* endif */

		    if (writemsg_col_present)
			free(cp);

		    /* setting val */
		    sdds_get_value(&SDDS_table, NELEM_COL, row, &l);
		    sdds_get_value(&SDDS_table, VAL_COL, row, &cp);
		    if (valbuff = prepare_val(cp, (int) l))
		    {
			(dp->get_val())->set(DBR_STRING, (int) l, valbuff);
			free(valbuff);
		    }
		    else
		    {
			fprintf(Log_fp, 
"ERROR: BurtProcessSnapshotFile() could not prepare_val() for nelem %d >%s<\n",
			    (int) l, cp);
			fflush(Log_fp);
			BurtExit(1);
		    } /* endif */
		    free(cp);

		    rc->append((Burtobj *) dp);

		    /* if readonlynotify list supplied */
		    if (readonlynotify && dp->read_only_notify())
			readonlynotify->append((Burtobj *) dp);

		    if (Verbose)
		    {
			if (dp->read_only_notify())
			    fprintf(Log_fp, 
			"dev >%s< writepv >%s< nelem %d <read only notify>\n", 
				dp->get_name(), 
				(dp->get_writechannel())->get_name(),
				(dp->get_val())->get_nelem());
			else if (dp->read_only())
			    fprintf(Log_fp, 
				"dev >%s< writepv >%s< nelem %d <read only>\n", 
				dp->get_name(), 
				(dp->get_writechannel())->get_name(),
				(dp->get_val())->get_nelem());
			else
			    fprintf(Log_fp, "dev >%s< writepv >%s< nelem %d\n",
				 dp->get_name(), 
				(dp->get_writechannel())->get_name(),
				(dp->get_val())->get_nelem());

			fflush(Log_fp);
		    } /* endif */

#endif
                }
                else
                {
                    fprintf(Log_fp, 
"ERROR: BurtProcessSnapshotFile() found unrecognizable type >%s< in file >%s<\n",
                        cp, filename);
                    fflush(Log_fp);
                    BurtExit(1);
                } /* endif */
            } /* endfor */
        } /* endwhile */
    }
    else
    {
	/* non-SDDS request file */

	if ((nbytes = BurtLongestLineLength(filename)) == UNKNOWN)
	{
	    fprintf(Log_fp,
"ERROR: BurtProcessSnapshotFile() unable to BurtLongestLineLength() file >%s<\n", 
		    filename);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	/* for \0 */
	nbytes ++; 

	if (!(buff = (char *) malloc(nbytes)))
	{
	    fprintf(Log_fp,
	    "ERROR: BurtProcessSnapshotFile() unable to malloc() %d bytes\n", 
		nbytes);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	if ((fp = fopen(filename, "r")) == NULL)
	{
	    fprintf(Log_fp,
		"ERROR: BurtProcessSnapshotFile() unable to open file >%s<\n", 
		    filename);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	if (!get_through_header(fp, rc, buff, nbytes))
	{
	    fprintf(Log_fp, 
	    "ERROR: BurtProcessSnapshotFile() file >%s< with illegal header\n", 
		filename);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
     
	/* through the header info */
	while (fgets(buff, nbytes, fp))
	{
	    if (*buff != '\n' && *buff != '#' && *buff != '%')
	    {

		/* got a pvar nelements ...data... */
		pp = new Burtpvar;
		(pp->get_channel())->set_val(pp->get_val());

		buff[strlen(buff)-1] = '\0';

		if (!(strncmp(buff, READONLYNOTIFYSTRING, 
			strlen(READONLYNOTIFYSTRING))) 
		    && buff[strlen(READONLYNOTIFYSTRING)] == ' ')
		{
		    /* we have a read only notify */
		    pp->set_read_only_notify();
		    pp->set_read_only();
		    startidx = strlen(READONLYNOTIFYSTRING);
		}
		else
		{
		    pp->clear_read_only_notify();

		    if (!(strncmp(buff, READONLYSTRING, strlen(READONLYSTRING)))
			&& buff[strlen(READONLYSTRING)] == ' ')
		    {
			/* we have a read only */
			pp->set_read_only();
			startidx = strlen(READONLYSTRING);
		    }
		    else
		    {
			pp->clear_read_only();
			startidx = 0;
		    } /* endif */

		} /* endif */

		/* sp->pvarname points to beg of record which is pvarname */
		/* then placing null character at first blank to make     */
		/* sp->pvarname a character string.                       */

		/**************************/
		/*                        */
		/* searching for pvarname */
		/*                        */
		/**************************/

		/* read over starting blanks */
		while (buff[startidx] == ' ')
		    startidx ++;

		/* loop to pass over pvarname */
		for (i = startidx;
		    i < nbytes && buff[i] != ' ' && buff[i] != '\n';
			i ++)
		;

		if (buff[i] != ' ')
		{
		    fprintf(Log_fp,
"ERROR: BurtProcessSnapshotFile() encountered invalid pvar name in line >%s<",
			buff);
		    fflush(Log_fp);
		    BurtExit(1);
		} /* endif */

		tempc = buff[i];
		buff[i] = '\0';
		(pp->get_channel())->set_name(buff+startidx);
		buff[i] = tempc;

		/***********************/
		/*                     */
		/* searching for nelem */
		/*                     */
		/***********************/

		/* loop to pass over interim blanks */
		startidx = i;
		while (buff[startidx] == ' ')
		    startidx ++;

		/* loop to pass over nelem */
		for (i = startidx;
		    i < nbytes && buff[i] != ' ' && buff[i] != '\n';
			i ++)
		;

		if (buff[i] != ' ')
		{
		    fprintf(Log_fp,
"ERROR: BurtProcessSnapshotFile() encountered invalid nelem in line >%s<",
			buff);
		    fflush(Log_fp);
		    BurtExit(1);
		} /* endif */

		tempc = buff[i];
		buff[i] = '\0';
		/* atoi() return 0 on failure ... not valid */
		if (!(nelem = atoi(buff+startidx)))
		{
		    fprintf(Log_fp, 
	    "ERROR: BurtProcessSnapshotFile() found nelem 0 in line >%s<\n",
			buff);
		    fflush(Log_fp);
		    BurtExit(1);
		} /* endif */
		buff[i] = tempc;

		/* loop to pass over interim blanks */
		while (buff[i] == ' ')
		    i ++;

		if (buff[i] == '\0' || buff[i] == '\n')
		{
		    fprintf(Log_fp, 
	    "ERROR: BurtProcessSnapshotFile() found no data in line >%s<\n",
			buff);
		    fflush(Log_fp);
		    BurtExit(1);
		} /* endif */

		if (valbuff = prepare_val(buff+i, nelem))
		{
		    (pp->get_val())->set(DBR_STRING, nelem, valbuff);
		    free(valbuff);
		}
		else
		{
		    fprintf(Log_fp, 
"ERROR: BurtProcessSnapshotFile() could not prepare_val() for nelem %d >%s<\n",
			(int) l, cp);
		    fflush(Log_fp);
		    BurtExit(1);
		} /* endif */

		rc->append((Burtobj *) pp);

		/* if readonlynotify list supplied */
		if (readonlynotify && pp->read_only_notify())
		    readonlynotify->append((Burtobj *) pp);

		if (Verbose)
		{
		    if (pp->read_only_notify())
			fprintf(Log_fp, "pv >%s< nelem %d <read only notify>\n",
			    pp->get_name(), 
			    ((pp->get_channel())->get_val())->get_nelem());
		    else if (pp->read_only())
			fprintf(Log_fp, "pv >%s< nelem %d <read only>\n", 
			    pp->get_name(), 
			    ((pp->get_channel())->get_val())->get_nelem());
		    else
			fprintf(Log_fp, "pv >%s< nelem %d\n", 
			    pp->get_name(), (pp->get_val())->get_nelem());

		    fflush(Log_fp);
		} /* endif */
	    } /* endif */
	} /* endwhile */

	fclose(fp);
	free(buff);

    } /* endif */

    return rc;

} /* end BurtProcessSnapshotFile() */

/****************************************************************
*
*
****************************************************************/

BOOL BurtSDDSColExists(SDDS_TABLE *const table, char *const colname,
    const long type)
{

BOOL rc;
long index;

    if (table && colname && *colname)
	if ((index = SDDS_GetColumnIndex(table, colname)) >= 0)
	    rc = (type == SDDS_GetColumnType(table, index));
	else
	    rc = FALSE;
    else
	rc = FALSE;

    return rc;

} /* end BurtSDDSColExists() */

/****************************************************************
*
*
****************************************************************/

BOOL BurtSDDSParmExists(SDDS_TABLE *const table, char *const parmname)
{

BOOL rc;

    if (table && parmname && *parmname)
	if (SDDS_GetParameterDefinition(table, parmname))
	    rc = TRUE;
	else
	    rc = FALSE;
    else
	rc = FALSE;

    return rc;

} /* end BurtSDDSParmExists() */

/****************************************************************
*
*
****************************************************************/

BOOL BurtSDDSFile(const char *const fname)
{

FILE *fp;
BOOL rc;

    if (fname)
    {
        if (*fname)
        {
            if (fp = fopen(fname, "r"))
            {
                if (rc = nextchar(fp, 'S'))
                    if (rc = nextchar(fp, 'D'))
                        if (rc = nextchar(fp, 'D'))
                            rc = nextchar(fp, 'S');

                fclose(fp);
            }
            else
                rc = FALSE;
        }
        else
            rc = FALSE;
    }
    else
        rc = FALSE;

    return rc;

} /* end BurtSDDSFile() */

/**************************************/
/*                                    */
/* Channel Access Interface Functions */
/*                                    */
/**************************************/

extern "C" {

/****************************************************************
*
*
****************************************************************/

void BurtInitializeChannelAccess()
{
    ca_task_initialize();
} /* end initialize_channel_access() */

/****************************************************************
*
*
****************************************************************/

BOOL BurtQueueCaArrayGetCallback(Channel *cp)
{

int status;
BOOL rc;

    status = ca_array_get_callback((long) cp->get_dbr_req_type(),
	(unsigned long) (cp->get_nreq() == UNKNOWN
	    ? cp->channel_nelem()
	    : cp->get_nreq()),
	*(cp->get_cid()), my_get_callback, (void *) cp);

    if (status == ECA_NORMAL)
    {
	rc = TRUE;
	cp->set_read_pending(); 

	if (Verbose)
	    fprintf(Log_fp, "%s ... ca_array_get_callback() nreq %d ... OK\n",
		cp->get_name(),
		(unsigned long) (cp->get_nreq() == UNKNOWN
		    ? cp->channel_nelem()
		    : cp->get_nreq()));
    }
    else
    {
	rc = FALSE;

	fprintf(Log_fp, "%s ... ca_array_get_callback() nreq %d ... %s\n",
	    cp->get_name(),
	    (unsigned long) (cp->get_nreq() == UNKNOWN
		? cp->channel_nelem()
		: cp->get_nreq()),
	    ca_message(status));
    } /* endif */

    fflush(Log_fp);

    return rc;

} /* end BurtQueueCaArrayGetCallback() */

/****************************************************************
*
*
****************************************************************/

BOOL BurtCaArrayPut(Channel *cp)
{

int status;
char *source, *ret_cp;
double *dp;
int i;
BOOL rc;

    rc = TRUE;

    if (!cp)
    {
	fprintf(Log_fp, "ERROR: BurtCaArrayPut() got NULL cp\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!((cp->get_val())->defined()))
    {
	fprintf(Log_fp, 
	    "ERROR: BurtCaArrayPut() encountered undefined value\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */


    /* older versions (pre-3.12) of IOC core could convert */
    /* arrays of strings to arrays of numeric types ... so */
    /* we have to attempt to do the conversion here to     */
    /* doubles and send down the array of doubles.         */

    if ((cp->get_val())->get_nelem() > 1 
	&& (cp->get_val())->get_dbr_type() == DBR_STRING
	&& cp->channel_type() != DBR_STRING
	&& cp->channel_type() != DBR_ENUM)
    {
	if (dp = (double *) malloc((cp->get_val())->get_nelem()*sizeof(double)))
	{
	    /* the conversion assumes that Values that are stored as  */
	    /* DBR_STRING have already been prepare_val()'d when they */
	    /* were read in via BurtProcessSnapshotFile().  In doing  */
	    /* so, individual values are placed into dbr_size_n(DBR_STRING,1) */
	    /* size slots (about 40 bytes '\0' terminated).           */

	    for (i = 0, source = (char *) (cp->get_val())->get_pval(); 
		    rc && i < (cp->get_val())->get_nelem(); 
			i ++, source += dbr_size_n(DBR_STRING,1))
	    {
		dp[i] = strtod(source, &ret_cp);

		if (ret_cp == source || *ret_cp)
		{
		    rc = FALSE;
		    fprintf(Log_fp, 
			"ERROR: BurtCaArrayPut() unable to strtod() >%s<\n", 
			source);
		    fflush(Log_fp);
		} /* endif */
	    } /* endfor */

	    if (rc)
		status = ca_array_put(DBR_DOUBLE, 
			    (unsigned long)(cp->get_val())->get_nelem(),
			    *(cp->get_cid()), (void *) dp);

	    free((char *) dp);
	}
	else
	{
	    rc = FALSE;
	    fprintf(Log_fp, 
		"ERROR: BurtCaArrayPut() unable to malloc %d bytes\n",
		(cp->get_val())->get_nelem()*sizeof(double));
	    fflush(Log_fp);
	} /* endif */
    }
    else
	status = ca_array_put((cp->get_val())->get_dbr_type(), 
		    (unsigned long)(cp->get_val())->get_nelem(),
		    *(cp->get_cid()), (cp->get_val())->get_pval());

    if (rc)
    {
	if (status == ECA_NORMAL)
	{
	    if (Verbose)
		fprintf(Log_fp, "%s ... ca_array_put() nelem %ld ... OK\n",
		    cp->get_name(), 
		    (unsigned long) (cp->get_val())->get_nelem());
	}
	else
	{
	    rc = FALSE;

	    fprintf(Log_fp, "%s ... ca_array_put() nelem %ld ... %s\n",
		cp->get_name(), (unsigned long) (cp->get_val())->get_nelem(), 
		ca_message(status));
	} /* endif */

	fflush(Log_fp);
    } /* endif */

    return rc;

} /* end BurtCaArrayPut() */

/****************************************************************
*
*
****************************************************************/

void BurtCaFlushIo()
{

int status;

    status = ca_flush_io();

    if (status == ECA_NORMAL)
    {
	if (Verbose)
	    fprintf(Log_fp, "flushing with ca_flush_io() ... OK\n");
    }
    else
    {
	fprintf(Log_fp, "flushing with ca_flush_io() ... %s\n", 
	    ca_message(status));
	BurtExit(1);
    } /* end switch() */

    fflush(Log_fp);

} /* end BurtCaFlushIo() */

/****************************************************************
*
*
****************************************************************/

void BurtQueueCaPendEvent(float sec)
{

    if (sec > 0)
	ca_pend_event(sec);
    else
	ca_pend_event(SHORT_TIME);

} /* end BurtQueueCaPendEvent() */

/****************************************************************
*
*
****************************************************************/

BOOL BurtQueueCaPendIO()
{

int status;
BOOL rc;

    status = ca_pend_io(5.0);

    if (status == ECA_NORMAL)
    {
	rc = TRUE;
	if (Verbose)
	    fprintf(Log_fp, "flushing with ca_pend_io() ... OK\n"); 
    }
    else
    {
	rc = FALSE;
	fprintf(Log_fp, "flushing with ca_pend_io() ... %s\n", 
	    ca_message(status)); 

	if (status != ECA_TIMEOUT)
	    BurtExit(1);
    } /* endif */

    fflush(Log_fp);

    return rc;

} /* end BurtQueueCaPendIO() */

/****************************************************************
*
*
****************************************************************/

void BurtQueueCaSearchAndConnect(Channel *cp)
{

int status;

    if (cp)
    {
	if (!(cp->ever_searched()))
	{
	    status = ca_search_and_connect((char *) cp->get_name(), 
			cp->get_cid(), my_connection_callback, (void *) NULL);

	    if (status == ECA_NORMAL)
	    {
		cp->set_ever_searched();
		NoutstandingSearches ++;

		if (Verbose)
		    fprintf(Log_fp, "%s ... ca_search_and_connect() ... OK\n", 
			cp->get_name());
	    }
	    else
		fprintf(Log_fp, "%s ... ca_search_and_connect() ... %s\n", 
		    cp->get_name(), ca_message(status));

	    fflush(Log_fp);
	} /* endif */
    }
    else
    {
	fprintf(Log_fp, "ERROR: BurtQueueCaSearchAndConnect() got NULL cp\n");
	BurtExit(1);
    } /* endif */

} /* end BurtQueueCaSearchAndConnect() */

} /* end extern c thing */

/***************************/
/*                         */
/* Private Local Functions */
/*                         */
/***************************/

#ifndef NODEVICES
/****************************************************************
*
*
****************************************************************/

static int dev_count(const char *const cp, devnamearr *const mynames)
{

int rc;
int status;
    
    if (cp && mynames)
    {
	if (*cp)
	{
	    status = devCount((char *) cp, &rc, mynames);

	    switch (status)
	    {
		case DEV_NORMAL: /* all OK */ break; 
		case DEV_NOSERVER:
		    fprintf(Log_fp, 
"ERROR: devCount() unable to contact device server to resolve dname >%s<\n", 
			cp);
		    fflush(Log_fp);
		    BurtExit(1);
		    break;
		case DEV_CALLFAIL:
		    fprintf(Log_fp, 
			"ERROR: devCount() dname >%s<: RPC call failed\n", cp);
		    fflush(Log_fp);
		    BurtExit(1);
		    break;
		case DEV_NODEV:
		    fprintf(Log_fp, 
	"ERROR: devCount() dname >%s<: dname is not a defined device name\n", 
			cp);
		    fflush(Log_fp);
		    BurtExit(1);
		    break;
		default:
		    fprintf(Log_fp, 
		    "ERROR: devCount() dname >%s<: got unrecognizable rc %d\n", 
			cp, status);
		    fflush(Log_fp);
		    BurtExit(1);
		    break;
	    } /* end switch() */
	}
	else
	{
	    fprintf(Log_fp, 
		"ERROR: dev_count() got cp >%s<\n", cp);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
    }
    else
    {
	fprintf(Log_fp, 
	    "ERROR: dev_count() got cp %x mynames %x\n", cp, mynames);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    return rc;

} /* end dev_count() */

/****************************************************************
*
*
****************************************************************/

static void dev_info(const char *const dname, devstructarr *const myinfo)
{

int status;

    if (dname && myinfo)
    {
	if (*dname)
	{
	    status = devInfo((char *) dname, myinfo);
	    switch (status)
	    {
		case DEV_NORMAL: /* all OK */ break; 
		case DEV_NOSERVER:
		    fprintf(Log_fp, 
"ERROR: devInfo() dname >%s<: unable to contact dev server to resolve dname\n", 
			dname);
		    fflush(Log_fp);
		    BurtExit(1);
		    break;
		case DEV_CALLFAIL:
		    fprintf(Log_fp, 
			"ERROR: devInfo() dname >%s<: RPC call failed\n", 
			dname);
		    fflush(Log_fp);
		    BurtExit(1);
		    break;
		case DEV_NODEV:
		    fprintf(Log_fp, 
	    "ERROR: devInfo() dname >%s<: dname is not a defined device name\n",
			dname);
		    fflush(Log_fp);
		    BurtExit(1);
		    break;
		default:
		    fprintf(Log_fp, 
		    "ERROR: devInfo() dname >%s<: got unrecognizable rc %d\n", 
			dname, status);
		    fflush(Log_fp);
		    BurtExit(1);
		    break;
	    } /* end switch() */

	}
	else
	{
	    fprintf(Log_fp, 
		"ERROR: dev_info() got dname >%s<\n", dname);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
    }
    else
    {
	fprintf(Log_fp, 
	    "ERROR: dev_info() got dname %x myinfo %x\n", dname, myinfo);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

} /* end dev_info() */
#endif

/****************************************************************
*
*
****************************************************************/

static BOOL file_exists(const char *const fname)
{

struct stat buff;
BOOL rc;

    if (fname)
        if (*fname)
            rc = !stat(fname, &buff);
        else
            rc = FALSE;
    else
        rc = FALSE;

    return rc;

} /* end file_exists() */

/******************************************************
*
* this should NOT be replaced by lex/yacc as it will be
* replaced by SDDS calls.
*
******************************************************/

static BOOL get_through_header(FILE *const fp, Burtobjlist *const snapshot, 
    char *const buff, const int nbytes)
{

char *t1p, *t2p, *t3p;
 
    if (!fp)
    {
        fprintf(Log_fp, "ERROR: get_through_header() got NULL fp\n");
        fflush(Log_fp);

	return FALSE;
    } /* endif */

    if (!snapshot)
    {
	fprintf(Log_fp, "ERROR: get_through_header() got NULL snapshot\n");
	fflush(Log_fp);

	return FALSE;
    } /* endif */

    if (!buff)
    {
	fprintf(Log_fp, "ERROR: get_through_header() got NULL buff\n");
	fflush(Log_fp);

	return FALSE;
    } /* endif */

    if (nbytes <= 0)
    {
	fprintf(Log_fp, 
	    "ERROR: get_through_header() got nbytes %d\n", nbytes);
	fflush(Log_fp);

	return FALSE;
    } /* endif */

    /* start header */
    if (fgets(buff, nbytes, fp))
    {
	buff[strlen(buff)-1] = '\0';
	if (strcmp(buff, STARTHEADERSTRING))
	{
	    fprintf(Log_fp,
	    "ERROR: get_through_header() found illegal first line\n");
	    fflush(Log_fp);

	    return FALSE;
	} /* endif */
    }
    else
    {
	fprintf(Log_fp,
	    "ERROR: get_through_header() could not read first line\n");
	fflush(Log_fp);

	return FALSE;
    } /* endif */

    /* Time */
    if (fgets(buff, nbytes, fp))
    {
	if (buff != strstr(buff, OLDBURT_TIMEHEADERSTRING))
	{
	    fprintf(Log_fp,
		"ERROR: get_through_header() found illegal time\n");
	    fflush(Log_fp);

	    return FALSE;
	}
	else
	{
	    (snapshot->get_parms())->
		set_currtime(buff+strlen(OLDBURT_TIMEHEADERSTRING)+1);
	} /* endif */
    }
    else
    {
	fprintf(Log_fp,
	    "ERROR: get_through_header() could not read time\n");
	fflush(Log_fp);

	return FALSE;
    } /* endif */

    /* Login Id */
    if (fgets(buff, nbytes, fp))
    {
	if (buff != strstr(buff, OLDBURT_LOGINHEADERSTRING))
	{
	    fprintf(Log_fp,
		"ERROR: get_through_header() found illegal login id\n");
	    fflush(Log_fp);

	    return FALSE;
	} /* endif */

	if (!(t1p = strchr(buff+strlen(OLDBURT_LOGINHEADERSTRING)+1, ' ')))
	{
	    fprintf(Log_fp,
		"ERROR: get_through_header() found illegal login id\n");
	    fflush(Log_fp);

	    return FALSE;
	} /* endif */

	*t1p = '\0';
	(snapshot->get_parms())->
	    set_loginid(buff+strlen(OLDBURT_LOGINHEADERSTRING)+1);
	*t1p = ' ';

	if (!(t2p = strchr(t1p, '(')))
	{
	    fprintf(Log_fp,
		"ERROR: get_through_header() found illegal login id\n");
	    fflush(Log_fp);

	    return FALSE;
	} /* endif */

	if (!(t3p = strchr(t2p, ')')))
	{
	    fprintf(Log_fp,
		"ERROR: get_through_header() found illegal login id\n");
	    fflush(Log_fp);

	    return FALSE;
	} /* endif */

	*t3p = '\0';
	(snapshot->get_parms())->
	    set_real_world_name(t2p+1);
	*t3p = ')';

    }
    else
    {
	fprintf(Log_fp,
	    "ERROR: get_through_header() could not read login id\n");
	fflush(Log_fp);

	return FALSE;
    } /* endif */

    /* Eff Id */
    if (fgets(buff, nbytes, fp))
    {
	if (buff != strstr(buff, OLDBURT_EFFUIDHEADERSTRING))
	{
	    fprintf(Log_fp,
	    "ERROR: get_through_header() found illegal effective id\n");
	    fflush(Log_fp);

	    return FALSE;
	}
	else
	{
	    (snapshot->get_parms())->
    set_effective_uid((uid_t) atoi(buff+strlen(OLDBURT_EFFUIDHEADERSTRING)+1));
	} /* endif */
    }
    else
    {
	fprintf(Log_fp,
	    "ERROR: get_through_header() could not read effective id\n");
	fflush(Log_fp);

	return FALSE;
    } /* endif */

    /* Group Id */
    if (fgets(buff, nbytes, fp))
    {
	if (buff != strstr(buff, OLDBURT_GROUPIDHEADERSTRING))
	{
	    fprintf(Log_fp,
		"ERROR: get_through_header() found illegal group id\n");
	    fflush(Log_fp);

	    return FALSE;
	}
	else
	{
	    (snapshot->get_parms())->
	set_group_id((gid_t) atoi(buff+strlen(OLDBURT_GROUPIDHEADERSTRING)+1));
	} /* endif */
    }
    else
    {
	fprintf(Log_fp,
	    "ERROR: get_through_header() could not read group id\n");
	fflush(Log_fp);

	return FALSE;
    } /* endif */

    /* Keywords */
    if (fgets(buff, nbytes, fp))
    {
	if (buff != strstr(buff, OLDBURT_KEYWORDSHEADERSTRING))
	{
	    fprintf(Log_fp,
		"ERROR: get_through_header() found illegal keywords\n");
	    fflush(Log_fp);

	    return FALSE;
	}
	else
	{
	    /* removing trailing <CR> ... if there is one */
	    if (t1p = strchr(buff, '\n'))
		*t1p = '\0';

	    (snapshot->get_parms())->
		set_keywords(buff+strlen(OLDBURT_KEYWORDSHEADERSTRING)+1);
	} /* endif */
    }
    else
    {
	fprintf(Log_fp,
	    "ERROR: get_through_header() could not read keywords\n");
	fflush(Log_fp);

	return FALSE;
    } /* endif */

    /* Comments */
    if (fgets(buff, nbytes, fp))
    {
	if (buff != strstr(buff, OLDBURT_COMMENTSHEADERSTRING))
	{
	    fprintf(Log_fp,
		"ERROR: get_through_header() found illegal comments\n");
	    fflush(Log_fp);

	    return FALSE;
	}
	else
	{
	    /* removing trailing <CR> ... if there is one */
	    if (t1p = strchr(buff, '\n'))
		*t1p = '\0';

	    (snapshot->get_parms())->
		set_comments(buff+strlen(OLDBURT_COMMENTSHEADERSTRING)+1);
	} /* endif */
    }
    else
    {
	fprintf(Log_fp,
	    "ERROR: get_through_header() could not read comments\n");
	fflush(Log_fp);

	return FALSE;
    } /* endif */

    /* Type */
    if (fgets(buff, nbytes, fp))
    {
	if (buff != strstr(buff, OLDBURT_TYPEHEADERSTRING))
	{
	    /* type not specified ... assuming absolute */
	    (snapshot->get_parms())->set_absolute();

	    fprintf(Log_fp, 
"WARNING: parm %s (Type of snapshot file) not specified ... assuming %s = %s\n",
		OLDBURT_TYPEHEADERSTRING, OLDBURT_TYPEHEADERSTRING,
		ABSOLUTESTRING);
	    fflush(Log_fp);

	    return process_optional_request_and_end(fp, snapshot, buff, nbytes);
	}
	else
	{
	    if (strstr(buff, ABSOLUTESTRING))
		(snapshot->get_parms())->set_absolute();
	    else if (strstr(buff, RELATIVESTRING))
		(snapshot->get_parms())->set_relative();
	    else if (strstr(buff, NOWRITESTRING))
		(snapshot->get_parms())->set_nowrite();
	    else
	    {
		fprintf(Log_fp,
		    "ERROR: get_through_header() found illegal type\n");
		fflush(Log_fp);

		return FALSE;
	    } /* endif */
	} /* endif */
    }
    else
    {
	fprintf(Log_fp,
	    "ERROR: get_through_header() could not read end header\n");
	fflush(Log_fp);

	return FALSE;
    } /* endif */

    /* {Directory + Req1 + Req2 + ... + Reqn} End header */
    if (fgets(buff, nbytes, fp))
	return process_optional_request_and_end(fp, snapshot, buff, nbytes);
    else
    {
	fprintf(Log_fp,
	    "ERROR: get_through_header() could not read end header\n");
	fflush(Log_fp);

	return FALSE;
    } /* endif */

} /* end get_through_header() */

/****************************************************************
*
*
****************************************************************/

static BOOL nextchar(FILE *const fp, const char c)
{

BOOL rc;

    if (fp)
        rc = (getc(fp) == (int) c);
    else
        rc = FALSE;

    return rc;

} /* end nextchar() */

/****************************************************************
*
* assumes values are spaced delimited in cp
* spaces inside quotes ignored
* all quotes must match ... must be even number of quotes
*
* dynamically allocates what is returned ... assumes caller frees it.
*
* places value(s) into char * buffer where the size of each value
* is dbr_size_n(DBR_STRING, 1) (about 40 bytes) long.
*
****************************************************************/

static char *prepare_val(char *const cp, const int nelem)
{

int i;
char *source;
char *dest;
char *tempp, tempc;
BOOL foundquote;
char *rc;

    if (cp)
    {
	if (rc = 
	(char *) calloc((unsigned) nelem, (unsigned) dbr_size_n(DBR_STRING, 1)))
	{
	    /* passing over leading whitespace */
	    for (source = cp; *source && isspace(*source); source ++)
	    ;

	    for (i = 0, dest = rc, source = cp; 
		rc && i < nelem; i ++, 
		    dest = dest + dbr_size_n(DBR_STRING, 1))
	    {
		if (*source == '\"')
		{
		    foundquote = TRUE;
		    source ++;

		    /* looking for ending quote */
		    for (tempp = source; *tempp && *tempp != '\"'; tempp ++)
		    ;

		    if (!(*tempp))
		    {
			/* did not find ending quote */
			fprintf(Log_fp, 
"ERROR: prepare_val() passed string with unmatching pair of quotes >%s<\n", 
			    cp);
			fflush(Log_fp);
			BurtExit(1);
		    } /* endif */
		}
		else
		{
		    foundquote = FALSE;
		    /* passing over value */
		    for (tempp = source; *tempp && !isspace(*tempp); tempp ++)
		    ;
		} /* endfor */

		if (tempp > source)
		{
		    /* got a value */
		    tempc = *tempp;
		    *tempp = '\0';
		    if (strcmp(source, NULLSTRING))
		    {
			if (strlen(source) < dbr_size_n(DBR_STRING, 1))
			    strcpy(dest, source);
			else
			{
			    fprintf(Log_fp, 
"ERROR: prepare_val() could not copy string >%s< length %d into CA buff that has length %d\n",
				source, strlen(source), 
				dbr_size_n(DBR_STRING, 1));
			    fflush(Log_fp);
			    BurtExit(1);
			} /* endif */
		    } /* endif */
		    *tempp = tempc;

		    /* passing over trailing whitespace */
		    for (source = (foundquote ? tempp+1 : tempp) ; 
			*source && isspace(*source); 
			    source ++)
		    ;

		    /* assuring NULL-terminated string */
		    *(dest+(dbr_size_n(DBR_STRING, 1)-1)) = '\0';

		}
		else
		{
		    fprintf(Log_fp, 
			"ERROR: prepare_val() found not enough values\n");
		    fflush(Log_fp);

		    free(rc);
		    rc = (char *) NULL;

		} /* endif */
	    } /* endfor */
	} 
	else
	{
	    fprintf(Log_fp, 
	    "ERROR: prepare_val() could not calloc %d x %d = %d bytes\n", 
		nelem, dbr_size_n(DBR_STRING, 1), 
		nelem*dbr_size_n(DBR_STRING, 1));
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
    }
    else
	rc = (char *) NULL;

    return rc;

} /* end prepare_val() */

/****************************************************************
*
* recursive function
*
****************************************************************/

#ifndef NODEVICES
static void process_device(Burtobjlist *const snapshot, 
    const char *const thisdevice, const char *const inlineage, 
    SDDS_TABLE *const SDDS_table, const int row, const BOOL mode_col_present,
    const BOOL nelem_col_present, const BOOL readmsg_col_present, 
    const BOOL writemsg_col_present)
{

int count;
devnamearr mynames;
devstructarr myinfo;
char *cp;
Burtatomdev *dp;
long l;
int nreq;
int i;
int nbytes;
char *outlineage;
BOOL needs_free;

    if (!thisdevice || !inlineage)
    {
	fprintf(Log_fp, 
	    "ERROR: process_device() got thisdevice %x inlineage %x\n",
	    thisdevice, inlineage);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!(*thisdevice))
    {
	fprintf(Log_fp, "ERROR: process_device() got empty thisdevice\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    /* getting information about this device */
    count = dev_count(thisdevice, &mynames);
    if (count == 1)
    {
	/* an atomic device */
	dp = new Burtatomdev;
	(dp->get_readchannel())->set_val(dp->get_val());
	(dp->get_writechannel())->set_val(dp->get_val());

	dp->set_name(mynames[0]);
	dp->set_lineage(inlineage);

	/* establishing readonly */
	if (mode_col_present)
	{
	    sdds_get_value(SDDS_table, MODE_COL, row, &cp);

	    if (strcmp(cp, READONLYNOTIFYSTRING))
	    {
		
		dp->clear_read_only_notify();

		if (strcmp(cp, READONLYSTRING))
		    dp->clear_read_only();
		else
		    dp->set_read_only();
	    }
	    else
	    {
		dp->set_read_only_notify();
		dp->set_read_only();
	    } /* endif */
		
	    free(cp);
	}
	else
	{
	    dp->clear_read_only_notify();
	    dp->clear_read_only();
	} /* endif */

	/* getting nelem */
	if (nelem_col_present)
	{
	    sdds_get_value(SDDS_table, NELEM_COL, row, &l);
	    nreq = ((int) l ? (int) l : UNKNOWN);
	} 
	else
	    nreq = UNKNOWN;
	(dp->get_readchannel())->set_nreq(nreq);

	/* getting read/write msgs and pvars */
	dev_info(mynames[0], &myinfo);

	/*********************************/
	/*                               */
	/* getting read message and pvar */
	/*                               */
	/*********************************/

	/* getting read msg */
	if (readmsg_col_present)
	{
	    sdds_get_value(SDDS_table, READMSG_COL, row, &cp);

	    if (strcmp(cp, DEFAULTSTRING))
		needs_free = TRUE;
	    else
	    {
		free(cp);
		needs_free = FALSE;

		cp = DEFAULTREADMSG;
	    } /* endif */
	}
	else
	{
	    needs_free = FALSE;
	    cp = DEFAULTREADMSG;
	} /* endif */

	dp->set_readmsg(cp);

	/* searching for pvar associated */
	/* with read msg in this device  */

	/* translating readmsg for this device */
	/* into the appropriate pvname         */

	for (i = 0; myinfo[i] && strcmp(myinfo[i]->msg, cp); i ++)
	;

	if (myinfo[i])
	    /* got it */
	    (dp->get_readchannel())->set_name(myinfo[i]->pvname);
	else
	{
	    /* not there */
	    fprintf(Log_fp, 
	    "ERROR: process_device() devname >%s< did not recognize msg >%s<\n",
		mynames[0], cp);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	if (needs_free)
	    free(cp);

	/**********************************/
	/*                                */
	/* getting write message and pvar */
	/*                                */
	/**********************************/

	/* geting write msg */
	if (writemsg_col_present)
	    sdds_get_value(SDDS_table, WRITEMSG_COL, row, &cp);
	else
	    cp = DEFAULTWRITEMSG;

	dp->set_writemsg(cp);

	/* searching for pvar associated  */
	/* with write msg in this device  */

	/* translating writemsg for this device */
	/* into the appropriate pvname          */

	for (i = 0; myinfo[i] && strcmp(myinfo[i]->msg, cp); i ++)
	;

	if (myinfo[i])
	    /* got it */
	    (dp->get_writechannel())->set_name(myinfo[i]->pvname);
	else
	{
	    /* not there */
	    fprintf(Log_fp, 
	    "ERROR: process_device() devname >%s< did not recognize msg >%s<\n",
		mynames[0], cp);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */

	if (writemsg_col_present)
	    free(cp);

	snapshot->append((Burtobj *) dp);

	if (Verbose)
	{
	    fprintf(Log_fp, 
	"dev >%s< readpv >%s< nreq=%d writepv >%s<\n", 
		dp->get_name(), 
		(dp->get_readchannel())->get_name(), 
		(dp->get_readchannel())->get_nreq());
		(dp->get_writechannel())->get_name(), 
	    fflush(Log_fp);
	} /*endif */

	free(mynames[0]);
    }
    else
    {
	/* a composite device */

	if (strcmp(inlineage, DEFAULTSTRING))
	{
	    /* lineage has already been started */
	    nbytes = strlen(inlineage) + strlen(LINEAGEDELIMETER)
		+ strlen(thisdevice) + 1;

	    if (outlineage = (char *) malloc(nbytes))
		sprintf(outlineage, "%s%s%s", 
		    inlineage, LINEAGEDELIMETER, thisdevice);
	    else
	    {
		fprintf(Log_fp, 
	"ERROR: process_device() could not malloc %d bytes for outlineage\n",
		    nbytes);
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */
	}
	else
	{

	    /* this composite device is the root of the lineage */

	    nbytes = strlen(thisdevice) + 1;

	    if (outlineage = (char *) malloc(nbytes))
		strcpy(outlineage, thisdevice);
	    else
	    {
		fprintf(Log_fp, 
	"ERROR: process_device() could not malloc %d bytes for outlineage\n",
		    nbytes);
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */
	} /* endif */

	for (i = 0; i < count; i ++)
	{
	    process_device(snapshot, mynames[i], outlineage, SDDS_table, row, 
		mode_col_present, nelem_col_present, readmsg_col_present, 
		writemsg_col_present);
	    free(mynames[i]);
	} /* endfor */

	free(outlineage);
    } /* endif */

    free((char *) mynames);

} /* end process_device() */
#endif

/****************************************************************
*
*
****************************************************************/

static BOOL process_optional_request_and_end(FILE *const fp, 
    Burtobjlist *const snapshot, char *const buff, const int nbytes)
{

char *dirname;
char *fname;

    dirname = (char *) NULL;

    buff[strlen(buff)-1] = '\0';
    if (buff == strstr(buff, OLDBURT_CWDHEADERSTRING))
    {
	/* encountered optional direcotry for list of request files */

	for (dirname = buff + strlen(OLDBURT_CWDHEADERSTRING); 
	    *dirname && isspace(*dirname); 
		dirname ++)
	;

	if (*dirname)
	    dirname = strdup(dirname);
	else
	    dirname = (char *) NULL;

	/* loop to process through all the request filenames */
	if (fgets(buff, nbytes, fp))
	{
	    buff[strlen(buff)-1] = '\0';
	    while (buff == strstr(buff, OLDBURT_REQFNAMEHEADERSTRING))
	    {
		for (fname = buff + strlen(OLDBURT_REQFNAMEHEADERSTRING);
		    *fname && isspace(*fname);
			fname ++)
		;

		if (*fname)
		{
		    if (*fname == '/')
			/* absolute filename */
			(snapshot->get_request_filenames())->
			    append(new Filename(fname));
		    else
			/* relative filename */
			(snapshot->get_request_filenames())->
			    append(new Filename(dirname, fname));
		} /* endif */

		if (!fgets(buff, nbytes, fp))
		{
		    fprintf(Log_fp,
		"ERROR: get_through_header() could not read end header\n");
		    fflush(Log_fp);

		    if (dirname)
			free(dirname);
		    return FALSE;
		} /* endif */
		buff[strlen(buff)-1] = '\0';
	    } /* endwhile */
	}
	else
	{
	    fprintf(Log_fp,
		"ERROR: get_through_header() could not read end header\n");
	    fflush(Log_fp);

	    if (dirname)
		free(dirname);
	    return FALSE;
	} /* endif */
    } /* endif */

    if (strcmp(buff, ENDHEADERSTRING))
    {
	fprintf(Log_fp,
	"ERROR: get_through_header() found illegal end header\n");
	fflush(Log_fp);

	if (dirname)
	    free(dirname);
	return FALSE;
    } /* endif */

    if (dirname)
	free(dirname);

    return TRUE;

} /* end process_optional_request_and_end() */

/****************************************************************
*
*
****************************************************************/

static BOOL sdds_get_parm(SDDS_TABLE *const SDDS_table, char *const parm, 
    void *const mem)
{

BOOL rc;

    if (SDDS_table && parm && mem)
    {
	if (*parm)
	{
	    if (BurtSDDSParmExists(SDDS_table, parm))
	    {
		if (SDDS_GetParameter(SDDS_table, parm, mem))
		    rc = TRUE;
		else
		{
		    rc = FALSE;

		    fprintf(Log_fp, 
			"SYS ERR: SDDS_GetParameter() for parm >%s<.\n", 
			parm);
		    fflush(Log_fp);
		} /* endif */
	    }
	    else
		rc = FALSE;
	}
	else
	{
	    rc = FALSE;

	    fprintf(Log_fp, "ERROR: sdds_get_parm() got empty parm\n");
	    fflush(Log_fp);
	} /* endif */
    }
    else
    {
	rc = FALSE;

	fprintf(Log_fp, 
	    "ERROR: sdds_get_parm() got SDDS_table %x parm %x mem %x\n",
	    SDDS_table, parm, mem);
	fflush(Log_fp);
    } /* endif */

    return rc;

} /* end sdds_get_parm() */

/****************************************************************
*
*
****************************************************************/

static void sdds_get_value(SDDS_TABLE *const SDDS_table, char *const colname, 
    const long rowidx, void *const mem)
{

    if (SDDS_table && colname && mem)
    {

	if (*colname)
	{
	    if (!SDDS_GetValue(SDDS_table, colname, rowidx, mem))
	    {
		fprintf(Log_fp, 
		    "ERROR: could not SDDS_GetValue() for column >%s<\n", 
		    colname);
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */
	}
	else
	{
	    fprintf(Log_fp, 
		"ERROR: sdds_get_value() got colname >%s<\n", colname);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
    }
    else
    {
	fprintf(Log_fp, 
	    "ERROR: sdds_get_value() got SDDS_table %x colname %x mem %x\n", 
	    SDDS_table, colname, mem);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

} /* end sdds_get_value() */

/*************/
/*           */
/* Callbacks */
/*           */
/*************/

/****************************************************************
*
* from epicsH/cadef.h
* struct  connection_handler_args
* {
*     chid chid;    Channel id
*     long op;      External codes for channel access operations
* };
*
* Presumably a ca_search_and_connect() has been previously called which 
* named this routine as the callback.  
*
* As an interesting side quesiton:  
*     When we did the the ca_search_and_connect() we could have passed
*     a user arg that would be passed to this routine.  We did not pass
*     one, but if we did, where would we find it in struct 
*     connection_handler_args?
*
****************************************************************/

extern "C" void my_connection_callback(struct connection_handler_args arg)
{
    /* totally useless executed statement, however only way I know how */
    /* to get rid of annoying compiler message that tells me that this */
    /* argument is not used.                                           */
    if (arg.op);

    NoutstandingSearches --;
} /* end my_connection_callback() */

/****************************************************************
*
* from epicsH/cadef.h
* struct  event_handler_args
* {
*     void *usr;   User argument supplied when event added
*     chid chid;   Channel id
*     long type;   the type of the value returned (long aligned)
*     long count;  the element count of the item returned
*     void *dbr;   Pointer to the value returned
* };
*
****************************************************************/
 
extern "C" void my_get_callback(struct event_handler_args arg)
{
 
Channel *cp;

    if (cp = (Channel *) arg.usr)
    {
	/* checking that channel access gave us what we asked for */
	if (arg.type == cp->get_dbr_req_type())
	{
	    if (cp->get_nreq() != UNKNOWN && arg.count != cp->get_nreq())
	    {
		fprintf(Log_fp,
	    "WARNING: %s requested %d values and actually read %d values\n", 
		    cp->get_nreq(), arg.count);
		fflush(Log_fp);
	    } /* endif */

	    (cp->get_val())->set((char) arg.type, (int) arg.count, arg.dbr);
	    cp->set_reported();
	}
	else
	{
	    fprintf(Log_fp,
	"ERROR: my_get_callback() ca delivered DBR %d while %d was expected\n",
		arg.type, cp->get_dbr_req_type());
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
    }
    else
    {
	fprintf(Log_fp, "ERROR: my_get_callback() got NULL cp\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */
 
} /* end my_get_callback() */

