/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#ifndef __burtcommon__
#define __burtcommon__

#include <stdlib.h> /* for exit() on HP's */
#include <stdio.h>
#include <math.h>      /* for log10() */
#ifndef WIN32
#include <sys/param.h> /* for MAXPATHLEN */
#else
#define MAXPATHLEN   512
#endif

#include "burtpublic.h"
#define BURT_CPP_ENV_NAME	"BURT_CPP_PARMS"

#ifdef MAINLINE
#define PUBLIC 
#else
#define PUBLIC extern
#endif

#define ABS(X) ((X) < 0 ? (0-(X)) : (X))

/* this only counts the bytes for the integer portion of X */
/* includes a byte for '\0' and another for "-" (if X<0) */
#define BYTESFORSTRING(X) \
    ((X) \
	? (((X) < 0 ? 1 : 0) + ((unsigned int) log10((double) ABS((X))) + 2)) \
        : 2)

/* snapshot arithmetic operations */
#define ADDITION       0
#define SUBTRACTION    1
#define MULTIPLICATION 2

/* the following 2 must agree ... (-x.)+6+(+exx)=13 */
#define FLOATFORMATSTRING "%.6e"
#define FLOATSTRINGLENGTH 13

/* the following 2 must agree ... (-x.)+15+(+exx)=22 */
#define DOUBLEFORMATSTRING "%.15e"
#define DOUBLESTRINGLENGTH 22

/* Start Non-SDDS Snapshot Parameters */
#define STARTHEADERSTRING "--- Start BURT header"
#define OLDBURT_TIMEHEADERSTRING     "Time:    "
#define OLDBURT_LOGINHEADERSTRING    "Login ID:"
#define OLDBURT_EFFUIDHEADERSTRING   "Eff  UID:"
#define OLDBURT_GROUPIDHEADERSTRING  "Group ID:"
#define OLDBURT_KEYWORDSHEADERSTRING "Keywords:"
#define OLDBURT_COMMENTSHEADERSTRING "Comments:"
#define OLDBURT_TYPEHEADERSTRING     "Type:    "
#define OLDBURT_CWDHEADERSTRING      "Directory"
#define OLDBURT_REQFNAMEHEADERSTRING "Req File:"
#define ENDHEADERSTRING   "--- End BURT header"
/* End Non-SDDS Snapshot Parameters */

#define BOOL char
#define FALSE 0
#define TRUE  1

#define SHORT_TIME   0.0001

/* for maximum -D and -I switches on command line */
#define MAXDEFINES  100
#define MAXINCLUDES 100

/* for command line sdds specification */
#define NOREQ     0
#define REQSDDS   1
#define REQNOSDDS 2

/* for input set characterization */
#define HETROINPUT   0
#define SDDSINPUT    1
#define NONSDDSINPUT 2

/* DBR types */
#define INVALID_DBR -1
#define NATIVE      -2

/* predicate directives on fail */
#define QUERYONFAIL 0 /* default */
#define HALTONFAIL  1
#define CONTONFAIL  2

/* Snap Shot Call Print Types */
#define SS_NOSTRING 0
#define SS_STRING   1
/* write directives */
#define DEFAULTWRITE 0
#define REPLACEWRITE 1 /* Relative snapshots written as replaces */
#define ADDWRITE     2 /* Absolute snapshots written as adds     */

#define PUBLICWRITEDIRECTORY "/tmp/"

/********************/
/*                  */
/* Global Variables */
/*                  */
/********************/

PUBLIC FILE *Log_fp;
PUBLIC FILE *Out_fp;
PUBLIC BOOL Verbose;
PUBLIC BOOL Sdds;
PUBLIC char *Outfilename;

PUBLIC char *Sh_path;
PUBLIC char *Sh_arg1;
PUBLIC char *Exec_args[4];

PUBLIC unsigned int NoutstandingSearches;

#endif 
