/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#define MAINLINE

#include "burtcommon.h"
#include "burtobjs.h"
#include "burtconvertsnap.h"

#include "list.h"
#include "burtlist.h"

/****************************************/
/*                                      */
/* Externally Visibile Remote Functions */
/*                                      */
/****************************************/

/* in burtcommon.cc */
void BurtCommonInit(const BOOL);
Burtobjlist *BurtProcessSnapshotFile(char *const);
BOOL BurtSDDSFile(const char *const);
void BurtExit(const int);

/***************************/
/*                         */
/* Private Local Functions */
/*                         */
/***************************/

static void parse_command_line_args(int, char **, char **, char **, char **,
    BOOL *, char *);
static void process_command_line_args(char *, char *, char *, Burtobjlist **, 
    char);

/****************************************************************
*
*
****************************************************************/

int main(int argc, char **argv)
{

char *logfilename;
char *snapfilename;
Burtobjlist *snapshot;
char output_spec;

    BurtCommonInit(FALSE);

    parse_command_line_args(argc, argv, &logfilename, &Outfilename,
	&snapfilename, &Verbose, &output_spec);

    process_command_line_args(logfilename, Outfilename, snapfilename, &snapshot,
	output_spec);

    snapshot->write_snapshot_to_file();

    BurtExit(0);

} /* end main() */

/******************************************************
*
*
******************************************************/

static void parse_command_line_args(int argc, char **argv, char **logfilename, 
    char **outfilename, char **snapfilename, BOOL *verbose, char *output_spec)
{

int arg;
int context;
int next_snap_file;
BOOL error;

    /* default values */
    *snapfilename = (char *) NULL;
    *logfilename = (char *) NULL;
    *outfilename = (char *) NULL;
    *verbose = FALSE;
    *output_spec = NOREQ;

    next_snap_file = 0;
    context = SNAP_FILES;
    arg = 1;
    error = FALSE;
    while (arg < argc && !error)
    {
/* printf("error %c argc %d arg %d >%s< context %s\n", (error ? 'T' : 'F'), argc, arg, argv[arg], (context == EXPECT_SWITCH ? "expect switch" : (context == SNAP_FILES ? "snap files" : (context == LOG_FILE ? "log file" : (context == OUT_FILE : "out file" : "unknown"))))); */
	switch (context)
	{
	    case EXPECT_SWITCH:
		if (strcmp(argv[arg], "-l") == 0)
		{
		    if (*logfilename)
			/* already have a logfile from */
			/* a previous -l switch        */
			error = TRUE;
		    else
		    {
			arg ++;
			context = LOG_FILE;
		    } /* endif */
		}
		else if (strcmp(argv[arg], "-v") == 0)
		{
		    arg ++;
		    *verbose = TRUE;
		}
                else if (strcmp(argv[arg], "-o") == 0)
                {
                    if (*outfilename)
                        /* already have a outfile from */
                        /* a previous -o switch        */
                        error = TRUE;
                    else
                    {
                        arg ++;
                        context = OUT_FILE;
                    } /* endif */
                }
                else if (strcmp(argv[arg], "-sdds") == 0)
                {
                    if (*output_spec == NOREQ)
                    {
                        arg ++;
                        *output_spec = REQSDDS;
                    }
                    else
                        error = TRUE;
                }
                else if (strcmp(argv[arg], "-nosdds") == 0)
                {
                    if (*output_spec == NOREQ)
                    {
                        arg ++;
                        *output_spec = REQNOSDDS;
                    }
                    else
                        error = TRUE;
                }
		else 
		    /* did not get recognizable switch */
		    error = TRUE;
		break;
	    case SNAP_FILES:
		if (*(argv[arg]) == '-')
		{
		    context = EXPECT_SWITCH;
		}
		else
		{
		    if (next_snap_file == 0)
		    {
			*snapfilename = argv[arg];
			next_snap_file ++;
			arg ++;
		    }
		    else
		    {
			error = TRUE;
		    } /* endif */
		} /* endif */
		break;
	    case LOG_FILE:
		if (*(argv[arg]) == '-')
		    error = TRUE;
		else
		{
		    /* got logfilename */
		    *logfilename = argv[arg];
		    arg ++;
		    context = EXPECT_SWITCH;
		} /* endif */
		break;
            case OUT_FILE:
                if (*(argv[arg]) == '-')
                    error = TRUE;
                else
                {
                    /* got outfilename */
                    *outfilename = argv[arg];
                    arg ++;
                    context = EXPECT_SWITCH;
                } /* endif */
                break;
	    default:
		error = TRUE;
		break;
	} /* end switch() */
    } /* endwhile */

/* printf("out of loop error %c arg %d context %s\n", (error ? 'T' : 'F'), arg, (context == EXPECT_SWITCH ? "expect switch" : (context == SNAP_FILES ? "snap files" : (context == LOG_FILE ? "log file" : (context == OUT_FILE ? "out file" : "unknown"))))); */

    error |= (context == LOG_FILE || context == OUT_FILE || !next_snap_file);

    if (error)
    {
printf("\n\tusage: %s snap {-l logfile} {-o outfile}\n", argv[0]);
printf("\t\t{-v} {-sdds or -nosdds}\n");

printf("where\n\n");

printf("\tsnap - Snapshot filename.  This is the only switch that is not\n");
printf("\t\toptional.  You must specify exactly one snapshot file.  This is\n");
printf("\t\tthe snapshot file that will be converted.\n\n");

printf("\t-l logfile - Log filename. The name of the file where all logging\n");
printf("\t\tmessages (e.g. error messages) go.  The default is stderr.\n\n");

printf("\t-o outfile - Output snapshot filename.  This is where the newly\n");
printf("\t\tgenerated converted snapshot file will be placed. The default\n");
printf("\t\tis stdout.\n\n");

printf("\t-v - Verbose.  This increases the amount of information displayed\n");
printf("\t\tin the logfile.\n\n");

printf("\t-sdds or -nosdds - SDDS/non-SDDS snapshot file.  Explicitly\n");
printf("\t\tspecifying that the generated snapshot file will be \n");
printf("\t\tSDDS/non-SDDS compliant.  The default is to produce whatever\n");
printf("\t\tthe input snapshot file is not.\n\n");

	exit(1);

    } /* endif */

} /* end parse_command_line_args() */

/******************************************************
*
*
******************************************************/

static void process_command_line_args(char *logfilename, char *outfilename,
    char *snapfilename, Burtobjlist **snapshot, char output_spec)
{

char input_profile;

    if (logfilename)
    {
	if (!(Log_fp = fopen(logfilename, "w")))
	{
	    printf("ERROR: unable to open logfile >%s<\n", logfilename);
	    exit(1);
	} /* endif */
    }
    else
	Log_fp = stderr;

    if (Verbose)
    {
	fprintf(Log_fp, "**** CONVERT SNAPSHOT BURT LOGFILE\n");
	fflush(Log_fp);
    } /* endif */

    if (!snapfilename)
    {
	fprintf(Log_fp,
	    "ERROR: process_command_line_args() got NULL snapfilename\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!(*snapfilename))
    {
	fprintf(Log_fp,
	    "ERROR: process_command_line_args() got empty snapfilename\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!snapshot)
    {
	fprintf(Log_fp,
	    "ERROR: process_command_line_args() got NULL snapshot\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (Verbose)
    {
	fprintf(Log_fp, "\n--- Start processing snapshot file\n");
	fflush(Log_fp);
    } /* endif */

    if (BurtSDDSFile(snapfilename))
	input_profile = SDDSINPUT;
    else
	input_profile = NONSDDSINPUT;

    if (*snapshot = BurtProcessSnapshotFile(snapfilename))
	(*snapshot)->set_owner();
    else
    {
	fprintf(Log_fp, 
"ERROR: process_command_line_args() got NULL from BurtProcessSnapshotFile()\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (Verbose)
    {
	fprintf(Log_fp, "--- End processing snapshot file\n");
	fflush(Log_fp);
    } /* endif */

    if (output_spec == REQSDDS)
        Sdds = TRUE;  /* explictly wants sdds */
    else if (output_spec == REQNOSDDS)
        Sdds = FALSE;  /* explictly does not want sdds */
    else if (input_profile == SDDSINPUT)
        Sdds = FALSE;  /* no specification, all input files sdds */
    else if (input_profile == NONSDDSINPUT)
        Sdds = TRUE;  /* no specification, all input files non-sdds */
    else
        Sdds = TRUE;  /* default ... no specifation, hetrogeneous input */

    if (!Sdds)
    {
        if (outfilename)
        {
            if (!(Out_fp = fopen(outfilename, "w")))
            {
                fprintf(Log_fp,
                    "ERROR: unable to open outfile >%s<\n", outfilename);
                fflush(Log_fp);
                BurtExit(1);
            } /* endif */
        }
        else
            Out_fp = stdout;
    } /* endif */

} /* end process_command_line_args() */

