/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#define MAINLINE

#include "burtgooey.h"
#include "burtgooeyfuncdecls.h"

#include "fallback.h"

/**************/
/*            */
/* main shell */
/*            */
/**************/

/* in read_burtgooey.c */
void main_read_button_cb();
void cleanup_read();

/* in write_burtgooey.c */
void main_write_button_cb();
void cleanup_write();

/* in addsub_burtgooey.c */
void main_addsub_button_cb();
void cleanup_addsub();

/* in mult_burtgooey.c */
void main_mult_button_cb();
void cleanup_mult();

/* in set_burtgooey.c */
void main_set_button_cb();
void cleanup_set();

/* in definc_burtgooey.c */
void main_definc_button_cb();

static void main_shared_button_cb();
static void dialog_ok_button_cb();

static void parse_command_line_args();
int getopt();

/****************************************************************
*
*
****************************************************************/

int main(argc, argv)
int argc;
char **argv;
{

/**************/
/*            */
/* main shell */
/*            */
/**************/

Widget shell, form, read_button, write_button; 
Widget addsub_button, mult_button, set_button, definc_button;
Widget help_button, done_button;

XmString xs;
Arg args[NARGS];
int nargs;
int i;
XtAppContext ac;

    /* initialization */
    Log_fp = stderr;

    Read_setup = FALSE;
    Write_setup = FALSE;
    Addsub_setup = FALSE;
    Mult_setup = FALSE;
    Set_setup = FALSE;
    Definc_setup = FALSE;

    /**************/
    /*            */
    /* main shell */
    /*            */
    /**************/

    nargs = 0;
    XtSetArg(args[nargs], XmNtitle, "BURT"); nargs ++;
    XtSetArg(args[nargs], XmNiconName, "BURT"); nargs ++;
    XtSetArg(args[nargs], XmNmwmDecorations, (MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU)); nargs ++;
    shell = XtAppInitialize(&ac, "shell", NULL, 0, &argc, argv, 
		fallback_resources, args, nargs);

    parse_command_line_args(argc, argv, Defines, Includes);

    /* XtVaSetValues(shell, */
	/* XmNallowShellResize, False, */
	/* NULL); */

    Red_pixel = color_string_to_pixel(shell, "red");
    Green_pixel = color_string_to_pixel(shell, "green");
    Blue_pixel = color_string_to_pixel(shell, "blue");

    /* Error Dialog */
    Error_dialog = XmCreateErrorDialog(shell, "Error_dialog", NULL, 0);
    /* disabling help and cancel buttons on error dialog */
    XtUnmanageChild(XmMessageBoxGetChild(Error_dialog, XmDIALOG_HELP_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(Error_dialog, XmDIALOG_CANCEL_BUTTON));

    /* Message Dialog */
    Message_dialog = XmCreateMessageDialog(shell, "Message_dialog", NULL, 0);
    /* disabling help and cancel buttons on message dialog */
    XtUnmanageChild(XmMessageBoxGetChild(Message_dialog, XmDIALOG_HELP_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(Message_dialog, 
	XmDIALOG_CANCEL_BUTTON));

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    form = XtCreateWidget("form", xmFormWidgetClass, shell, args, nargs);

    /* for all the main buttons */
    nargs = 0;
    XtSetArg(args[nargs], XmNheight, 50); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 50); nargs ++;

    read_button = XmCreatePushButton(form, "read_button", args, nargs);
    XtManageChild(read_button);

    write_button = XmCreatePushButton(form, "write_button", args, nargs);
    XtManageChild(write_button);

    addsub_button = XmCreatePushButton(form, "addsub_button", args, nargs);
    XtManageChild(addsub_button);

    mult_button = XmCreatePushButton(form, "mult_button", args, nargs);
    XtManageChild(mult_button);

    set_button = XmCreatePushButton(form, "set_button", args, nargs);
    XtManageChild(set_button);

    definc_button = XmCreatePushButton(form, "definc_button", args, nargs);
    XtManageChild(definc_button);

    help_button = XmCreatePushButton(form, "help_button", args, nargs);
    XtManageChild(help_button);

    done_button = XmCreatePushButton(form, "done_button", args, nargs);
    XtManageChild(done_button);

    XtManageChild(form);

    /*************/
    /*           */
    /* Callbacks */
    /*           */
    /*************/

    /* main shell */
    XtAddCallback(read_button, XmNactivateCallback, main_read_button_cb, 
	(Widget) shell);
    XtAddCallback(write_button, XmNactivateCallback, main_write_button_cb, 
	(Widget) shell);
    XtAddCallback(addsub_button, XmNactivateCallback, main_addsub_button_cb, 
	(Widget) shell);
    XtAddCallback(mult_button, XmNactivateCallback, main_mult_button_cb, 
	(Widget) shell);
    XtAddCallback(set_button, XmNactivateCallback, main_set_button_cb, 
	(Widget) shell);
    XtAddCallback(definc_button, XmNactivateCallback, main_definc_button_cb, 
	(Widget) shell);
    XtAddCallback(help_button, XmNactivateCallback, shared_temp_help_button_cb, 
	NULL);
    XtAddCallback(done_button, XmNactivateCallback, main_shared_button_cb, 
	NULL);

    XtAddCallback(XmMessageBoxGetChild(Error_dialog, XmDIALOG_OK_BUTTON), 
	XmNactivateCallback, dialog_ok_button_cb, (Widget) Error_dialog);
    XtAddCallback(XmMessageBoxGetChild(Message_dialog, XmDIALOG_OK_BUTTON), 
	XmNactivateCallback, dialog_ok_button_cb, (Widget) Message_dialog);

    /* let's get this thing going */

    XtRealizeWidget(shell);
    XtAppMainLoop(ac);

} /* end main() */

/****************************************************************
*
*
****************************************************************/

static void parse_command_line_args(argc, argv, defines, includes)
int argc;
char **argv;
char *defines;
char *includes;
{

extern char *optarg;
extern int optind, opterr;
int c;
BOOL error;

    /* default values */
    *defines = '\0';
    *includes = '\0';

    error = FALSE;
    while (!error && ((c = getopt(argc, argv, "D:I:")) != -1))
    {
	switch (c)
	{
	    case 'D':
		/* -D option */
		if ((strlen(defines)+strlen(argv[optind-1])+1) < CMDSIZE)
		{
		    strcat(defines, argv[optind-1]);
		    strcat(defines, " ");
		}
		else
		    /* too many -D specifications */
		    error = TRUE;
		break;
	    case 'I':
		/* -I option */
		if ((strlen(includes)+strlen(argv[optind-1])+1) < CMDSIZE)
		{
		    strcat(includes, argv[optind-1]);
		    strcat(includes, " ");
		}
		else
		    /* too many -I specifications */
		    error = TRUE;
		break;
	    default: error = TRUE; break;
	} /* end switch() */
    } /* endwhile */

    if (error)
    {
printf("\n\tusage: %s {-Dname{=def}} {-Ipathname}\n\n", argv[0]);

printf("where\n\n");

printf("\t-Dname{=def} - Defines a symbol name to the C preprocessor.\n");
printf("\t\tEquivalent to a #define directive.  If no def is given, name\n");
printf("\t\tis defined as `1'.  The maximum length that can be specified\n");
printf("\t\tin this manner is %d.\n\n", CMDSIZE);

printf("\t-Ipathname - Add pathname to the list of directories in which\n");
printf("\t\tto search for #include files with relative filenames (not\n");
printf("\t\tbeginning with slash `/').  The preprocessor first searches\n");
printf("\t\tfor #include files in the directory containing the file,\n");
printf("\t\tthen in the directories named with -I option (if any).\n");
printf("\t\tThe maximumn length that can be specified in this manner\n");
printf("\t\tis %d.\n\n", CMDSIZE);

	exit(1);
    } /* endif */

} /* end parse_command_line_args() */

/****************************************************************
*
*
****************************************************************/

static void main_shared_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

Widget my_shell;

/* printf("entering main_shared_button_cb() >%s<\n", (client_data ? XtName((Widget) client_data) : "nowidget")); */
    if (my_shell = (Widget) client_data)
	XtMapWidget(my_shell);
    else
    {
	/* Done was pressed */
	cleanup_read();
	cleanup_write();
	cleanup_addsub();
	cleanup_mult();
	cleanup_set();

	exit(0);
    } /* endif */

} /* end main_shared_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void dialog_ok_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

    /* a dialog OK button pressed ... making dialog disappear */
    XtUnmanageChild((Widget) client_data);

} /* end dialog_ok_button_cb() */

/****************************************************************
*
*
****************************************************************/

void shared_temp_help_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

    popup_message("Help not implemented yet.");

} /* end shared_temp_help_button_cb() */

