/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#include <sys/types.h> /* for stat() */
#include <sys/stat.h>  /* for stat() */
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

#include "burtgooey.h"
#include "burtgooeyfuncdecls.h"

/*********************/
/*                   */
/* Utility Functions */
/*                   */
/*********************/

/****************************************************************
*
* it is assumed that filename is a character string at 
* least MAXPATHLEN (from <sys/param.h> big.
*
****************************************************************/

BOOL BurtCreateUniqueFilename(prefix, suffix, filename)
char *prefix;
char *suffix;
char *filename;
{

time_t curr_time; 
struct tm *pnow_tm;
int attempt;
unsigned nbytes;
struct passwd *pp;
BOOL rc;

    if (!filename)
    {
	fprintf(Log_fp, 
	    "ERROR: BurtCreateUniqueFilename() got NULL filename\n");
	fflush(Log_fp);
	exit(1);
    } /* endif */

    if (pp = getpwuid(geteuid()))
    {
	time(&curr_time);

	pnow_tm = localtime(&curr_time);

	nbytes = strlen((prefix ? prefix : ""));
	nbytes = nbytes + strlen(pp->pw_name) + 1;
	nbytes = nbytes + (2*3) + 1 + (2*3) + 1 + 1;
	nbytes = strlen((suffix ? suffix : ""));
	nbytes ++;

	if (nbytes > MAXPATHLEN)
	{
	    fprintf(Log_fp, 
		"ERROR: BurtCreateUniqueFilename() prefix >%s< (%d) suffix >%s< (%d) cannot create filename that large ... maxlen %d\n", 
		(prefix ? prefix : ""), strlen((prefix ? prefix : "")), 
		(suffix ? suffix : ""), strlen((suffix ? suffix : "")), 
		MAXPATHLEN);
	    fflush(Log_fp);
	    exit(1);
	} /* endif */

	for (rc = FALSE, attempt = 0; !rc && attempt < 10; attempt ++)
	{
	    sprintf(filename, "%s%s_%02d%02d%02d_%02d%02d%02d_%1d%s", 
		(prefix ? prefix : ""), pp->pw_name, pnow_tm->tm_year, 
		pnow_tm->tm_mon+1, pnow_tm->tm_mday, pnow_tm->tm_hour, 
		pnow_tm->tm_min, pnow_tm->tm_sec, attempt, 
		(suffix ? suffix : ""));

	    rc = !file_exists(filename);
	} /* endfor */

	if (!rc)
	    *filename = '\0';
    }
    else
    {
	fprintf(Log_fp, 
	    "ERROR: BurtCreateUniqueFilename() could not id user\n");
	fflush(Log_fp);
	exit(1);
    } /* endif */

    return rc;

} /* end BurtCreateUniqueFilename() */

/****************************************************************
*
*
****************************************************************/

void clear_scrolled_text(scrolled_text)
Widget scrolled_text;
{

    /* clearing logfile's scrolled text */
    XtVaSetValues(scrolled_text,
	XmNvalue, "",
	NULL);
    /* need this or the scrolled txt is not cleared */
    XFlush(XtDisplay(scrolled_text));

} /* end clear_scrolled_text() */

/****************************************************************
*
*
****************************************************************/

Pixel color_string_to_pixel(w, color_name)
Widget w;
char *color_name;
{

XColor  color;
Colormap cmap;

    if (color_name)
    {
	cmap = DefaultColormap(XtDisplay(w),DefaultScreen(XtDisplay(w)));
	color.pixel = 0;
	XParseColor(XtDisplay(w), cmap, color_name, &color);
	XAllocColor(XtDisplay(w), cmap, &color);
    }
    else
    {
	fprintf(Log_fp, "ERROR: color_string_to_pixel() got NULL color_name\n");
	exit(1);
    } /* endif */

    return(color.pixel);

} /* end color_string_to_pixel() */

/****************************************************************
*
*
****************************************************************/
 
BOOL file_exists(fname)
char *fname;
{
 
struct stat buff;
BOOL rc;
 
    if (fname)
        if (*fname)
            rc = !stat(fname, &buff);
        else
            rc = FALSE;
    else
        rc = FALSE;
 
    return rc;
 
} /* end file_exists() */
 
/****************************************************************
*
*
****************************************************************/

void file_to_scrolled_text(fname, scrolled_text)
char *fname;
Widget scrolled_text;
{

struct stat stat_buff;
int fd;
char *logbuff;
int buffsize;

    /* dumping logfile into scrolled text */
    if (fname)
    {
	if ((fd = open(fname, O_RDONLY)) != -1)
	{
	    if (!fstat(fd, &stat_buff))
	    {
		/* adding one for terminating null char */
		buffsize = stat_buff.st_size + 1;

		if (logbuff = (char *) calloc (buffsize, 1))
		{
		    if (read(fd, logbuff, stat_buff.st_size) != -1)
		    {
			/* although buffer was allocated with calloc(),   */
			/* which presumably allocates a block with all    */
			/* zeros, still need to explicitly place the null */
			/* char at the end so as not to get extra garbage */
			/* characters in the scroll text ... can't        */
			/* explain why this works just happy that it      */
			/* does.                                          */

			*(logbuff+stat_buff.st_size) = '\0';

			XtVaSetValues(scrolled_text,
				XmNvalue, logbuff,
				NULL);
		    }
		    else
			XtVaSetValues(scrolled_text,
			    XmNvalue, "ERROR: read(). Cannot view log file.",
				NULL);

		    free(logbuff);
		}
		else
		{
		    XtVaSetValues(scrolled_text,
			XmNvalue, "ERROR: calloc(). Cannot view log file.",
			NULL);
		} /* endif */
	    }
	    else
	    {
		XtVaSetValues(scrolled_text,
		    XmNvalue, "ERROR: fstat(). Cannot view log file.",
		    NULL);
	    } /* endif */
	    close(fd);
	}
	else
	{
	    XtVaSetValues(scrolled_text,
		XmNvalue, "ERROR: open(). Cannot view log file.",
		NULL);
	} /* endif */
    }
    else
    {
	XtVaSetValues(scrolled_text,
	    XmNvalue, "SYS ERR:: file_to_scrolled_text() got NULL fname.",
	    NULL);
    } /* endif */

} /* end file_to_scrolled_text() */

/****************************************************************
*
*
****************************************************************/

void popup_error(cp)
char *cp;
{


XmString xs;

    if (cp && *cp)
    {
	xs = XmStringCreateSimple(cp);
	XtVaSetValues(Error_dialog, XmNmessageString, xs, NULL);
	XmStringFree(xs);

	XtManageChild(Error_dialog);
    } /* endif */

} /* end popup_error() */

/****************************************************************
*
*
****************************************************************/

void popup_message(cp)
char *cp;
{


XmString xs;

    if (cp && *cp)
    {
	xs = XmStringCreateSimple(cp);
	XtVaSetValues(Message_dialog, XmNmessageString, xs, NULL);
	XmStringFree(xs);

	XtManageChild(Message_dialog);
    } /* endif */

} /* end popup_message() */

/****************************************************************
*
* Allows only digits 0-9 (isdigit()) in the TextField
*
* from page 314 AT&T Motif Course Manual
*
****************************************************************/

void verify_int_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

XmTextVerifyCallbackStruct *cbs;
int i;

    cbs = (XmTextVerifyCallbackStruct *) call_data;

    if (cbs->text->ptr)
	/* something has been added */
	for (i = 0; i < cbs->text->length; i ++)
	    if (!isdigit(cbs->text->ptr[i]))
		cbs->doit = False; /* disallow the whole thing */

} /* end verify_int_cb() */

/****************************************************************
*
*
****************************************************************/

void print_file(fname)
char *fname;
{

char command[CMDSIZE];
int i;
int rc;
char *buff;
int fd;
struct stat stat_buff;
BOOL all_ascii;
char mbuff[1000];


    if (*fname)
    {
	if ((fd = open(fname, O_RDONLY)) != -1)
	{
	    if (!fstat(fd, &stat_buff))
	    {
		if (S_ISREG(stat_buff.st_mode))
		{
		    if (stat_buff.st_size)
		    {
			if (buff = (char *) calloc (stat_buff.st_size,1))
			{
			    if (read(fd, buff, stat_buff.st_size) != -1)
			    {
				/* loop to check file is printable */
				for (i = 0, all_ascii = TRUE; 
				    all_ascii && i < stat_buff.st_size; 
					i ++)
					    all_ascii = isascii(buff[i]);

				if (all_ascii)
				{
				    sprintf(command, "lpr %s", fname);
				    rc = system(command);

				    if (WIFEXITED(rc))
				    {
					if (WEXITSTATUS(rc))
					{
					    /* something wrong */
					    sprintf(mbuff, 
				    "SYS FAULT: error detected in print of %s.",
						fname);
					    popup_message(mbuff);
					} /* endif */
				    }
				    else
				    {
					sprintf(mbuff, 
			"SYS FAULT: print %s command returned but not exited.",
					    fname);
					popup_message(mbuff);
				    } /* endif */
				}
				else
				{
				    sprintf(mbuff, 
					"ERROR: file %s not printable.", fname);
				    popup_error(mbuff);
				} /* endif */
			    }
			    else
			    {

				sprintf(mbuff, 
				"SYS FAULT: could not read() file %s.", fname);
				popup_message(mbuff);
			    } /* endif */

			    free(buff);
			}
			else
			{
			sprintf(mbuff, "SYS FAULT: failed calloc() %d bytes.", 
				stat_buff.st_size);
			    popup_message(mbuff);
			} /* endif */
		    }
		    else
		    {
			sprintf(mbuff, "%s empty ... not printed.", fname);
			popup_message(mbuff);
		    } /* endif */
		}
		else
		{
		    sprintf(mbuff, 
		"NOTE: cannot print file %s because it is not a regular file.", 
			fname);
		popup_message(mbuff);
		} /* endif */
	    }
	    else
		popup_message("SYS FAULT: error in fstat().");

	    close(fd);
	}
	else
	{
	    sprintf(mbuff, "SYS FAULT: cannot open() file %s.", fname);
	    popup_message(mbuff);
	} /* endif */
    } /* endif */

} /* end print_file() */

/****************************************************************
*
*
****************************************************************/

void set_not_ok(status_label)
Widget status_label;
{

XmString xs;

    XmChangeColor(status_label, Red_pixel);
    XFlush(XtDisplay(status_label));

    xs = XmStringCreateSimple(NOTOKSTRING);
    XtVaSetValues(status_label,
	XmNlabelString, xs,
	NULL);
    XmStringFree(xs);

    XFlush(XtDisplay(status_label));

    XSync(XtDisplay(status_label), False);

} /* end set_running() */

/****************************************************************
*
*
****************************************************************/

void set_ok(status_label)
Widget status_label;
{

XmString xs;

    XmChangeColor(status_label, Green_pixel);
    XFlush(XtDisplay(status_label));

    xs = XmStringCreateSimple(OKSTRING);
    XtVaSetValues(status_label,
	XmNlabelString, xs,
	NULL);
    XmStringFree(xs);

    XFlush(XtDisplay(status_label));

    XSync(XtDisplay(status_label), False);

} /* end set_running() */

/****************************************************************
*
*
****************************************************************/

void set_running(status_label)
Widget status_label;
{

XmString xs;

    XmChangeColor(status_label, Blue_pixel);
    XFlush(XtDisplay(status_label));

    xs = XmStringCreateSimple(RUNNINGSTRING);
    XtVaSetValues(status_label,
	XmNlabelString, xs,
	NULL);
    XmStringFree(xs);

    XFlush(XtDisplay(status_label));

    XSync(XtDisplay(status_label), False);

} /* end set_running() */

/********************/
/*                  */
/* Shared Callbacks */
/*                  */
/********************/

/****************************************************************
*
*
****************************************************************/

void shared_done_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

    XtUnmapWidget((Widget) client_data);

} /* end shared_done_button_cb() */

/****************************************************************
*
*
****************************************************************/

void shared_print_log_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{
    print_file((char *) client_data);
} /* end shared_print_log_button_cb() */

