/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#ifndef __burtlist__
#define __burtlist__

#include <stdio.h>

#include "SDDS.h"

#include "burtobjs.h"

#ifndef BOOL
#define BOOL char
#endif
#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE 1
#endif

/***************/
/*             */
/* Burtobjlist */
/*             */
/***************/

class Burtobj;
class SnapParms;
class Filename;

class Burtobjlist : public List<Burtobj>
{

public:
    Burtobjlist();

    /* start inline methods */
    SnapParms *get_parms() const; 
    List<Filename> *const look_at_request_filenames() const;
    List<Filename> *get_request_filenames();
    /* end inline methods */

    Burtobj *find(Burtobj *const) const;
    Burtobj *find_name(const char *const) const;
    Burtobj *find_name_nreq(const char *const, const int) const;

    BOOL read_snapshot_from_iocs(const int);
    BOOL read_snapshot_from_iocs(const int, char *const, char *const);

    void write_req_to_file() const;

    void write_snapshot_to_file() const;

    BOOL write_snapshot_to_iocs(const int, const char, const char);

    void add_request_filename(const char *const);

private:
    BOOL issue_reads();
    void search_for_read_pvars(const int);
    void search_for_write_pvars(const int);

    BOOL wait_for_reads(const int) const;
    BOOL write_snapshot_to_iocs_absolute(const int);
    BOOL write_snapshot_to_iocs_relative(const int) const;
    void write_header() const;
    /* must be a pointer here and allocated in Snaplist() because */
    /* list.h CANNOT include burtobjs.h. it cannont because burtobjs.h */
    /* needs to include list.h */
    SnapParms *parms;
    List<Filename> request_filenames;

}; /* end class Burtobjlist */

inline SnapParms *Burtobjlist::get_parms() const { return parms; }
inline List<Filename> *const Burtobjlist::look_at_request_filenames() const { return (List<Filename> *const) &request_filenames; }
inline List<Filename> *Burtobjlist::get_request_filenames() { return &request_filenames; }

/*******************/
/*                 */
/* Burtobjlistlist */
/*                 */
/*******************/

class Burtobjlist;

class Burtobjlistlist : public List<Burtobjlist>
{

public:
    Burtobj *find_name(const char *);

}; /* end class Burtobjlistlist */

/*****************/
/*               */
/* Predicatelist */
/*               */
/*****************/

class Predicate;

class Predicatelist : public List<Predicate>
{

public:
    Predicatelist();
    Burtobjlist *get_burtobjs();

private:
    Burtobjlist burtobjs;

}; /* end class Predicatelist */

inline Burtobjlist *Predicatelist::get_burtobjs() { return &burtobjs; }
inline Predicatelist::Predicatelist() { get_burtobjs()->set_view(); }

#endif
