/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* for strncpy() */
#include <pwd.h>   /* for getpwuid() */

/* for time() */
#include <sys/types.h>
#include <sys/time.h>

#include <time.h> /* for ctime() */

#include <unistd.h>    /* for geteuid() */
#include <db_access.h> /* for dbr_size_n() */
#include <ctype.h>     /* for isspace() */

#include "burtcommon.h"
#include "burtobjs.h"

#include "list.h"
/*#include "cursor.h"*/
#include "burtlist.h"

#include "SDDS.h"

/****************************************/
/*                                      */
/* Externally Visibile Remote Functions */
/*                                      */
/****************************************/

/* in burtcommon.cc */
BOOL BurtSDDSColExists(SDDS_TABLE *const, char *const, const long);
void BurtExit(const int);

extern "C" {
/* Channel Access Interface Functions */
BOOL BurtCaArrayPut(Channel *);
} /* end extern c thing */

/************/
/*          */
/* Filename */
/*          */
/************/

Filename::Filename(const char *const d, const char *const f)
{
    if (d) 
    {
	strncpy(directory, d, sizeof(directory)-1);
	directory[sizeof(directory)-1] = '\0';
    }
    else
	*directory = '\0';

    if (f)
    {
	strncpy(filename, f, sizeof(filename)-1);
	filename[sizeof(filename)-1] = '\0';
    } 
    else
	*filename = '\0';
}

Filename::Filename(const char *const f)
{

    directory[0] = '\0';

    if (f)
    {
	strncpy(filename, f, sizeof(filename)-1);
	filename[sizeof(filename)-1] = '\0';
    } 
    else
	*filename = '\0';
}

/*********/
/*       */
/* Value */
/*       */
/*********/

/**********/
/* Public */
/**********/

/****************************************************************
*
*
****************************************************************/

Value::Value()
{
    set_dbr_type(INVALID_DBR);
    set_pval((void *) NULL);
    set_nelem(UNKNOWN);
}

/****************************************************************
*
*
****************************************************************/

void Value::clear_value() 
{
    set_dbr_type(INVALID_DBR);

    free((char *) get_pval());
    set_pval((void *) NULL);

    set_nelem(UNKNOWN);

} 

/****************************************************************
*
*
****************************************************************/

BOOL Value::defined() const
{

BOOL rc;

    rc = (get_dbr_type() != INVALID_DBR 
	    && get_pval() 
		&& get_nelem() != UNKNOWN);
    
    return rc;

}

/****************************************************************
*
*
****************************************************************/

void Value::set(const char dbr_type, const int nelem, const void *const pval)
{

    if (VALID_ATOMIC_DBR_REQ(dbr_type) && nelem > 0 && pval)
    {
	clear_value();

	/* dbr_size_n() macro in db_access.h */
	set_pval((void *) malloc(dbr_size_n(dbr_type, nelem)));

	if (get_pval())
	{
	    /* this simple memcpy() works because we are assured that */
	    /* we never requested any DBR_STS_xxx types.  if we did   */
	    /* then we would have to copy from                        */
	    /* ((struct dbr_sts_xxx *) arg.dbr)->value with an ugly   */
	    /* switch so as to cast the correct struct dbr_sts_xxx.   */

#if 1
        if (dbr_type==DBR_STRING && nelem==1)
        {
	        strcpy((char*)get_pval(), (const char*)pval);
        }else {
#endif
	        memcpy(get_pval(), pval, dbr_size_n(dbr_type, nelem));
#if 1
        }
#endif
	    set_dbr_type(dbr_type);
	    set_nelem(nelem);
	}
	else
	{
	    fprintf(Log_fp, "ERROR:  Value::set() failed malloc %d bytes\n",
		dbr_size_n(dbr_type, nelem));
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
    }
    else
    {
	fprintf(Log_fp, 
	    "ERROR: Value::set() got dbr_type %d nelem %d pval %x\n", 
	    dbr_type, nelem, pval);
	BurtExit(1);
    } /* endif */

}

/****************************************************************
*
*
****************************************************************/

void Value::write_val(const char *const name, const BOOL read_only_notify, 
    const BOOL read_only) const
{

int elem;
char *cp;
short *sp;
unsigned short *usp;
unsigned char *ucp;
long *lp;
double *dp;
float *fp;
int needs_quotes;
char *temp_cp;

    if (defined())
    {
	if (name)
	{
	    if (read_only_notify)
		fprintf(Out_fp, "%s ", READONLYNOTIFYSTRING);
	    else if (read_only)
		fprintf(Out_fp, "%s ", READONLYSTRING);
	    fprintf(Out_fp, "%s %d", name, get_nelem());
	    fflush(Out_fp);

	    switch (get_dbr_type())
	    {
		case DBR_STRING:
		    for (elem = 0, cp = (char *) get_pval();
			elem < get_nelem();
			    elem ++, cp += MAX_STRING_SIZE)
		    {
			if (*cp)
			{
			    /* checking for any white space */
			    for (temp_cp = cp, needs_quotes = FALSE;
				!needs_quotes && *temp_cp 
				&& temp_cp < cp + MAX_STRING_SIZE;
				    temp_cp ++)
					needs_quotes = isspace(*temp_cp);

			    if (needs_quotes)
				fprintf(Out_fp, " \"");
			    else
				fprintf(Out_fp, " ");
				
			    fprintf(Out_fp, "%s", cp);

			    if (needs_quotes)
				fprintf(Out_fp, "\"");

			}
			else
			    fprintf(Out_fp, " %s", NULLSTRING);
			fflush(Out_fp);
		    } /* endfor */
		    fprintf(Out_fp, "\n");
		    fflush(Out_fp);
		    break;
		/* case DBR_INT: */
		case DBR_SHORT:
		    for (elem = 0, sp = (short *) get_pval();
			elem < get_nelem();
			    elem ++, sp ++)
		    {
			fprintf(Out_fp, " %d", *sp);
			fflush(Out_fp);
		    } /* endfor */
		    fprintf(Out_fp, "\n");
		    fflush(Out_fp);
		    break;
		case DBR_FLOAT:
		    for (elem = 0, fp = (float *) get_pval();
			elem < get_nelem();
			    elem ++, fp ++)
		    {
			fprintf(Out_fp, " ");
			fprintf(Out_fp, FLOATFORMATSTRING, *fp);
			fflush(Out_fp);
		    } /* endfor */
		    fprintf(Out_fp, "\n");
		    fflush(Out_fp);
		    break;
		case DBR_ENUM:
		    for (elem = 0, usp = (unsigned short *) get_pval();
			elem < get_nelem();
			    elem ++, usp ++)
		    {
			fprintf(Out_fp, " %d", *usp);
			fflush(Out_fp);
		    } /* endfor */
		    fprintf(Out_fp, "\n");
		    fflush(Out_fp);
		    break;
		case DBR_CHAR:
		    for (elem = 0, ucp = (unsigned char *) get_pval();
			elem < get_nelem();
			    elem ++, ucp ++)
		    {
		    if (get_nelem()==1 && *ucp==0 ) fprintf(Out_fp, " 0");
		else	fprintf(Out_fp, " %c", *ucp);
			fflush(Out_fp);
		    } /* endfor */
		    fprintf(Out_fp, "\n");
		    fflush(Out_fp);
		    break;
		case DBR_LONG:
		    for (elem = 0, lp = (long *) get_pval();
			elem < get_nelem();
			    elem ++, lp ++)
		    {
			fprintf(Out_fp, " %ld", *lp);
			fflush(Out_fp);
		    } /* endfor */
		    fprintf(Out_fp, "\n");
		    fflush(Out_fp);
		    break;
		case DBR_DOUBLE:
		    for (elem = 0, dp = (double *) get_pval();
			elem < get_nelem();
			    elem ++, dp ++)
		    {
			fprintf(Out_fp, " ");
			fprintf(Out_fp, DOUBLEFORMATSTRING, *dp);
			fflush(Out_fp);
		    } /* endfor */
		    fprintf(Out_fp, "\n");
		    fflush(Out_fp);
		    break;
		default:
		    fprintf(Log_fp, 
	"ERROR: Value::write_val() encountered unrecognizable dbr_type %d\n", 
			get_dbr_type());
		    BurtExit(1);
		    break;
	    } /* end switch() */
	}
	else
	{
	    fprintf(Log_fp, "ERROR: Value::write_val() got name %x\n", name);
	    BurtExit(1);
	} /* endif */
    } /* endif */

}

/****************************************************************
*
* writes a value to an SDDS table into a specific row.
* 
* converts the value into a string and places it into outbuff before writing.
*
****************************************************************/

void Value::write_val_sdds(SDDS_TABLE *table, const long &row, 
    const char *const name, const char *const type, const char *const lineage, 
    const char *const bmsg, const char *const rmsg, const BOOL read_only_notify,
    const BOOL read_only) const
{

int elem;
char *cp;
short *sp;
unsigned short *usp;
unsigned char *ucp;
long *lp;
double *dp;
float *fp;
char *outbuff, *tempbuff;
unsigned out_nbytes, temp_nbytes;
int needs_quotes;
char *temp_cp;

    if (defined())
    {
	if (table && name && type && lineage && bmsg && rmsg)
	{
	    if (*name && *type && *lineage && *bmsg && *rmsg)
	    {
		/* checking all the columns are there */
		if (!BurtSDDSColExists(table, NAME_COL, SDDS_STRING)
		    || !BurtSDDSColExists(table, TYPE_COL, SDDS_STRING)
		    || !BurtSDDSColExists(table, LINEAGE_COL, SDDS_STRING)
		    || !BurtSDDSColExists(table, READMSG_COL, SDDS_STRING)
		    || !BurtSDDSColExists(table, WRITEMSG_COL, SDDS_STRING)
		    || !BurtSDDSColExists(table, MODE_COL, SDDS_STRING)
		    || !BurtSDDSColExists(table, NELEM_COL, SDDS_LONG)
		    || !BurtSDDSColExists(table, VAL_COL, SDDS_STRING))
		{
		    fprintf(Log_fp, 
"ERROR: Value::write_val_sdds() got table missing columns of proper types.\n");
		    fflush(Log_fp);
		    BurtExit(1);
		} /* endif */
		    
		/* calculating required size for outbuff and tempbuff */
		out_nbytes = 0;
		temp_nbytes = 0;
		switch (get_dbr_type())
		{
		    case DBR_STRING:
			out_nbytes = get_nelem()*(MAX_STRING_SIZE+5)+1;
			temp_nbytes = MAX_STRING_SIZE+5+1;
			break;
		    /* case DBR_INT: */
		    case DBR_SHORT:
			for (elem = 0, sp = (short *) get_pval();
			    elem < get_nelem();
				elem ++, sp ++)
			{
			    out_nbytes = out_nbytes + BYTESFORSTRING(*sp) + 1;
			    if (temp_nbytes < BYTESFORSTRING(*sp))
				temp_nbytes = BYTESFORSTRING(*sp);
			} /* endfor */
			out_nbytes ++;
			temp_nbytes ++;
			break;
		    case DBR_FLOAT:
			out_nbytes = get_nelem()*(FLOATSTRINGLENGTH+1)+1;
			temp_nbytes = (FLOATSTRINGLENGTH+1)+1;
			break;
		    case DBR_ENUM:
			for (elem = 0, usp = (unsigned short *) get_pval();
			    elem < get_nelem();
				elem ++, usp ++)
			{
			    out_nbytes = out_nbytes + BYTESFORSTRING(*usp) + 1;
			    if (temp_nbytes < BYTESFORSTRING(*usp))
				temp_nbytes = BYTESFORSTRING(*usp);
			} /* endfor */
			out_nbytes ++;
			temp_nbytes ++;
			break;
		    case DBR_CHAR:
			out_nbytes = get_nelem()*2+1;
			temp_nbytes = 2+1;
			break;
		    case DBR_LONG:
			for (elem = 0, lp = (long *) get_pval();
			    elem < get_nelem();
				elem ++, lp ++)
			{
			    out_nbytes = out_nbytes + BYTESFORSTRING(*lp) + 1;
			    if (temp_nbytes < BYTESFORSTRING(*lp))
				temp_nbytes = BYTESFORSTRING(*lp);
			} /* endfor */
			out_nbytes ++;
			temp_nbytes ++;
			break;
		    case DBR_DOUBLE:
			out_nbytes = get_nelem()*(DOUBLESTRINGLENGTH+1)+1;
			temp_nbytes = (DOUBLESTRINGLENGTH+1)+1;
			break;
		    default:
			fprintf(Log_fp, 
    "ERROR: Value::write_val_sdds() encountered unrecognizable dbr_type %d\n", 
			    get_dbr_type());
			BurtExit(1);
			break;
		} /* end switch() */

/* printf("Value::write_val_sdds() dbr_type %d nelem %d ---> out_nbytes %d temp_nbytes\n", out_nbytes, temp_nbytes); */

		if (!(outbuff = (char *) malloc(out_nbytes)))
		{
			fprintf(Log_fp, 
"ERROR: Value::write_val_sdds() dbr_type %d nelem %d unable to malloc %d bytes for outbuff\n", 
			    get_dbr_type(), get_nelem(), out_nbytes);
			BurtExit(1);
		} /* endif */

		if (!(tempbuff = (char *) malloc(temp_nbytes)))
		{
			fprintf(Log_fp, 
"ERROR: Value::write_val_sdds() dbr_type %d nelem %d unable to malloc %d bytes for tempbuff\n", 
			    get_dbr_type(), get_nelem(), temp_nbytes);
			BurtExit(1);
		} /* endif */

		/* constructing outbuff from value */
		*outbuff = '\0';
		switch (get_dbr_type())
		{
		    case DBR_STRING:
			for (elem = 0, cp = (char *) get_pval();
			    elem < get_nelem();
				elem ++, cp += MAX_STRING_SIZE)
			{
			    if (*cp)
			    {
				/* checking for any white space */
				for (temp_cp = cp, needs_quotes = FALSE;
				    !needs_quotes && *temp_cp 
				    && temp_cp < cp + MAX_STRING_SIZE;
					temp_cp ++)
					    needs_quotes = isspace(*temp_cp);

				if (needs_quotes) 
				    strcat(outbuff, "\"");
				strcat(outbuff, cp);
				if (needs_quotes) 
				    strcat(outbuff, "\"");

				if (elem < get_nelem() - 1)
				    strcat(outbuff, " ");
			    }
			    else
			    {
				/* empty string */
				strcat(outbuff, NULLSTRING);
				if (elem < get_nelem() - 1)
				    strcat(outbuff, " ");
			    } /* endif */
			} /* endfor */
			break;
		    /* case DBR_INT: */
		    case DBR_SHORT:
			for (elem = 0, sp = (short *) get_pval();
			    elem < get_nelem();
				elem ++, sp ++)
			{
			    sprintf(tempbuff, "%d", *sp);

			    strcat(outbuff, tempbuff);
			    if (elem < get_nelem() - 1)
				strcat(outbuff, " ");
			} /* endfor */
			break;
		    case DBR_FLOAT:
			for (elem = 0, fp = (float *) get_pval();
			    elem < get_nelem();
				elem ++, fp ++)
			{
			    sprintf(tempbuff, FLOATFORMATSTRING, *fp);

			    strcat(outbuff, tempbuff);
			    if (elem < get_nelem() - 1)
				strcat(outbuff, " ");
			} /* endfor */
			break;
		    case DBR_ENUM:
			for (elem = 0, usp = (unsigned short *) get_pval();
			    elem < get_nelem();
				elem ++, usp ++)
			{
			    sprintf(tempbuff, "%d", *usp);

			    strcat(outbuff, tempbuff);
			    if (elem < get_nelem() - 1)
				strcat(outbuff, " ");
			} /* endfor */
			break;
		    case DBR_CHAR:
			for (elem = 0, ucp = (unsigned  char *) get_pval();
			    elem < get_nelem();
				elem ++, ucp ++)
			{
			    sprintf(tempbuff, "%c", *ucp);

			    strcat(outbuff, tempbuff);
			    if (elem < get_nelem() - 1)
				strcat(outbuff, " ");
			} /* endfor */
			break;
		    case DBR_LONG:
			for (elem = 0, lp = (long *) get_pval();
			    elem < get_nelem();
				elem ++, lp ++)
			{
			    sprintf(tempbuff, "%ld", *lp);

			    strcat(outbuff, tempbuff);
			    if (elem < get_nelem() - 1)
				strcat(outbuff, " ");
			} /* endfor */
			break;
		    case DBR_DOUBLE:
			for (elem = 0, dp = (double *) get_pval();
			    elem < get_nelem();
				elem ++, dp ++)
			{
			    sprintf(tempbuff, DOUBLEFORMATSTRING, *dp);

			    strcat(outbuff, tempbuff);
			    if (elem < get_nelem() - 1)
				strcat(outbuff, " ");
			} /* endfor */
			break;
		    default:
			fprintf(Log_fp, 
    "ERROR: Value::write_val_sdds() encountered unrecognizable dbr_type %d\n", 
			    get_dbr_type());
			BurtExit(1);
			break;
		} /* end switch() */

/* printf("about to set row %ld name >%s< type >%s< bkup >%s< rstr >%s< mode >%s< nelem %ld dval %f sval >%s<\n", row, name, type, bmsg, rmsg, (read_only_notify ? READONLYNOTIFYSTRING : (read_only ? READONLYSTRING : DEFAULTSTRING)), (long) get_nelem(), (double) 0, outbuff); */
		if (!SDDS_SetRowValues(table, 
		    (long) (SDDS_SET_BY_NAME | SDDS_PASS_BY_VALUE), row,
			NAME_COL, name,
			TYPE_COL, type,
			LINEAGE_COL, lineage,
			READMSG_COL, bmsg,
			WRITEMSG_COL, rmsg,
			MODE_COL, (read_only_notify ? READONLYNOTIFYSTRING
				    : (read_only ? 
					READONLYSTRING : DEFAULTSTRING)),
			NELEM_COL, (long) get_nelem(),
			VAL_COL, outbuff,
			NULL))
		{
		    fprintf(Log_fp, 
	    "SYS ERR: Value::write_val_sdds() SDDS_SetRowValues()\n");
		    fflush(Log_fp);
		    BurtExit(1);
		} /* endif */

		if (outbuff)
		    free(outbuff);

		if (tempbuff)
		    free(tempbuff);
	    }
	    else
	    {
		fprintf(Log_fp, 
"ERROR: Value::write_val_sdds() got name >%s< type >%s< lineage >%s< bmsg >%s< rmsg >%s<\n",
		    name, type, lineage, bmsg, rmsg);
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */
	}
	else
	{
	    fprintf(Log_fp, 
"ERROR: Value::write_val_sdds() got table %x name %x type %x lineage %x bmsg %x rmsg %x\n",
		table, name, type, lineage, bmsg, rmsg);
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
    } /* endif */

}

/***********/
/* Private */
/***********/

/****************************************************************
*
*
****************************************************************/

void Value::set_dbr_type(const char t)
{

    if (t == INVALID_DBR || VALID_ATOMIC_DBR_REQ(t))
	dbr_type = t;
    else
    {
	fprintf(Log_fp, 
	    "ERROR: Value::set_dbr_type() got invalid type %d\n", t);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

}

/****************************************************************
*
*
****************************************************************/

void Value::set_nelem(const int n)
{
    if (n == UNKNOWN || n > 0)
	nelem = n;
    else
    {
	fprintf(Log_fp, "ERROR: Value::set_nelem() got invalid n=%d\n", n);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

}

/***********/
/*         */
/* Channel */
/*         */
/***********/

/****************************************************************
*
*
****************************************************************/

Channel::Channel()
{

    *name = '\0';

    clear_ever_searched();

    set_nreq(UNKNOWN);
    set_dbr_req_type(INVALID_DBR);

    clear_read_pending();
    clear_reported();

    set_val((Value *) NULL);

}

/****************************************************************
*
*
****************************************************************/

BOOL Channel::connected() const
{ 

BOOL rc;

    if (ever_searched())
	rc = (ca_field_type(cid) != TYPENOTCONN);
    else
	rc = FALSE;

    return rc;
}

/****************************************************************
*
*
****************************************************************/

int Channel::channel_nelem() const
{

int rc;

    if (connected())
	rc =  ca_element_count(cid);
    else
	rc = 0;

    return rc;

}

/****************************************************************
*
*
****************************************************************/

char Channel::channel_type() const
{

char rc;

    if (connected())
	rc =  ca_field_type(cid);
    else
	rc = INVALID_DBR;

    return rc;

}

/****************************************************************
*
*
****************************************************************/

void Channel::set_name(const char *const cp)
{

    if (cp)
    {
	strncpy(name, cp, sizeof(name)-1);
	name[sizeof(name)-1] = '\0';
    } /* endif */

}

/****************************************************************
*
*
****************************************************************/

BOOL Channel::write_to_ioc()
{

BOOL rc;

    if (!(connected()))
    {
	rc = FALSE;

	fprintf(Log_fp, "channel >%s< not connected, no ca_array_put()\n",
	    get_name());
	fflush(Log_fp);
    }
    else if (!(get_val()->defined()))
    {
	rc = FALSE;

	fprintf(Log_fp, 
"ERROR: Channel::write_to_ioc() channel >%s< not defined, no ca_array_put()\n",
	    get_name());
	fflush(Log_fp);
	BurtExit(1);
    }
    else
	rc = BurtCaArrayPut(this);

    return rc;

}

/****************************************************************
*
*
****************************************************************/

void Channel::set_dbr_req_type(const char t) 
{

    if (t == INVALID_DBR || t == NATIVE || VALID_ATOMIC_DBR_REQ(t))
	dbr_req_type = t;
    else
    {
	fprintf(Log_fp, 
	    "ERROR: Channel::set_dbr_req_type() got invalid type %d\n", t);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

}

/****************************************************************
*
*
****************************************************************/

void Channel::set_nreq(const int n)
{
    if (n == UNKNOWN || n > 0)
	nreq = n;
    else
    {
	fprintf(stderr, "ERROR: Channel::set_nreq() got invalid n=%d\n", n);
	BurtExit(1);
    } /* endif */

}

/***********/
/*         */
/* Iocpvar */
/*         */
/***********/

/****************************************************************
*
*
****************************************************************/

Iocpvar::Iocpvar()
{
    set_start((char *) NULL);
    set_end((char *) NULL);
}

/****************************************************************
*
*
****************************************************************/

void Iocpvar::print() const
{
    printf("IOC pvar start %x end %x pvar >%s<\n", 
	get_start(), get_end(), (get_burtpvar()->get_channel())->get_name());
}

/**********/
/*        */
/* Sspvar */
/*        */
/**********/

/****************************************************************
*
*
****************************************************************/

Sspvar::Sspvar()
{

    set_start((char *) NULL);
    set_end((char *) NULL);
    name[0] = '\0';
    set_val((Value *) NULL);
    set_print_type(SS_NOSTRING);

}

/****************************************************************
*
*
****************************************************************/

void Sspvar::set_print_type(const char t)
{

    if (t == SS_STRING || t == SS_NOSTRING)
	print_type = t;
    else
    {
	fprintf(Log_fp, 
	    "ERROR: Sspvar::set_print_type() got invalid type %d\n", t);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

}

/****************************************************************
*
*
****************************************************************/

void Sspvar::set_name(const char *const cp)
{

    if (cp)
    {
	strncpy(name, cp, sizeof(name)-1);
	name[sizeof(name)-1] = '\0';
    } /* endif */

}

/****************************************************************
*
*
****************************************************************/

void Sspvar::print() const
{
    printf("Snapshot pvar start %x end %x printtype %d pvar >%s<\n", 
	get_start(), get_end(), get_print_type(), get_name());
}

/*************/
/*           */
/* SnapParms */
/*           */
/*************/

/****************************************************************
*
*
****************************************************************/

SnapParms::SnapParms()
{
    set_effective_uid((uid_t) -1);
    set_group_id((gid_t) -1);
    set_type((char) UNKNOWN);

    /* bypassing the set_XXX() methods here because the first thing */
    /* they do is free() these values ... we cannot be sure what    */
    /* garbage might be in these variables when they are malloc()'d */
    /* to us.                                                       */
    currtime = (char *) NULL;
    loginid = (char *) NULL;
    real_world_name = (char *) NULL;
    comments = (char *) NULL;
    keywords = (char *) NULL;
}

/****************************************************************
*
*
****************************************************************/

SnapParms::~SnapParms()
{
    free(get_currtime());
    free(get_loginid());
    free(get_real_world_name());
    free(get_comments());
    free(get_keywords());
}

/****************************************************************
*
*
****************************************************************/

void SnapParms::set_who_and_time_stamp()
{

int effective_uid;
time_t now;
struct passwd *pp;

    /* uid - person who logged in ... effective uid person as */
    /* defined by "set-user-ID" (if done at all)              */

    effective_uid = geteuid();

    if (time(&now) != (time_t) -1)
    {
        if (pp = getpwuid(effective_uid))
        {
            set_currtime(ctime(&now));
            set_loginid(pp->pw_name);
            set_real_world_name(pp->pw_gecos);
            set_effective_uid(pp->pw_uid);
            set_group_id(pp->pw_gid);
        }
        else
        {
            fprintf(Log_fp,
	    "ERROR: SnapParms::set_who_and_time_stamp() could not id user\n");
            fflush(Log_fp);
            BurtExit(1);
        } /* endif */
    }
    else
    {
        fprintf(Log_fp,
    "ERROR: SnapParms::set_who_and_time_stamp() could not get time of day\n");
        fflush(Log_fp);
        BurtExit(1);
    } /* endif */

}

/****************************************************************
*
*
****************************************************************/

void SnapParms::set_currtime(const char *const c) 
{

    free(get_currtime()); 

    if (c)
	currtime = strdup(c);
    else
	currtime = (char *) NULL;

}

/****************************************************************
*
*
****************************************************************/

void SnapParms::set_loginid(const char *const c) 
{

    free(get_loginid()); 

    if (c)
	loginid = strdup(c);
    else
	loginid = (char *) NULL;

}

/****************************************************************
*
*
****************************************************************/

void SnapParms::set_real_world_name(const char *const c) 
{

    free(get_real_world_name()); 

    if (c)
	real_world_name = strdup(c);
    else
	real_world_name = (char *) NULL;

}

/****************************************************************
*
*
****************************************************************/

void SnapParms::set_comments(const char *const c) 
{

    free(get_comments()); 

    if (c)
	comments = strdup(c);
    else
	comments = (char *) NULL;

}

/****************************************************************
*
*
****************************************************************/

void SnapParms::set_keywords(const char *const c) 
{

    free(get_keywords()); 

    if (c)
	keywords = strdup(c);
    else
	keywords = (char *) NULL;

}

/****************************************************************
*
*
****************************************************************/

void SnapParms::set_type(const char c)
{

    if (c == UNKNOWN || c == ABSOLUTE || c == RELATIVE || c == NOWRITE)
	type = c;
    else
    {
	fprintf(Log_fp, 
	    "ERROR: SnapParms::set_type() got invalid type %d\n", c);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

}

/****************************************************************
*
*
****************************************************************/

void SnapParms::clear_snapparms()
{

    free(get_currtime());
    set_currtime((char *) NULL);

    free(get_loginid());
    set_loginid((char *) NULL);

    free(get_real_world_name());
    set_real_world_name((char *) NULL);

    set_effective_uid((uid_t) -1);
    set_group_id((gid_t) -1);

    free(get_comments());
    set_comments((char *) NULL);

    free(get_keywords());
    set_keywords((char *) NULL);

    set_type((char) UNKNOWN);

}

/****************************************************************
*
*
****************************************************************/

BOOL SnapParms::defined() const
{

BOOL rc;

    rc = (get_currtime() 
	    && get_loginid()
	    && get_real_world_name()
	    && get_group_id() != (gid_t) -1
	    && get_comments()
	    && get_keywords()
	    && get_type() != UNKNOWN);

    return rc;

}

/***********/
/*         */
/* Burtobj */
/*         */
/***********/

void Burtobj::write_sdds(SDDS_TABLE *table, const long &row) const
{

    if (table)
    {
	/* checking all the columns are there */
	if (!BurtSDDSColExists(table, NAME_COL, SDDS_STRING)
	    || !BurtSDDSColExists(table, TYPE_COL, SDDS_STRING)
	    || !BurtSDDSColExists(table, READMSG_COL, SDDS_STRING)
	    || !BurtSDDSColExists(table, WRITEMSG_COL, SDDS_STRING)
	    || !BurtSDDSColExists(table, NELEM_COL, SDDS_LONG)
	    || !BurtSDDSColExists(table, MODE_COL, SDDS_STRING))
	{
	    fprintf(Log_fp, 
"ERROR: Value::write_val() got table missing columns of proper types.\n");
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
	    
	if (!SDDS_SetRowValues(table, 
	    (long) (SDDS_SET_BY_NAME | SDDS_PASS_BY_VALUE), row,
		NAME_COL, get_name(),
		TYPE_COL, (isa() == PVAR ? PVSTRING : DEVSTRING),
		READMSG_COL, get_readmsg(),
		WRITEMSG_COL, get_writemsg(),
		NELEM_COL, (get_val_RO()->get_nelem() == UNKNOWN 
				? 0 : (long) get_val_RO()->get_nelem()),
		MODE_COL, (read_only_notify() ? READONLYNOTIFYSTRING
			    : (read_only() ? 
				READONLYSTRING : DEFAULTSTRING)),
		NULL))
	{
	    fprintf(Log_fp, 
    "SYS ERR: Burtobj::write_sdds() SDDS_SetRowValues()\n");
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
    }
    else
    {
	fprintf(Log_fp, "ERROR: Burtobj::write_sdds() got NULL table\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

}

/***************/
/*             */
/* Burtatomdev */
/*             */
/***************/

/****************************************************************
*
*
****************************************************************/

Burtatomdev::Burtatomdev() 
{
    *name = '\0'; 
    *lineage = '\0';
    *readmsg = '\0';
    *writemsg = '\0';
}

/****************************************************************
*
*
****************************************************************/

void Burtatomdev::set_name(const char *const cp)
{

    if (cp)
    {
	strncpy(name, cp, sizeof(name)-1);
	name[sizeof(name)-1] = '\0';
    } /* endif */

}

/****************************************************************
*
*
****************************************************************/

void Burtatomdev::set_lineage(const char *const cp)
{

    if (cp)
    {
	strncpy(lineage, cp, sizeof(lineage)-1);
	lineage[sizeof(lineage)-1] = '\0';
    } /* endif */

}

/****************************************************************
*
*
****************************************************************/

void Burtatomdev::set_readmsg(const char *const cp)
{

    if (cp)
    {
	strncpy(readmsg, cp, sizeof(readmsg)-1);
	readmsg[sizeof(readmsg)-1] = '\0';
    } /* endif */

}

/****************************************************************
*
*
****************************************************************/

void Burtatomdev::set_writemsg(const char *const cp)
{

    if (cp)
    {
	strncpy(writemsg, cp, sizeof(writemsg)-1);
	writemsg[sizeof(writemsg)-1] = '\0';
    } /* endif */

}

/*************/
/*           */
/* Predicate */
/*           */
/*************/

/****************************************************************
*
*
****************************************************************/

Predicate::Predicate() 
{ 
    get_ioc_pvars_list()->set_owner(); 
    get_ss_pvars_list()->set_owner(); 

    set_directive_on_fail((char) UNKNOWN);

    raw_input_line = (char *) NULL;
    input_line = (char *) NULL;

    clear_printed();
}

/****************************************************************
*
*
****************************************************************/

Predicate::~Predicate()
{
    free(get_raw_input_line());
    free(get_input_line());
}

/****************************************************************
*
*
****************************************************************/

void Predicate::set_raw_input_line(const char *const c) 
{

    free(get_raw_input_line()); 

    if (c)
	raw_input_line = strdup(c);
    else
	raw_input_line = (char *) NULL;

}

/****************************************************************
*
*
****************************************************************/

void Predicate::set_input_line(const char *const c) 
{

    free(get_input_line()); 

    if (c)
	input_line = strdup(c);
    else
	input_line = (char *) NULL;

}

/****************************************************************
*
*
****************************************************************/

void Predicate::set_directive_on_fail(const char c)
{

    if (c == UNKNOWN || c == QUERYONFAIL || c == HALTONFAIL || c == CONTONFAIL)
	directive_on_fail = c;
    else
    {
	fprintf(Log_fp, 
	"ERROR: Predicate::set_directive_on_fail() got invalid directive %d\n", 
	    c);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

}

/****************************************************************
*
*
****************************************************************/

void Predicate::print() const
{

Cursor<Iocpvar> *icp;
Iocpvar *ip;
Cursor<Sspvar> *scp;
Sspvar *sp;

    printf("input line %x start %x dironfail %d >%s<\n", 
	get_input_line(), get_start(), get_directive_on_fail(), 
	get_input_line());

    printf("-- IOC Pvars\n");
    icp = new Cursor<Iocpvar>(get_ioc_pvars_list());
    for (ip = icp->getfirst(); ip; ip = icp->getnext())
	ip->print();
    delete icp;

    printf("-- Snapshot Pvars\n");
    scp = new Cursor<Sspvar>(get_ss_pvars_list());
    for (sp = scp->getfirst(); sp; sp = scp->getnext())
	sp->print();
    delete scp;

}

