/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#define MAINLINE

#include "burtcommon.h"
#include "burtobjs.h"
#include "burtset.h"

#include "list.h"
/*#include "cursor.h"*/
#include "burtlist.h"

/****************************************/
/*                                      */
/* Externally Visibile Remote Functions */
/*                                      */
/****************************************/

/* in burtcommon.cc */
void BurtCommonInit(const BOOL);
BOOL BurtPreprocessFile(const char *const, const char **const,
    const char **const, char *const);
void BurtProcessRequestFile(char *const, Burtobjlist *const);
BOOL BurtSDDSFile(const char *const);
void BurtExit(const int);

/***************************/
/*                         */
/* Private Local Functions */
/*                         */
/***************************/

static void init();
static void parse_command_line_args(int argc, char **argv, char **logfilename, 
    char **outfilename, char **reqfilenames, char **defines, char **includes,
    char *operation, BOOL *keep_generated, BOOL *verbose, char *output_spec);
static void perform_operation(Burtobjlist *op1_list, 
    Burtobjlist *op2_list, char operation, Burtobjlist *result_list);
static void process_command_line_args(char *logfilename, char *outfilename, 
    char **reqfilenames, char **defines, char **includes, Burtobjlist *op1_list,
     Burtobjlist *op2_list, BOOL keep_generated, char output_spec);

/****************************************************************
*
*
****************************************************************/

int main(int argc, char **argv)
{

char *logfilename;
char *reqfilenames[2];
char *defines[MAXDEFINES];
char *includes[MAXINCLUDES];
Burtobjlist op1_list, op2_list, result_list;
BOOL keep_generated;
char operation;
char output_spec;

    init();

    op1_list.set_owner();
    op2_list.set_owner();
    result_list.set_view();

    parse_command_line_args(argc, argv, &logfilename, &Outfilename,
        reqfilenames, defines, includes, &operation, &keep_generated, 
	&Verbose, &output_spec);

    process_command_line_args(logfilename, Outfilename, reqfilenames, defines, 
	includes, &op1_list, &op2_list, keep_generated, output_spec);

    perform_operation(&op1_list, &op2_list, operation, &result_list);

    result_list.write_req_to_file();

    BurtExit(0);

} /* end main() */

/******************************************************
*
*
******************************************************/

static void init()
{

    BurtCommonInit(FALSE);

} /* end init() */


/******************************************************
*
*
******************************************************/

static void parse_command_line_args(int argc, char **argv, char **logfilename, 
    char **outfilename, char **reqfilenames, char **defines, char **includes,
    char *operation, BOOL *keep_generated, BOOL *verbose, char *output_spec)
{

int arg;
int file;
int context;
int next_req_file;
int nincludes;
int ndefines;
BOOL got_all_request_files;
BOOL gotoperation;
BOOL error;

    /* default values */
    *logfilename = (char *) NULL;
    *outfilename = (char *) NULL;
    *operation = UNION;
    *keep_generated = FALSE;
    *verbose = FALSE;
    *output_spec = NOREQ;
    for (file = 0; file < 2; file ++)
	reqfilenames[file] = (char *) NULL;
    for (ndefines = 0; ndefines < MAXDEFINES; ndefines ++)
        defines[ndefines] = (char *) NULL;
    for (nincludes = 0; nincludes < MAXINCLUDES; nincludes ++)
        includes[nincludes] = (char *) NULL;

    nincludes = ndefines = next_req_file = 0;
    got_all_request_files = FALSE;
    gotoperation = FALSE;
    context = EXPECT_SWITCH;
    arg = 1;
    error = FALSE;
    while (arg < argc && !error)
    {
/* printf("arg %d >%s< context %s\n", arg, argv[arg], (context == EXPECT_SWITCH ? "expect switch" : (context == REQ_FILES ? "request files" : (context == LOG_FILE ? "log file" : "unknown")))); */
	switch (context)
	{
	    case EXPECT_SWITCH:
                if (strncmp(argv[arg], "-D", 2) == 0)
                {
                    /* -D option */
                    if (ndefines < MAXDEFINES)
                    {
                        defines[ndefines] = argv[arg];
                        arg ++;
                        ndefines ++;
                    }
                    else
                        /* too many -D specifications */
                        error = TRUE;
                }
                else if (strncmp(argv[arg], "-I", 2) == 0)
                {
                    /* -I option */
                    if (nincludes < MAXINCLUDES)
                    {
                        includes[nincludes] = argv[arg];
                        arg ++;
                        nincludes ++;
                    }
                    else
                        /* too many -I specifications */
                        error = TRUE;
                }
		else if (strcmp(argv[arg], "-f") == 0)
		{
		    if (got_all_request_files)
			/* already got all the request files from */
			/* from a previous -f switch              */
			error = TRUE;
		    else
		    {
			arg ++;
			context = REQ_FILES;
		    } /* endif */
		}
		else if (strcmp(argv[arg], "-l") == 0)
		{
		    if (*logfilename)
			/* already have a logfile from */
			/* a previous -l switch        */
			error = TRUE;
		    else
		    {
			arg ++;
			context = LOG_FILE;
		    } /* endif */
		}
		else if (strcmp(argv[arg], "-o") == 0)
		{
		    if (*outfilename)
			/* already have a outfile from */
			/* a previous -l switch        */
			error = TRUE;
		    else
		    {
			arg ++;
			context = OUT_FILE;
		    } /* endif */
		}
		else if (strcmp(argv[arg], "-d") == 0)
		{
		    arg ++;
		    *keep_generated = TRUE;
		}
		else if (strcmp(argv[arg], "-v") == 0)
		{
		    arg ++;
		    *verbose = TRUE;
		}
		else if (strcmp(argv[arg], "-union") == 0)
		{
		    if (gotoperation)
			/* already got operation from previous switch */
			error = TRUE;
		    else
		    {
			arg ++;
			gotoperation = TRUE;
			*operation = UNION;
		    } /* endif */
		}
		else if (strcmp(argv[arg], "-inter") == 0)
		{
		    if (gotoperation)
			/* already got operation from previous switch */
			error = TRUE;
		    else
		    {
			arg ++;
			gotoperation = TRUE;
			*operation = INTERSECTION;
		    } /* endif */
		}
		else if (strcmp(argv[arg], "-diff") == 0)
		{
		    if (gotoperation)
			/* already got operation from previous switch */
			error = TRUE;
		    else
		    {
			arg ++;
			gotoperation = TRUE;
			*operation = DIFFERENCE;
		    } /* endif */
		}
                else if (strcmp(argv[arg], "-sdds") == 0)
                {
                    if (*output_spec == NOREQ)
                    {
                        arg ++;
                        *output_spec = REQSDDS;
                    }
                    else
                        error = TRUE;
                }
                else if (strcmp(argv[arg], "-nosdds") == 0)
                {
                    if (*output_spec == NOREQ)
                    {
                        arg ++;
                        *output_spec = REQNOSDDS;
                    }
                    else
                        error = TRUE;
                }
		else 
		    /* did not get recognizable switch */
		    error = TRUE;
		break;
	    case REQ_FILES:
		if (*(argv[arg]) == '-')
		{
		    context = EXPECT_SWITCH;
		    got_all_request_files = TRUE;
		}
		else
		{
		    if (next_req_file < 2)
		    {
			reqfilenames[next_req_file] = argv[arg];
			next_req_file ++;
			arg ++;
		    }
		    else
                        error = TRUE;
		} /* endif */
		break;
	    case LOG_FILE:
		if (*(argv[arg]) == '-')
		    error = TRUE;
		else
		{
		    /* got logfilename */
		    *logfilename = argv[arg];
		    arg ++;
		    context = EXPECT_SWITCH;
		} /* endif */
		break;
	    case OUT_FILE:
		if (*(argv[arg]) == '-')
		    error = TRUE;
		else
		{
		    /* got logfilename */
		    *outfilename = argv[arg];
		    arg ++;
		    context = EXPECT_SWITCH;
		} /* endif */
		break;
	    default:
		error = TRUE;
		break;
	} /* end switch() */
    } /* endwhile */

    error |= (context == LOG_FILE || context == OUT_FILE 
		|| next_req_file != 2);

    if (error)
    {
printf("\n\tusage: %s -f req1 req2 {-union or -inter or -diff} {-l logfile}\n",
    argv[0]);
printf("\t\t{-o outfile} {-d} {-v} {-sdds or -nosdds} {-Dname{=def}}\n");
printf("\t\t{-Ipathname}\n\n");

printf("where\n\n");

printf("\t-f req1 req2 - Request filenames. This is the only switch\n");
printf("\t\tthat is not optional.  You must specify exactly request\n");
printf("\t\tfiles.\n\n");

printf("\t-union or -inter or -diff - Set operation.  The set operation\n");
printf("\t\t(union, intersection, or difference) to be performed on the two\n");
printf("\t\trequest files.  The default is union.\n\n");

printf("\t-l logfile - Log filename. The name of the file where all logging\n");
printf("\t\tmessages (e.g. error messages) go.  The default is stderr.\n\n");

printf("\t-o outfile - Request filename.  The name of the file where the\n");
printf("\t\tresult of the set operation goes.  The default is stdout.\n\n");

printf("\t-d - Debug.  Save the files created by processing the request\n");
printf("\t\tfiles with the C preprocessor.  The default is to delete these\n");
printf("\t\tfiles.\n\n");

printf("\t-v - Verbose.  This increases the amount of information displayed\n");
printf("\t\tin the logfile.\n\n");

printf("\t-sdds or -nosdds - SDDS/non-SDDS snapshot file.  Explicitly\n");
printf("\t\tspecifying that the generated request file will be \n");
printf("\t\tSDDS/non-SDDS compliant.  The default is to adopt the SDDS\n");
printf("\t\ttype from the inputs.  If there is a heterogeneous set of\n");
printf("\t\tinputs (one SDDS and the other non-SDDS), the default is to\n");
printf("\t\tproduce an SDDS compliant request file.\n\n");

printf("\t-Dname{=def} - Defines a symbol name to the C preprocessor.\n");
printf("\t\tEquivalent to a #define directive.  If no def is given, name\n");
printf("\t\tis defined as `1'.  The maximum number that can be specified\n");
printf("\t\tin this manner is %d.\n\n", MAXDEFINES);

printf("\t-Ipathname - Add pathname to the list of directories in which\n");
printf("\t\tto search for #include files with relative filenames (not\n");
printf("\t\tbeginning with slash `/').  The preprocessor first searches\n");
printf("\t\tfor #include files in the directory containing the request\n");
printf("\t\tfile, then in the directories named with -I option (if any).\n\n");
printf("\t\tThe maximumn number that can be specified in this manner\n");
printf("\t\tis %d.\n\n", MAXINCLUDES);

	exit(1);
    } /* endif */

} /* end parse_command_line_args() */

/******************************************************
*
* This function assumes that there may be duplicates (name, nreq) in op1_list 
* and op2_list. However, it assures that there will be no duplicates 
* (name, nreq) in result_list.  
*
* The RO values of the result_list are set if EITHER element in op1_list or 
* op2_list is RO.  In the case where there are duplicates (name, nreq) in 
* either op1_list or op2_list, the RO values in the result_list is up for
* grabs ... most likely adopting the RO value of the first occurrence in
* either list.
* 
*
* NOTE: that this function does not bother to make a copy of each item
*       before placing into the result_list.  It takes each instance and 
*       appends it into the result_list.  Therefore, the result_list MUST 
*       be a view only.
*
* NOTE: this function can potentially change the RO values of elements
*       in op1_list and op2_list.
*
******************************************************/

static void perform_operation(Burtobjlist *op1_list, 
    Burtobjlist *op2_list, char operation, Burtobjlist *result_list)
{

Cursor<Burtobj> *rcp;
Burtobj *rp, *temp_rp;

    if (!op1_list || !op2_list || !result_list)
    {
	fprintf(Log_fp, 
"ERROR: perform_operation() found op1_list, %x op2_list %x result_list %x\n",
	op1_list, op2_list, result_list);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!(result_list->empty()))
    {
	fprintf(Log_fp, 
	    "ERROR: perform_operation() found non-empty result_list\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    result_list->set_view();

    switch (operation)
    {
	case UNION:
	    if (Verbose)
	    {
		fprintf(Log_fp, "\n**** operation = Union\n\n");
		fflush(Log_fp);
	    } /* endif */

	    /* inserting op1_list */
	    rcp = new Cursor<Burtobj>(op1_list);
	    for (rp = rcp->getfirst(); rp; rp = rcp->getnext())
	    {
		if (!(result_list->find(rp)))
		{
		    /* not in result_list ... must append */

		    /* preserving RO */

		    /* setting RO in op1_list if not already set */
		    /* and corresponding value set in op2_list   */

		    if (!(rp->read_only_notify()))
		    {
			/* rp not RON */
			temp_rp = op2_list->find(rp);
			if (temp_rp && temp_rp->read_only_notify())
			{
			    /* rp not RON, temp_rp RON */
			    rp->set_read_only_notify();
			    rp->set_read_only();
			} 
			else
			{
			    /* neither RON */

			    if (!(rp->read_only()))
				if (temp_rp && temp_rp->read_only())
				    rp->set_read_only();
			} /* endif */
		    } /* endif */

		    result_list->append(rp);

		} /* endif */
	    } /* endfor */
	    delete rcp;

	    /* inserting op2_list */
	    rcp = new Cursor<Burtobj>(op2_list);
	    for (rp = rcp->getfirst(); rp; rp = rcp->getnext())
	    {
		if (!(result_list->find(rp)))
		{
		    /* not in result_list ... must append */

		    /* preserving RO */

		    /* setting RO in op2_list if not already set */
		    /* and corresponding value set in op1_list   */

		    if (!(rp->read_only_notify()))
		    {
			/* rp not RON */
			temp_rp = op1_list->find(rp);
			if (temp_rp && temp_rp->read_only_notify())
			{
			    /* rp not RON, temp_rp RON */
			    rp->set_read_only_notify();
			    rp->set_read_only();
			} 
			else
			{
			    /* neither RON */

			    if (!(rp->read_only()))
				if (temp_rp && temp_rp->read_only())
				    rp->set_read_only();
			} /* endif */
		    } /* endif */

		    result_list->append(rp);

		} /* endif */
	    } /* endfor */
	    delete rcp;

	    break;
	case INTERSECTION:
	    if (Verbose)
	    {
		fprintf(Log_fp, "\n**** operation = Intersection\n\n");
		fflush(Log_fp);
	    } /* endif */

	    rcp = new Cursor<Burtobj>(op1_list);
	    for (rp = rcp->getfirst(); rp; rp = rcp->getnext())
	    {
		if ((temp_rp = op2_list->find(rp)) && !(result_list->find(rp)))
		{
		    /* preserving RO */
		    if (rp->read_only_notify() || temp_rp->read_only_notify())
		    {
			rp->set_read_only_notify();
			rp->set_read_only();
		    }
		    else if (rp->read_only() || temp_rp->read_only())
			rp->set_read_only();

		    result_list->append(rp);
		} /* endif */
	    } /* endfor */
	    delete rcp;
	    break;
	case DIFFERENCE:
	    if (Verbose)
	    {
		fprintf(Log_fp, "\n**** operation = Difference\n\n");
		fflush(Log_fp);
	    } /* endif */

	    rcp = new Cursor<Burtobj>(op1_list);
	    for (rp = rcp->getfirst(); rp; rp = rcp->getnext())
		if (!(op2_list->find(rp)) && !(result_list->find(rp)))
		    /* NOTE: RO automatically preserved */
		    result_list->append(rp);
	    delete rcp;
	    break;
	default:
	    fprintf(Log_fp, 
	"ERROR: perform_operation() encountered unrecognizable operation %d\n", 
		operation);
	    fflush(Log_fp);
	    BurtExit(1);
	    break;
    } /* end switch() */

} /* end perform_operation() */

/******************************************************
*
*
******************************************************/

static void process_command_line_args(char *logfilename, char *outfilename, 
    char **reqfilenames, char **defines, char **includes, Burtobjlist *op1_list,
     Burtobjlist *op2_list, BOOL keep_generated, char output_spec)
{

char newfilename[MAXPATHLEN];
int input_profile;

    if (logfilename)
    {
	if (!(Log_fp = fopen(logfilename, "w")))
	{
	    printf("ERROR: unable to open logfile >%s<\n", logfilename);
	    exit(1);
	} /* endif */
    }
    else
	Log_fp = stderr;

    if (Verbose)
    {
	fprintf(Log_fp, "**** SET BURT LOGFILE\n");
	fflush(Log_fp);
    } /* endif */

    if (!reqfilenames || !op1_list || !op2_list)
    {
	fprintf(Log_fp,
"ERROR: process_command_line_args() reqfilenames %x op1_list %x op2_list %x\n",
	    reqfilenames, op1_list, op2_list);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!defines)
    {
        fprintf(Log_fp,
            "ERROR: process_command_line_args() got NULL defines\n");
        fflush(Log_fp);
        BurtExit(1);
    } /* endif */

    if (!includes)
    {
        fprintf(Log_fp,
            "ERROR: process_command_line_args() got NULL includes\n");
        fflush(Log_fp);
        BurtExit(1);
    } /* endif */

    if (!(op1_list->empty()) || !(op2_list->empty()))
    {
	fprintf(Log_fp, 
"ERROR: process_command_line_args() found op1_list nelem %d op2_list nelem %d\n", 
	    op1_list->nelem(), op2_list->nelem());
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (Verbose)
    {
	fprintf(Log_fp, "\n--- Start processing files\n");
	fflush(Log_fp);
    } /* endif */

    if (reqfilenames[0])
    {
	if (Verbose)
	{
	    fprintf(Log_fp, "file >%s<\n", reqfilenames[0]);
	    fflush(Log_fp);
	} /* endif */

	if (BurtSDDSFile(reqfilenames[0]))
	{
	    input_profile = SDDSINPUT;

	    BurtProcessRequestFile(reqfilenames[0], op1_list);
	}
	else
	{
	    input_profile = NONSDDSINPUT;

            if (!BurtPreprocessFile(reqfilenames[0], (const char **)defines, 
		    (const char **)includes, newfilename))
            {
                fprintf(Log_fp, 
"ERROR: process_command_line_args() could not preprocess non-SDDS file >%s<\n", 
                    reqfilenames[0]);
                fflush(Log_fp);

                BurtExit(1);
            } /* endif */

	    BurtProcessRequestFile(newfilename, op1_list);

	    if (!keep_generated)
		unlink(newfilename);
	} /* endif */

    }
    else
    {
	fprintf(Log_fp, 
	"ERROR: process_command_line_args() found NULL reqfilenames[0]\n");
	    fflush(Log_fp);
	    BurtExit(1);
    } /* endif */

    if (reqfilenames[1])
    {
	if (Verbose)
	{
	    fprintf(Log_fp, "file >%s<\n", reqfilenames[1]);
	    fflush(Log_fp);
	} /* endif */

	if (BurtSDDSFile(reqfilenames[1]))
	{
	    if (input_profile != SDDSINPUT)
		input_profile = HETROINPUT;

	    BurtProcessRequestFile(reqfilenames[1], op2_list);
	}
	else
	{
	    if (input_profile != NONSDDSINPUT)
		input_profile = HETROINPUT;

            if (!BurtPreprocessFile(reqfilenames[1], (const char **)defines, 
		    (const char **)includes, newfilename))
            {
                fprintf(Log_fp, 
"ERROR: process_command_line_args() could not preprocess non-SDDS file >%s<\n", 
                    reqfilenames[1]);
                fflush(Log_fp);

                BurtExit(1);
            } /* endif */

	    BurtProcessRequestFile(newfilename, op2_list);

	    if (!keep_generated)
		unlink(newfilename);
	} /* endif */

    }
    else
    {
	fprintf(Log_fp, 
	"ERROR: process_command_line_args() found NULL reqfilenames[1]\n");
	    fflush(Log_fp);
	    BurtExit(1);
    } /* endif */

    if (Verbose)
    {
	fprintf(Log_fp, "--- End processing files\n");
	fflush(Log_fp);
    } /* endif */

    if (output_spec == REQSDDS)
        Sdds = TRUE;  /* explictly wants sdds */
    else if (output_spec == REQNOSDDS)
        Sdds = FALSE;  /* explictly does not want sdds */
    else if (input_profile == SDDSINPUT)
        Sdds = TRUE;  /* no specification, all input files sdds */
    else if (input_profile == NONSDDSINPUT)
        Sdds = FALSE;  /* no specification, all input files non-sdds */
    else
        Sdds = TRUE;  /* default ... no specifation, hetrogeneous input */

    if (!Sdds)
    {
	if (outfilename)
	{
	    if (!(Out_fp = fopen(outfilename, "w")))
	    {
		fprintf(Log_fp,
		    "ERROR: unable to open outfile >%s<\n", outfilename);
		fflush(Log_fp);
		BurtExit(1);
	    } /* endif */
	}
	else
	    Out_fp = stdout;
    } /* endif */

} /* end process_command_line_args() */

