/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#define MAINLINE

#include <stdlib.h>

#include "burtcommon.h"
#include "burtobjs.h"
#include "burtwb.h"

#include "list.h"
/*#include "cursor.h"*/
#include "burtlist.h"

#include <db_access.h> /* for DBR_STRING */

/****************************************/
/*                                      */
/* Externally Visibile Remote Functions */
/*                                      */
/****************************************/

/* in burtcommon.cc */
void BurtCommonInit(const BOOL);
BOOL BurtCreateUniqueFilename(const char *const, const char *const,char *const);
int BurtLongestLineLength(const char *const);
void BurtPipeWait(int *const, int *const);
BOOL BurtPreprocessFile(const char *const, const char **const,
    const char **const, char *const);
Burtobjlist *BurtProcessSnapshotFile(char *const, Burtobjlist *const);
BOOL BurtSDDSFile(const char *const);
void BurtExit(const int);

/***************************/
/*                         */
/* Private Local Functions */
/*                         */
/***************************/

static BOOL dependency_check(Burtobjlistlist *, Predicatelist *, BOOL, int);
static void get_dependency_pvars(Burtobjlistlist *, Predicatelist *, int);
static void get_ss_pvars(Burtobjlistlist *, Predicatelist *);
static void init();
static BOOL is_pvname_and_right_paren(char *, char **, char **);
static BOOL make_program();
static void parse_command_line_args(int argc, char **argv, char **logfilename, 
    char **outfilename, char ***snapfilenames, int *nsnapfilenames,
    char ***dependencyfilenames, int *ndepfilenames, char **defines, 
    char **includes, int *retry_count, BOOL *keep_generated, BOOL *verbose, 
    char *absolute_write_directive, char *relative_write_directive, 
    char *output_spec, char **comments, char **keywords);
static void print_predicate_in_printf(char *);
static void process_command_line_args(char *logfilename, char **snapfilenames, 
    int nsnapfilenames, Burtobjlistlist *snapshots, Predicatelist *predicates,
    char **dependencyfilenames, int ndepfilenames, char **defines, 
    char **includes, BOOL keep_generated, char output_spec, 
    Burtobjlist *readonlynotify);
static void process_dependency_input(Predicatelist *);
static void read_dependency_input(char *, Predicatelist *);
static BOOL run_program();
static void write_epilogue();
static void write_predicates(Predicatelist *);
static void write_program(Predicatelist *);
static void write_prologue();

/****************************************************************
*
*
****************************************************************/

int main(int argc, char **argv)
{

char *logfilename;
char **snapfilenames;
int nsnapfilenames;
char **dependencyfilenames;
int ndepfilenames;
char *defines[MAXDEFINES];
char *includes[MAXINCLUDES];
int retry_count;
BOOL keep_generated;
Burtobjlistlist snapshots;
Predicatelist predicates;
Cursor<Burtobjlist> *blcp;
Burtobjlist *blp;
Cursor<Burtobj> *bcp;
Burtobj *bp;
char absolute_write_directive, relative_write_directive;
char output_spec;
Burtobjlist readonlynotify;
char *comments;
char *keywords;
BOOL found_all;

    init();
    snapshots.set_owner();
    predicates.set_owner();

    readonlynotify.set_view();

    parse_command_line_args(argc, argv, &logfilename, &Outfilename, 
	&snapfilenames, &nsnapfilenames, &dependencyfilenames, &ndepfilenames,
	defines, includes, &retry_count, &keep_generated, &Verbose, 
	&absolute_write_directive, &relative_write_directive, &output_spec, 
	&comments, &keywords);

    process_command_line_args(logfilename, snapfilenames, nsnapfilenames,
	&snapshots, &predicates, dependencyfilenames, ndepfilenames,
	defines, includes, keep_generated, output_spec, &readonlynotify);

    if (snapfilenames)
	free((char *) snapfilenames);

    if (dependencyfilenames)
	free((char *) dependencyfilenames);

    found_all = TRUE;
    if (dependency_check(&snapshots, &predicates, keep_generated, retry_count))
    {
	blcp = new Cursor<Burtobjlist>(&snapshots);
	for (blp = blcp->getfirst(); blp; blp = blcp->getnext())
	    found_all &= (blp->write_snapshot_to_iocs(retry_count, 
			absolute_write_directive, relative_write_directive));
	delete blcp;
    } /* endif */

    /* writing the read only notify values to the file */
    /* only if there are any                           */
    if ((readonlynotify.empty()))
    {
	if (Verbose)
	{
	    fprintf(Log_fp, 
"\ndid not find any read only notify values ... NO OUTPUT FILE CREATED\n\n");
	    fflush(Log_fp);
	} /* endif */
    }
    else
    {
	/* must only open file here because it should not even  */
	/* be created unless there is something to put in there */

	if (!Sdds)
	{
	    if (Outfilename)
	    {
		if (!(Out_fp = fopen(Outfilename, "w")))
		{
		    fprintf(Log_fp, "ERROR: unable to open outfile >%s<\n", 
			Outfilename);
		    fflush(Log_fp);
		    BurtExit(1);
		} /* endif */
	    }
	    else
		Out_fp = stdout;
	} /* endif */

	(readonlynotify.get_parms())->set_keywords((keywords ? keywords : ""));
	(readonlynotify.get_parms())->set_comments((comments ? comments : ""));
	(readonlynotify.get_parms())->set_nowrite();
	(readonlynotify.get_parms())->set_who_and_time_stamp();

	readonlynotify.write_snapshot_to_file();

	if (comments)
	    free(comments);

	if (keywords)
	    free(keywords);
    } /* endif */

    /* loop to explicitly free all the Value objects that were */
    /* dynamically allocated during process_dependency_input() */
    /* ... not really necessary because about to exit program  */
    /* anyway ... but doing it just to be complete.            */
    bcp = new Cursor<Burtobj>(predicates.get_burtobjs());
    for (bp = bcp->getfirst(); bp; bp = bcp->getnext())
	delete (bp->get_readchannel())->get_val();
    delete bcp;

    if (found_all)
	BurtExit(0);
    else
	BurtExit(1);

} /* end main() */

/****************************************************************
*
*
****************************************************************/

static BOOL dependency_check(Burtobjlistlist *snapshots, 
    Predicatelist *predicates, BOOL keep_generated, int retry_count)
{

BOOL rc;

    if (!snapshots)
    {
	fprintf(Log_fp, "ERROR: dependency_check() got NULL snapshots\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!predicates)
    {
	fprintf(Log_fp, "ERROR: dependency_check() got NULL predicates\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!predicates->empty())
    {
	get_dependency_pvars(snapshots, predicates, retry_count);

	write_program(predicates);
	if (rc = make_program())
	    rc = run_program();

	if (!keep_generated)
	{
	    unlink(SourceFilename);
	    unlink(ExecutableFilename);
	} /* endif */
    }
    else
	rc = TRUE;

    return rc;

} /* end dependency_check() */

/****************************************************************
*
*
****************************************************************/

static void get_dependency_pvars(Burtobjlistlist *snapshots, 
    Predicatelist *predicates, int retry_count)
{

    if (!snapshots)
    {
	fprintf(Log_fp, "ERROR: get_dependency_pvars() got NULL snapshots\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!predicates)
    {
	fprintf(Log_fp, "ERROR: get_dependency_pvars() got NULL predicates\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    /* getting snapshot pvars */
    get_ss_pvars(snapshots, predicates);

    /* getting IOC pvars */
    (predicates->get_burtobjs())->read_snapshot_from_iocs(retry_count);

} /* end get_dependency_pvars() */

/****************************************************************
*
*
****************************************************************/

static void get_ss_pvars(Burtobjlistlist *snapshots, Predicatelist *predicates)
{

Cursor<Predicate> *pcp;
Predicate *pp;
Cursor<Sspvar> *scp;
Sspvar *sp;
Burtobj *cp;

    if (!snapshots)
    {
	fprintf(Log_fp, "ERROR: get_dependency_pvars() got NULL snapshots\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!predicates)
    {
	fprintf(Log_fp, "ERROR: get_dependency_pvars() got NULL predicates\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    pcp = new Cursor<Predicate>(predicates);
    for (pp = pcp->getfirst(); pp; pp = pcp->getnext())
    {
	scp = new Cursor<Sspvar>(pp->get_ss_pvars_list());
	for (sp = scp->getfirst(); sp; sp = scp->getnext())
	{
	    if (cp = snapshots->find_name(sp->get_name()))
	    {
		if ((cp->get_val())->defined() 
		    && (cp->get_val())->get_nelem() == 1)
			sp->set_val(cp->get_val());
		else
		{

		    sp->set_val((Value *) NULL);

		    if (!(pp->printed()))
		    {
			pp->set_printed();
			fprintf(Log_fp, "%s\n", pp->get_raw_input_line());
			fflush(Log_fp);
		    } /* endif */

		    fprintf(Log_fp, "nelem != 1 in snapshot for pvar >%s<\n", 
			sp->get_name());
		    fflush(Log_fp);

		} /* endif */
	    }
	    else
	    {
		sp->set_val((Value *) NULL);

		if (!(pp->printed()))
		{
		    pp->set_printed();
		    fprintf(Log_fp, "%s\n", pp->get_raw_input_line());
		    fflush(Log_fp);
		} /* endif */

		fprintf(Log_fp, "pvar >%s< not found in snapshot\n", 
		    sp->get_name());
		fflush(Log_fp);
	    } /* endif */
	} /* endfor */
	delete scp;
    } /* endfor */
    delete pcp;

} /* end get_ss_pvars() */

/****************************************************************
*
*
****************************************************************/

static void init()
{

    if (VALID_ATOMIC_DBR_REQ(NATIVE))
    {
	fprintf(stderr, 
	    "ERROR: init() found that NATIVE (%d) is a valid dbr req type\n", 
	    NATIVE);
	fflush(stderr);
	exit(1);
    } /* endif */

    if (VALID_ATOMIC_DBR_REQ(INVALID_DBR))
    {
	fprintf(stderr, 
	"ERROR: init() found that INVALID_DBR (%d) is a valid dbr req type\n", 
	    INVALID_DBR);
	fflush(stderr);
	exit(1);
    } /* endif */

    BurtCommonInit(TRUE);

    Log_fp = stderr;

    /* ca_get() function names */
    /* it is important that any one function name not be a  */
    /* subset of another function name (i.e. fff() fffgg()) */
    /* with the allowable exception that Get_native be a    */
    /* subset of anything. This is because of the way the   */
    /* input strings are parsed in read_dependency_input(). */

    strcpy(Get_native, Get_Base);

    sprintf(Get_string, "%s_string", Get_Base);
    sprintf(Get_int, "%s_int", Get_Base);
    sprintf(Get_short, "%s_short", Get_Base);
    sprintf(Get_float, "%s_float", Get_Base);
    sprintf(Get_enum, "%s_enum", Get_Base);
    sprintf(Get_char, "%s_char", Get_Base);
    sprintf(Get_long, "%s_long", Get_Base);
    sprintf(Get_double, "%s_double", Get_Base);

    strcpy(SnapShotGet_native, SnapShotGet_Base);
    sprintf(SnapShotGet_string, "%s_string", SnapShotGet_Base);

} /* end init() */

/****************************************************************
*
* start points to right after '(' upon entry.
*
* upon exit TRUE --> pvar points to beginning of pvar
*                    right_paren points to ')'
*
****************************************************************/

static BOOL is_pvname_and_right_paren(char *start, char **pvar, char **right_paren)
{

BOOL rc;

    if (!start)
    {
	fprintf(Log_fp, "ERROR: is_pvname_and_right_paren() got NULL start\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!pvar)
    {
	fprintf(Log_fp, "ERROR: is_pvname_and_right_paren() got NULL pvar\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!right_paren)
    {
	fprintf(Log_fp, 
	    "ERROR: is_pvname_and_right_paren() got NULL right_paren\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    /* passing over any blanks between ( and pvarname */

    while (*start == ' ')
	start ++;

    if (*start == '\0')
	/* was not really a call to read from ioc */
	rc = FALSE;
    else
    {
	/* found the pvarname ... still need to find ')' */

	*pvar = start;

	for (*right_paren = start; 
	    *(*right_paren) != '(' && *(*right_paren) != ')' && *(*right_paren);
		*right_paren = *right_paren + 1)
	;

	rc = (*(*right_paren) == ')');

    } /* endif */

    if (!rc)
    {
	*pvar = (char *) NULL;
	*right_paren = (char *) NULL;
    } /* endif */

    return rc;

} /* end is_pvname_and_right_paren() */

/****************************************************************
*
*
****************************************************************/

static BOOL make_program()
{

long pid;
int exit_status, signal;
int status;
char *cp;
char system_command[100];
BOOL rc;

    rc = TRUE;

    if (Verbose)
    {
	fprintf(Log_fp, "making program ... ");
	fflush(Log_fp);
    } /* endif */

    strcpy(ExecutableFilename, SourceFilename);

    if (cp = strstr(ExecutableFilename, ".c"))
    {
	*cp = '\0';

	sprintf(system_command, "cc -o %s %s", 
	    ExecutableFilename, SourceFilename);
	Exec_args[2] = system_command;

	pid = fork();

	if (pid >= 0)
	{
	    if (pid > 0)
	    {
		/* I am parent */
		BurtPipeWait(&exit_status, &signal);
		if (exit_status)
		{
		    if (Verbose)
		    {
			fprintf(Log_fp, "ERROR exit_status %d signal %d\n", 
			    exit_status, signal);
		    } /* endif */

		    rc = FALSE;
		}
		else
		{
		    if (Verbose)
		    {
			fprintf(Log_fp, "done\n");
			fflush(Log_fp);
		    } /* endif */
		} /* endif */
	    } 
	    else
	    {
		/* I am child, pid == 0 */
		status = execv(Sh_path, Exec_args);

		/* child process must die */
		exit(0); 
	    } /* endif */
	}
	else
	{
	    fprintf(Log_fp, 
		"ERROR: make_program() illegal pid %d from fork()\n", pid);

	    rc = FALSE;

	} /* endif */

    } 
    else
    {

	fprintf(Log_fp, 
	    "ERROR: make_program() could not find .c extension in >%s<\n", 
		ExecutableFilename);
	fflush(Log_fp);

	rc = FALSE;

    } /* endif */


    return rc;

} /* end make_program() */

/******************************************************
*
*
******************************************************/

static void parse_command_line_args(int argc, char **argv, char **logfilename, 
    char **outfilename, char ***snapfilenames, int *nsnapfilenames,
    char ***dependencyfilenames, int *ndepfilenames, char **defines, 
    char **includes, int *retry_count, BOOL *keep_generated, BOOL *verbose, 
    char *absolute_write_directive, char *relative_write_directive, 
    char *output_spec, char **comments, char **keywords)
{

int arg;
int context;
int nbytes;
int i;
int nincludes;
int ndefines;
BOOL got_all_snap_files;
BOOL got_all_dep_files;
BOOL gotretrycount;
BOOL got_all_comments;
BOOL got_all_keywords;
BOOL error;

    /* default values */
    *logfilename = (char *) NULL;
    *outfilename = (char *) NULL;
    *snapfilenames = (char **) NULL;
    *nsnapfilenames = 0;
    *dependencyfilenames = (char **) NULL;
    *ndepfilenames = 0;
    *retry_count = 0;
    *keep_generated = FALSE;
    *verbose = FALSE;
    *absolute_write_directive = DEFAULTWRITE;
    *relative_write_directive = DEFAULTWRITE;
    *output_spec = NOREQ;
    *comments = (char *) NULL;
    *keywords = (char *) NULL;
    for (ndefines = 0; ndefines < MAXDEFINES; ndefines ++)
        defines[ndefines] = (char *) NULL;
    for (nincludes = 0; nincludes < MAXINCLUDES; nincludes ++)
        includes[nincludes] = (char *) NULL;

    got_all_snap_files = FALSE;
    got_all_dep_files = FALSE;
    gotretrycount = FALSE;
    got_all_comments = FALSE;
    got_all_keywords = FALSE;
    context = EXPECT_SWITCH;
    nincludes = ndefines = 0;
    arg = 1;
    error = FALSE;
    while (arg < argc && !error)
    {
/* printf("error %c argc %d arg %d >%s< context %s\n", (error ? 'T' : 'F'), argc, arg, argv[arg], (context == EXPECT_SWITCH ? "expect switch" : (context == SNAP_FILES ? "snap files" : (context == LOG_FILE ? "log file" : (context == DEP_FILES ? "dep files" : (context == RETRY_COUNT ? "retry count" : (context == OUT_FILE : "out file" : (context == COMMENTS ? "comments" : (context = KEYWORDS ? "keywords" : "unknown"))))))))); */
	switch (context)
	{
	    case EXPECT_SWITCH:
                if (strncmp(argv[arg], "-D", 2) == 0)
                {
                    /* -D option */
                    if (ndefines < MAXDEFINES)
                    {
                        defines[ndefines] = argv[arg];
                        arg ++;
                        ndefines ++;
                    }
                    else
                        /* too many -D specifications */
                        error = TRUE;
                }
                else if (strncmp(argv[arg], "-I", 2) == 0)
                {
                    /* -I option */
                    if (nincludes < MAXINCLUDES)
                    {
                        includes[nincludes] = argv[arg];
                        arg ++;
                        nincludes ++;
                    }
                    else
                        /* too many -I specifications */
                        error = TRUE;
                }
		else if (strcmp(argv[arg], "-f") == 0)
		{
		    if (got_all_snap_files)
			/* already got all the request files from */
			/* from a previous -f switch              */
			error = TRUE;
		    else
		    {
			arg ++;
			context = SNAP_FILES;
		    } /* endif */
		}
		else if (strcmp(argv[arg], "-p") == 0)
		{
		    if (got_all_dep_files)
			/* already got all the dep files from */
			/* from a previous -p switch              */
			error = TRUE;
		    else
		    {
			arg ++;
			context = DEP_FILES;
		    } /* endif */
		}
		else if (strcmp(argv[arg], "-l") == 0)
		{
		    if (*logfilename)
			/* already have a logfile from */
			/* a previous -l switch        */
			error = TRUE;
		    else
		    {
			arg ++;
			context = LOG_FILE;
		    } /* endif */
		}
		else if (strcmp(argv[arg], "-r") == 0)
		{
		    if (gotretrycount)
			/* already have seconds from */
			/* a previous -s switch        */
			error = TRUE;
		    else
		    {
			arg ++;
			context = RETRY_COUNT;
		    } /* endif */
		}
		else if (strcmp(argv[arg], "-d") == 0)
		{
		    arg ++;
		    *keep_generated = TRUE;
		}
		else if (strcmp(argv[arg], "-v") == 0)
		{
		    arg ++;
		    *verbose = TRUE;
		}
		else if (strcmp(argv[arg], "-add") == 0)
		{
		    arg ++;
		    *absolute_write_directive = ADDWRITE;
		}
		else if (strcmp(argv[arg], "-replace") == 0)
		{
		    arg ++;
		    *relative_write_directive = REPLACEWRITE;
		}
                else if (strcmp(argv[arg], "-o") == 0)
                {
                    if (*outfilename)
                        /* already have a outfile from */
                        /* a previous -o switch        */
                        error = TRUE;
                    else
                    {
                        arg ++;
                        context = OUT_FILE;
                    } /* endif */
                }
                else if (strcmp(argv[arg], "-c") == 0)
                {
                    if (got_all_comments)
                        /* already got all the comments from */
                        /* from a previous -c switch         */
                        error = TRUE;
                    else
                    {
                        arg ++;
                        context = COMMENTS;
                    } /* endif */
                }
                else if (strcmp(argv[arg], "-k") == 0)
                {
                    if (got_all_keywords)
                        /* already got all the keywords from */
                        /* from a previous -k switch         */
                        error = TRUE;
                    else
                    {
                        arg ++;
                        context = KEYWORDS;
                    } /* endif */
                }
                else if (strcmp(argv[arg], "-sdds") == 0)
                {
                    if (*output_spec == NOREQ)
                    {
                        arg ++;
                        *output_spec = REQSDDS;
                    }
                    else
                        error = TRUE;
                }
                else if (strcmp(argv[arg], "-nosdds") == 0)
                {
                    if (*output_spec == NOREQ)
                    {
                        arg ++;
                        *output_spec = REQNOSDDS;
                    }
                    else
                        error = TRUE;
                }
		else 
		    /* did not get recognizable switch */
		    error = TRUE;
		break;
	    case SNAP_FILES:
                /* processing all the snap files at once */

                /* counting all the snap files */
                for (i = arg; i < argc && *(argv[i]) != '-'; i ++)
                    *nsnapfilenames = *nsnapfilenames + 1;

                if (*nsnapfilenames)
                {
                    if (*snapfilenames =
                        (char **) malloc(*nsnapfilenames*sizeof(char *)))
                    {
                        for (i = 0; arg+i < argc && *(argv[arg+i]) != '-'; i ++)
                            (*snapfilenames)[i] = argv[arg+i];

                        if (arg+i < argc)
                        {
                            context = EXPECT_SWITCH;
                            got_all_snap_files = TRUE;
                        } /* endif */

                        arg += i;

                    }
                    else
                    {
                        fprintf(stderr,
"ERROR: unable to malloc nfiles %d * sizeof(char**) = %d bytes for fnames\n",
                            *nsnapfilenames, sizeof(char *),
                            *nsnapfilenames*sizeof(char *));
                        fflush(stderr);
                        exit(1);
                    } /* endif */
                }
                else
                {
                    arg ++;
                    context = EXPECT_SWITCH;
                } /* endif */
		break;
	    case DEP_FILES:
                /* processing all the dependency files at once */

                /* counting all the dependency files */
                for (i = arg; i < argc && *(argv[i]) != '-'; i ++)
                    *ndepfilenames = *ndepfilenames + 1;

                if (*ndepfilenames)
                {
                    if (*dependencyfilenames =
                        (char **) malloc(*ndepfilenames*sizeof(char *)))
                    {
                        for (i = 0; arg+i < argc && *(argv[arg+i]) != '-'; i ++)
                            (*dependencyfilenames)[i] = argv[arg+i];

                        if (arg+i < argc)
                        {
                            context = EXPECT_SWITCH;
                            got_all_dep_files = TRUE;
                        } /* endif */

                        arg += i;

                    }
                    else
                    {
                        fprintf(stderr,
"ERROR: unable to malloc nfiles %d * sizeof(char**) = %d bytes for fnames\n",
                            *ndepfilenames, sizeof(char *),
                            *ndepfilenames*sizeof(char *));
                        fflush(stderr);
                        exit(1);
                    } /* endif */
                }
                else
                {
                    arg ++;
                    context = EXPECT_SWITCH;
                } /* endif */
		break;
	    case LOG_FILE:
		if (*(argv[arg]) == '-')
		    error = TRUE;
		else
		{
		    /* got logfilename */
		    *logfilename = argv[arg];
		    arg ++;
		    context = EXPECT_SWITCH;
		} /* endif */
		break;
            case COMMENTS:
                /* processing all the comments at once */

                /* counting the bytes in all the comments */
                for (i = arg, nbytes = 0; i < argc && *(argv[i]) != '-'; i ++)
                    nbytes = nbytes + strlen(argv[i]) + 1;

		if (nbytes)
		{
		    if (*comments = (char *) malloc(nbytes))
		    {
			**comments = '\0';

			for (i= arg; i < argc && *(argv[i]) != '-'; i ++)
			{
			    if (i != arg)
				strcat(*comments, " ");

			    strcat(*comments, argv[i]);
			} /* endfor */

			if (i < argc)
			{
			    context = EXPECT_SWITCH;
			    got_all_comments = TRUE;
			} /* endif */

			arg = i;
		    }
		    else
		    {
			fprintf(stderr,
			    "ERROR: unable to malloc %d bytes for comments\n",
			    nbytes);
			fflush(stderr);
			exit(1);
		    } /* endif */
		}
		else
		{
		    arg ++;
		    context = EXPECT_SWITCH;
		} /* endif */
                break;
            case KEYWORDS:
                /* processing all the keywords at once */

                /* counting the bytes in all the keywords */
                for (i = arg, nbytes = 0; i < argc && *(argv[i]) != '-'; i ++)
                    nbytes = nbytes + strlen(argv[i]) + 1;

		if (nbytes)
		{
		    if (*keywords = (char *) malloc(nbytes))
		    {
			**keywords = '\0';

			for (i= arg; i < argc && *(argv[i]) != '-'; i ++)
			{
			    if (i != arg)
				strcat(*keywords, " ");

			    strcat(*keywords, argv[i]);
			} /* endfor */

			if (i < argc)
			{
			    context = EXPECT_SWITCH;
			    got_all_keywords = TRUE;
			} /* endif */

			arg = i;
		    }
		    else
		    {
			fprintf(stderr,
			    "ERROR: unable to malloc %d bytes for keywords\n",
			    nbytes);
			fflush(stderr);
			exit(1);
		    } /* endif */
		}
		else
		{
		    arg ++;
		    context = EXPECT_SWITCH;
		} /* endif */
                break;
            case OUT_FILE:
                if (*(argv[arg]) == '-')
                    error = TRUE;
                else
                {
                    /* got outfilename */
                    *outfilename = argv[arg];
                    arg ++;
                    context = EXPECT_SWITCH;
                } /* endif */
                break;
            case RETRY_COUNT:
                if (*(argv[arg]) == '-')
                    error = TRUE;
                else
                {
                    /* got retry count */
                    *retry_count = atoi(argv[arg]);
                    arg ++;
                    context = EXPECT_SWITCH;
		    if (*retry_count < 0)
			*retry_count = 0;
                } /* endif */
		break;
	    default:
		error = TRUE;
		break;
	} /* end switch() */
    } /* endwhile */

/* printf("out of loop error %c arg %d context %s\n", (error ? 'T' : 'F'), arg, (context == EXPECT_SWITCH ? "expect switch" : (context == SNAP_FILES ? "snap files" : (context == LOG_FILE ? "log file" : (context == DEP_FILES ? "dep files" : (context == RETRY_COUNT ? "retry count" : (context == OUT_FILE ? "out file" : (context == KEYWORDS ? "keywords" : (context == COMMENTS ? "comments" : "unknown"))))))))); */

    error |= (context == RETRY_COUNT 
		|| (context == DEP_FILES && !(*ndepfilenames))
                || (context == COMMENTS && !(*comments))
                || (context == KEYWORDS && !(*keywords))
		|| context == LOG_FILE || context == OUT_FILE 
		|| !(*nsnapfilenames));

    if (error)
    {
printf("\n\tusage: %s -f snap1 {snap2 ...} {-l logfile} {-o outfile}\n", argv[0]);
printf("\t\t{-c ... comments ...} {-k keyword1 ... keywordn}\n");
printf("\t\t{-d} {-v} {-p dep1 ... depn} {-r retry count} {-add} {-replace}\n");
printf("\t\t{-sdds or -nosdds} {-Dname{=def}} {-Ipathname}\n\n");

printf("where\n\n");

printf("\t-f snap1 {snap2 ...} - Snapshot filenames.  This is the only\n");
printf("\t\tswitch that is not optional.  You must specify at least one\n");
printf("\t\tsnapshot file.\n\n");

printf("\t-l logfile - Log filename. The name of the file where all logging\n");
printf("\t\tmessages (e.g. error messages, reports of process variables\n");
printf("\t\tthat were not found) go.  The default is stderr.\n\n");

printf("\t-o outfile - Snapshot filename.  If any of the snapshot files\n");
printf("\t\tread only notify values, this file is created and those values\n");
printf("\t\tare placed there.  If none of the snapshot files have read only\n");
printf("\t\tnotify values, then no file is created.  The default is stdout.\n\n");

printf("\t-c ... comments ... - Comments.  Adds comments to the header of\n");
printf("\t\tthe snapshot file.\n\n");

printf("\t-k keyword1 ... keywordn - Keywords.  Adds keywords to the header\n");
printf("\t\tof the snapshot file.\n\n");

printf("\t-d - Debug.  Save the files created by processing the dependency\n");
printf("\t\tfiles with the C preprocessor.  The default is to delete these\n");
printf("\t\tfiles.\n\n");

printf("\t-v - Verbose.  This increases the amount of information displayed\n");
printf("\t\tin the logfile.\n\n");

printf("\t-p dep1 ... depn - Dependency filenames.  The names of the\n");
printf("\t\tdependency files containing predicates to be evaluated before\n");
printf("\t\twriting the values from the snapshot files.\n\n");

printf("\t-r retry count - Number of additional attempts to wait for\n");
printf("\t\tconnections.  The program will attempt to find all the process\n");
printf("\t\tvariables.  If it is unsuccessful, it will try this many more\n");
printf("\t\ttimes to establish connections.  The default value is 0.\n\n");

printf("\t-add - Absolute snapshots written as adds.  All the absolute\n");
printf("\t\tsnapshots, i.e., those taken directly off IOCs, will be written\n");
printf("\t\tas additions to the values found on the IOCs.  The default is\n");
printf("\t\tto write the absolute snapshots as replacement values on the \n");
printf("\t\tIOCs.\n\n");

printf("\t-replace - Relative snapshots written as replacements.  All the\n");
printf("\t\trelative snapshots, i.e., those generated by adding or\n");
printf("\t\tsubtracting two snapshots, will be written to prelace the\n");
printf("\t\tvalues on the IOCs.  The default is to write the relative\n");
printf("\t\tsnapshots as additions to the values on the IOCs.\n\n");

printf("\t-sdds or -nosdds - SDDS/non-SDDS snapshot file.  Explicitly\n");
printf("\t\tspecifying that the generated snapshot file will be \n");
printf("\t\tSDDS/non-SDDS compliant.  The default is to adopt the SDDS\n");
printf("\t\ttype from the input(s).  If there is a heterogeneous set of\n");
printf("\t\tinputs (some SDDS and some non-SDDS), the default is to produce\n");
printf("\t\tand SDDS compliant snapshot file.\n\n");

printf("\t-Dname{=def} - Defines a symbol name to the C preprocessor.\n");
printf("\t\tEquivalent to a #define directive.  If no def is given, name\n");
printf("\t\tis defined as `1'.  The maximum number that can be specified\n");
printf("\t\tin this manner is %d.\n\n", MAXDEFINES);

printf("\t-Ipathname - Add pathname to the list of directories in which\n");
printf("\t\tto search for #include files with relative filenames (not\n");
printf("\t\tbeginning with slash `/').  The preprocessor first searches\n");
printf("\t\tfor #include files in the directory containing the dependency\n");
printf("\t\tfile, then in the directories named with -I option (if any).\n\n");
printf("\t\tThe maximumn number that can be specified in this manner\n");
printf("\t\tis %d.\n\n", MAXINCLUDES);

	exit(1);

    } /* endif */

} /* end parse_command_line_args() */

/****************************************************************
*
*
****************************************************************/

static void print_predicate_in_printf(char *pred)
{

int i;

    for (i = 0; pred[i]; i ++)
	if (pred[i] == '"')
	    fprintf(Prog_fp, "\\\"");
	else if (pred[i] == '\\')
	    fprintf(Prog_fp, "\\\\");
	else if (pred[i] == '%')
	    fprintf(Prog_fp, "%%%%");
	else
	    fprintf(Prog_fp, "%c", pred[i]);

} /* end print_predicate_in_printf() */

/******************************************************
*
*
******************************************************/

static void process_command_line_args(char *logfilename, char **snapfilenames, 
    int nsnapfilenames, Burtobjlistlist *snapshots, Predicatelist *predicates,
    char **dependencyfilenames, int ndepfilenames, char **defines, 
    char **includes, BOOL keep_generated, char output_spec, 
    Burtobjlist *readonlynotify)
{

int i;
char newfilename[MAXPATHLEN];
Burtobjlist *clp;
char input_profile;

    if (logfilename)
    {
	if (!(Log_fp = fopen(logfilename, "w")))
	{
	    printf("ERROR: unable to open logfile >%s<\n", logfilename);
	    exit(1);
	} /* endif */
    }
    else
	Log_fp = stderr;

    if (Verbose)
    {
	fprintf(Log_fp, "**** WRITE BURT LOGFILE\n");
	fflush(Log_fp);
    } /* endif */

    if (!snapfilenames)
    {
	fprintf(Log_fp,
	    "ERROR: process_command_line_args() got NULL snapfilenames\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!snapshots)
    {
	fprintf(Log_fp,
	    "ERROR: process_command_line_args() got NULL snapshots\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!predicates)
    {
	fprintf(Log_fp,
	    "ERROR: process_command_line_args() got NULL predicates\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (ndepfilenames && !dependencyfilenames)
    {
	fprintf(Log_fp,
	"ERROR: process_command_line_args() got NULL dependencyfilenames\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!defines)
    {
        fprintf(Log_fp,
            "ERROR: process_command_line_args() got NULL defines\n");
        fflush(Log_fp);
        BurtExit(1);
    } /* endif */

    if (!includes)
    {
        fprintf(Log_fp,
            "ERROR: process_command_line_args() got NULL includes\n");
        fflush(Log_fp);
        BurtExit(1);
    } /* endif */

    if (!readonlynotify)
    {
	fprintf(Log_fp,
	    "ERROR: process_command_line_args() got NULL readonlynotify\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (Verbose)
    {
	fprintf(Log_fp, "\n--- Start processing snapshot files\n");
	fflush(Log_fp);
    } /* endif */

    for (i = 0; i < nsnapfilenames; i ++)
    {
	if (BurtSDDSFile(snapfilenames[i]))
	{
	    if (i == 0)
		input_profile = SDDSINPUT;
	    else if (input_profile != SDDSINPUT)
		input_profile = HETROINPUT;
	}
	else
	{
	    if (i == 0)
		input_profile = NONSDDSINPUT;
	    else if (input_profile != NONSDDSINPUT)
		input_profile = HETROINPUT;
	} /* endif */

	if (clp = BurtProcessSnapshotFile(snapfilenames[i], readonlynotify))
	    snapshots->append(clp);
	else
	{
	    fprintf(Log_fp, 
"ERROR: process_command_line_args() got NULL from BurtProcessSnapshotFile()\n");
	    fflush(Log_fp);
	    BurtExit(1);
	} /* endif */
    } /* endfor */

    if (Verbose)
    {
	fprintf(Log_fp, "--- End processing snapshot files\n");
	fflush(Log_fp);
    } /* endif */

    if (output_spec == REQSDDS)
        Sdds = TRUE;  /* explictly wants sdds */
    else if (output_spec == REQNOSDDS)
        Sdds = FALSE;  /* explictly does not want sdds */
    else if (input_profile == SDDSINPUT)
        Sdds = TRUE;  /* no specification, all input files sdds */
    else if (input_profile == NONSDDSINPUT)
        Sdds = FALSE;  /* no specification, all input files non-sdds */
    else
        Sdds = TRUE;  /* default ... no specifation, hetrogeneous input */

    if (Verbose)
    {
	fprintf(Log_fp, "\n--- Start reading dependency files\n");
	fflush(Log_fp);
    } /* endif */

    for (i = 0; i < ndepfilenames; i ++)
    {
	if (Verbose)
	{
	    fprintf(Log_fp, "file >%s<\n", dependencyfilenames[i]);
	    fflush(Log_fp);
	} /* endif */

	if (!BurtPreprocessFile(dependencyfilenames[i], (const char**)defines, 
		(const char**)includes, newfilename))
	{
	    fprintf(Log_fp, 
"ERROR: process_command_line_args() could not preprocess non-SDDS file >%s<\n", 
		dependencyfilenames[i]);
	    fflush(Log_fp);

	    BurtExit(1);
	} /* endif */

	read_dependency_input(newfilename, predicates);

	if (!keep_generated)
	    unlink(newfilename);

    } /* endfor */

    if (Verbose)
    {
	fprintf(Log_fp, "--- End reading dependency files\n");
	fflush(Log_fp);
    } /* endif */

    process_dependency_input(predicates);

} /* end process_command_line_args() */

/****************************************************************
*
*
****************************************************************/

static void process_dependency_input(Predicatelist *predicates)
{

char *pfunctor;
char *pvar;
char *start, *end;
char ss_print_type;
char dbr_type;
char tempc;
BOOL gotone;
BOOL foundfunctor;
Predicate *pp;
Iocpvar *ip;
Sspvar *sp;
Cursor<Predicate> *pcp;

    if (!predicates)
    {
	fprintf(Log_fp, 
	    "ERROR: process_dependency_input() got NULL predicates\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (Verbose)
    {
	fprintf(Log_fp, "\nprocessing all dependency files ... ");
	fflush(Log_fp);
    } /* endif */

    pcp = new Cursor<Predicate>(predicates);
    for (pp = pcp->getfirst(); pp; pp = pcp->getnext())
    {

	/* looking for failure directive at beginning of predicate */

	if (!(strncmp(pp->get_input_line(), HALTONFAILSTRING, 
	    strlen(HALTONFAILSTRING))))
	{
	    pp->set_directive_on_fail((char) HALTONFAIL);
	    pp->set_start(pp->get_input_line() + strlen(HALTONFAILSTRING));
	}
	else if (!(strncmp(pp->get_input_line(), CONTONFAILSTRING, 
	    strlen(CONTONFAILSTRING))))
	{
	    pp->set_directive_on_fail((char) CONTONFAIL);
	    pp->set_start(pp->get_input_line() + strlen(CONTONFAILSTRING));
	}
	else
	{
	    /* default */
	    pp->set_directive_on_fail((char) QUERYONFAIL);
	    pp->set_start(pp->get_input_line());
	} /* endif */

	/* looking for all the ca_get...() functors */
	pfunctor = strstr(pp->get_start(), Get_Base);

	while (pfunctor)
	{
	    /* found a candidate ca_get...() functor */
	    /* trying to identify which one it is.   */

	    /* should use a hash table here */
	    if (!strncmp(pfunctor, Get_string, strlen(Get_string)) 
		&& *(pfunctor+strlen(Get_string)) == '(')
	    {
		foundfunctor = TRUE;
		dbr_type = DBR_STRING;
		pvar = pfunctor + strlen(Get_string) + 1;
	    }
	    else if (!strncmp(pfunctor, Get_int, strlen(Get_int))
		&& *(pfunctor+strlen(Get_int)) == '(')
	    {
		foundfunctor = TRUE;
		dbr_type = DBR_INT;
		pvar = pfunctor + strlen(Get_int) + 1;
	    }
	    else if (!strncmp(pfunctor, Get_short, strlen(Get_short))
		&& *(pfunctor+strlen(Get_short)) == '(')
	    {
		foundfunctor = TRUE;
		dbr_type = DBR_SHORT;
		pvar = pfunctor + strlen(Get_short) + 1;
	    }
	    else if (!strncmp(pfunctor, Get_float, strlen(Get_float))
		&& *(pfunctor+strlen(Get_float)) == '(')
	    {
		foundfunctor = TRUE;
		dbr_type = DBR_FLOAT;
		pvar = pfunctor + strlen(Get_float) + 1;
	    }
	    else if (!strncmp(pfunctor, Get_enum, strlen(Get_enum))
		&& *(pfunctor+strlen(Get_enum)) == '(')
	    {
		foundfunctor = TRUE;
		dbr_type = DBR_ENUM;
		pvar = pfunctor + strlen(Get_enum) + 1;
	    }
	    else if (!strncmp(pfunctor, Get_char, strlen(Get_char))
		&& *(pfunctor+strlen(Get_char)) == '(')
	    {
		foundfunctor = TRUE;
		dbr_type = DBR_CHAR;
		pvar = pfunctor + strlen(Get_char) + 1;
	    }
	    else if (!strncmp(pfunctor, Get_long, strlen(Get_long))
		&& *(pfunctor+strlen(Get_long)) == '(')
	    {
		foundfunctor = TRUE;
		dbr_type = DBR_LONG;
		pvar = pfunctor + strlen(Get_long) + 1;
	    }
	    else if (!strncmp(pfunctor, Get_double, strlen(Get_double))
		&& *(pfunctor+strlen(Get_double)) == '(')
	    {
		foundfunctor = TRUE;
		dbr_type = DBR_DOUBLE;
		pvar = pfunctor + strlen(Get_double) + 1;
	    }
	    else if (!strncmp(pfunctor, Get_native, strlen(Get_native))
		&& *(pfunctor+strlen(Get_native)) == '(')
	    {
		foundfunctor = TRUE;
		dbr_type = NATIVE;
		pvar = pfunctor + strlen(Get_native) + 1;
	    }
	    else
		foundfunctor = FALSE;

	    if (foundfunctor)
	    {
		/* checking to see if rest is pvar followed by ')' */
		if (is_pvname_and_right_paren(pvar, &start, &end))
		{
		    gotone = TRUE;

		    ip = new Iocpvar;
		    ((ip->get_burtpvar())->get_channel())->set_val(new Value);

		    ip->set_start(pfunctor);
		    ip->set_end(end);

		    (ip->get_burtpvar())->clear_read_only_notify();
		    (ip->get_burtpvar())->clear_read_only();

		    tempc = *end;
		    *end = '\0';
		    ((ip->get_burtpvar())->get_channel())->set_name(start);
		    *end = tempc;

		    ((ip->get_burtpvar())->get_channel())->set_nreq(1);
		    ((ip->get_burtpvar())->get_channel())->
			set_dbr_req_type(dbr_type);

		    (predicates->get_burtobjs())->
			append((Burtobj *) ip->get_burtpvar());

		    (pp->get_ioc_pvars_list())->append(ip);
		}
		else
		    gotone = FALSE;
	    } 
	    else
		gotone = FALSE;

	    if (gotone)
		pfunctor = end + 1;
	    else
		pfunctor = pfunctor + 1;

	    pfunctor = strstr(pfunctor, Get_Base);

	} /* endwhile */

	/* looking for all the ss_get...() functors */
	pfunctor = strstr(pp->get_start(), SnapShotGet_Base);

	while (pfunctor)
	{
	    /* found a candidate ca_get...() functor */
	    /* trying to identify which one it is.   */

	    if (!strncmp(pfunctor, SnapShotGet_string, 
		strlen(SnapShotGet_string))
		    && *(pfunctor+strlen(SnapShotGet_string)) == '(')
	    {
		foundfunctor = TRUE;
		ss_print_type = SS_STRING;
		pvar = pfunctor + strlen(SnapShotGet_string) + 1;
	    }
	    else if (!strncmp(pfunctor, SnapShotGet_native, 
		strlen(SnapShotGet_native))
		    && *(pfunctor+strlen(SnapShotGet_native)) == '(')
	    {
		foundfunctor = TRUE;
		ss_print_type = SS_NOSTRING;
		pvar = pfunctor + strlen(SnapShotGet_native) + 1;
	    }
	    else
		foundfunctor = FALSE;

	    if (foundfunctor)
	    {
		/* checking to see if rest is pvar followed by ')' */
		if (is_pvname_and_right_paren(pvar, &start, &end))
		{
		    gotone = TRUE;

		    sp = new Sspvar;

		    sp->set_start(pfunctor);
		    sp->set_end(end);
		    sp->set_print_type(ss_print_type);

		    tempc = *end;
		    *end = '\0';
		    sp->set_name(start);
		    *end = tempc;

		    (pp->get_ss_pvars_list())->append(sp);

		}
		else
		    gotone = FALSE;
	    } 
	    else
		gotone = FALSE;

	    if (gotone)
		pfunctor = sp->get_end() + 1;
	    else
		pfunctor = pfunctor + 1;

	    pfunctor = strstr(pfunctor, SnapShotGet_Base);

	} /* endwhile */
    } /* endfor */
    delete pcp;

    if (Verbose)
    {
	fprintf(Log_fp,"done\n");
	fflush(Log_fp);
    } /* endif */

} /* end process_dependency_input() */

/****************************************************************
*
*
****************************************************************/

static void read_dependency_input(char *filename, Predicatelist *predicates)
{

char *buff;
int nbytes;
Predicate *pp;

FILE *fp;

    if (!filename)
    {
	fprintf(Log_fp, "ERROR: read_dependency_input() got NULL filename\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (!predicates)
    {
	fprintf(Log_fp, "ERROR: read_dependency_input() got NULL predicates\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if ((nbytes = BurtLongestLineLength(filename)) == UNKNOWN)
    {
	fprintf(Log_fp, 
    "ERROR: read_dependency_input() could not find BurtLongestLineLength()\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    /* for \0 */
    nbytes ++;

    if (!(buff = (char *) malloc(nbytes)))
    {
	fprintf(Log_fp, 
	    "ERROR: read_dependency_input() could not malloc %d bytes\n", 
	    nbytes);
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if ((fp = fopen(filename, "r")) == NULL)
    {
	 fprintf(Log_fp, "ERROR: read_dependency_input() could not open >%s<\n",
	    filename);
	 fflush(Log_fp);
	 BurtExit(1);
    } /* endif */

    while (fgets(buff, nbytes, fp))
    {
	if (*buff != '#' && *buff != '%' && *buff != '\n')
	{
	    buff[strlen(buff)-1] = '\0';
	    pp = new Predicate;

	    /* copying input line into predicate */

	    pp->set_raw_input_line(buff);
	    pp->set_input_line(buff);

	    predicates->append(pp);

	    if (Verbose)
	    {
		fprintf(Log_fp, ">%s<\n", pp->get_raw_input_line());
		fflush(Log_fp);
	    } /* endif */

	} /* endif */
    } /* endwhile */

    fclose(fp);
    free(buff);

} /* end read_dependency_input() */

/****************************************************************
*
*
****************************************************************/

static BOOL run_program()
{

int rc, signal;
long pid;
int status;

    if (Verbose)
    {
	fprintf(Log_fp, "\n--- start running program\n");
	fflush(Log_fp);
    } /* endif */

    Exec_args[2] = ExecutableFilename;

    pid = fork();

    if (pid < 0)
    {
	fprintf(Log_fp, 
	    "ERROR: run_program() illegal pid %d from fork()\n", pid);
	BurtExit(1);
    } /* endif */

    if (pid > 0)
    {
	/* I am parent */

	BurtPipeWait(&rc, &signal);

	if (Verbose)
	{
	    if (rc)
		fprintf(Log_fp, 
		    "--- done running program exit_status %d signal %d\n", 
			rc, signal);
	    else
		fprintf(Log_fp, "--- done running program\n");
	    fflush(Log_fp);
	} /* endif */

	return !rc;
    } 
    else
    {
	/* I am child */
	status = execv(Sh_path, Exec_args);
	exit(0);
    } /* endif */

} /* end run_program() */

/****************************************************************
*
*
****************************************************************/

static void write_epilogue()
{

    fprintf(Prog_fp, "    exit(0);\n\n");
    fprintf(Prog_fp, "} /* end main() */\n\n");

} /* end write_epilogue() */

/****************************************************************
*
*
****************************************************************/

static void write_predicates(Predicatelist *predicates)
{

Cursor<Predicate> *pcp;
Predicate *pp;
Cursor<Iocpvar> *icp;
Iocpvar *ip;
Cursor<Sspvar> *scp;
Sspvar *sp;
char *cp;
int predcount;
BOOL all_vals_present;

    if (!predicates)
    {
	fprintf(Log_fp, "ERROR: write_predicates() got NULL predicates\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    predcount = 1;

    fprintf(Prog_fp, "    /**************/\n");
    fprintf(Prog_fp, "    /*            */\n");
    fprintf(Prog_fp, "    /* predicates */\n");
    fprintf(Prog_fp, "    /*            */\n");
    fprintf(Prog_fp, "    /**************/\n\n");

    pcp = new Cursor<Predicate>(predicates);
    for (pp = pcp->getfirst(); pp; pp = pcp->getnext())
    {
	icp = new Cursor<Iocpvar>(pp->get_ioc_pvars_list());
	scp = new Cursor<Sspvar>(pp->get_ss_pvars_list());

	fprintf(Prog_fp, "    /* Predicate %d */\n", predcount);

	fprintf(Prog_fp, "    printf(\"");
	print_predicate_in_printf(pp->get_raw_input_line());
	fprintf(Prog_fp, " ... \");\n");
	    
	if (!((pp->get_ioc_pvars_list())->empty()) 
	    || !((pp->get_ss_pvars_list())->empty()))
	{
	    /* this predicate has ioc calls and/or ss calls    */
	    /* checking to see if there are ANY missing values */
	    /* from either.                                    */

	    all_vals_present = TRUE;

	    for (ip = icp->getfirst(); 
		all_vals_present && ip; 
		    ip = icp->getnext())
			all_vals_present = 
		(((ip->get_burtpvar())->get_readchannel())->get_val())->defined();

	    for (sp = scp->getfirst(); 
		all_vals_present && sp; 
		    sp = scp->getnext())
			all_vals_present = 
		    (sp->get_val() && (sp->get_val())->defined());

	    if (!all_vals_present)
	    {
		/* this predicate has missing ioc values ... */
		/* the "if" statement should not be written  */  
		fprintf(Prog_fp, 
		    "    printf(\"missing pvars not evaluated\\n\\n\");\n");
		switch (pp->get_directive_on_fail())
		{
		    case QUERYONFAIL:
			fprintf(Prog_fp, "    done = FALSE;\n");
			fprintf(Prog_fp, "    while (!done)\n");
			fprintf(Prog_fp, "    {\n");
			fprintf(Prog_fp, 
			    "\tprintf(\"Continue ? enter 'y' or 'n' \");\n");
			fprintf(Prog_fp, 
		    "\tif (gets(buff) && (*buff == 'y' || *buff == 'n'))\n");
			fprintf(Prog_fp, "    \tdone = TRUE;\n");
			fprintf(Prog_fp, "    } /* endwhile */\n");
			fprintf(Prog_fp, "    if (*buff == 'n')\n");
			fprintf(Prog_fp, "\texit(1);\n");
			fprintf(Prog_fp, "    printf(\"\\n\");\n");
			break;
		    case HALTONFAIL:
			fprintf(Prog_fp, 
			    "    printf(\"Halting on failure\\n\\n\");\n");
			fprintf(Prog_fp, "    exit(1);\n");
			break;
		    case CONTONFAIL:
			fprintf(Prog_fp, 
			    "    printf(\"Continuing on failure\\n\\n\");\n");
			break;
		    default:
			break;
		} /* end switch() */
		fprintf(Prog_fp, "\n");
	    } /* endif */
	} /* endif */

	if (!(!((pp->get_ioc_pvars_list())->empty())
	        || !((pp->get_ss_pvars_list())->empty()))
	    || all_vals_present)

	{
	    /* the "if" statement should be written */
	    fprintf(Prog_fp, "    if (");

	    cp = pp->get_start();
	    ip = icp->getfirst();
	    sp = scp->getfirst();

	    while (*cp)
	    {
		if (ip && cp == ip->get_start())
		{

		    /* came across an IOC pvar */

		    switch ((((ip->get_burtpvar())->get_readchannel())->
			get_val())->get_dbr_type())
		    {
			case DBR_STRING:
			    fprintf(Prog_fp, "\"%s\"", 
((char *) (((ip->get_burtpvar())->get_readchannel())->get_val())->get_pval()));
			    break;
			/* case DBR_INT: */
			case DBR_SHORT:
			    fprintf(Prog_fp, "%d", 
*((short *) (((ip->get_burtpvar())->get_readchannel())->get_val())->get_pval()));
			    break;
			case DBR_FLOAT:
			    fprintf(Prog_fp, FLOATFORMATSTRING, 
*((float *) (((ip->get_burtpvar())->get_readchannel())->get_val())->get_pval()));
			    break;
			case DBR_ENUM:
			    fprintf(Prog_fp, "%d", 
*((unsigned short *) (((ip->get_burtpvar())->get_readchannel())->get_val())->get_pval()));
			    break;
			case DBR_CHAR:
			    fprintf(Prog_fp, "'%c'", 
*((unsigned char *) (((ip->get_burtpvar())->get_readchannel())->get_val())->get_pval()));
			    break;
			case DBR_LONG:
			    fprintf(Prog_fp, "%ld", 
*((long *) (((ip->get_burtpvar())->get_readchannel())->get_val())->get_pval()));
			    break;
			case DBR_DOUBLE:
			    fprintf(Prog_fp, DOUBLEFORMATSTRING, 
*((double *) (((ip->get_burtpvar())->get_readchannel())->get_val())->get_pval()));
			    break;
			default:
			    fprintf(Log_fp, 
	    "ERROR: write_predicates() found unrecognizable ip dbr type %d\n",
	(((ip->get_burtpvar())->get_readchannel())->get_val())->get_dbr_type());
			    fflush(Log_fp);
			    BurtExit(1);
			    break;
		    } /* end switch() */

		    cp = ip->get_end() + 1;

		    ip = icp->getnext();
		}
		else if (sp && cp == sp->get_start())
		{
		    /* came across an snapshot pvar */

		    /* if (sp->get_print_type() == SS_STRING  */
			/* && (sp->get_val())->get_dbr_type() != DBR_STRING) */
			    /* fprintf(Prog_fp, "\""); */
		    if (sp->get_print_type() == SS_STRING)
			    fprintf(Prog_fp, "\"");

		    switch ((sp->get_val())->get_dbr_type())
		    {
			case DBR_STRING:
			    fprintf(Prog_fp, "%s", 
				((char *) (sp->get_val())->get_pval()));
			    break;
			/* case DBR_INT: */
			case DBR_SHORT:
			    fprintf(Prog_fp, "%d", 
				*((short *) (sp->get_val())->get_pval()));
			    break;
			case DBR_FLOAT:
			    fprintf(Prog_fp, FLOATFORMATSTRING, 
				*((float *) (sp->get_val())->get_pval()));
			    break;
			case DBR_ENUM:
			    fprintf(Prog_fp, "%d", 
			*((unsigned short *) (sp->get_val())->get_pval()));
			    break;
			case DBR_CHAR:
			    fprintf(Prog_fp, "%c", 
			*((unsigned char *) (sp->get_val())->get_pval()));
			    break;
			case DBR_LONG:
			    fprintf(Prog_fp, "%ld", 
				*((long *) (sp->get_val())->get_pval()));
			    break;
			case DBR_DOUBLE:
			    fprintf(Prog_fp, DOUBLEFORMATSTRING, 
				*((double *) (sp->get_val())->get_pval()));
			    break;
			default:
			    fprintf(Log_fp, 
	    "ERROR: write_predicates() found unrecognizable sp dbr type %d\n",
				(sp->get_val())->get_dbr_type());
			    fflush(Log_fp);
			    BurtExit(1);
			    break;
		    } /* end switch() */

		    /* if (sp->get_print_type() == SS_STRING  */
			/* && (sp->get_val())->get_dbr_type() != DBR_STRING) */
			    /* fprintf(Prog_fp, "\""); */
		    if (sp->get_print_type() == SS_STRING)
			    fprintf(Prog_fp, "\"");


		    cp = sp->get_end() + 1;

		    sp = scp->getnext();
		}
		else
		{
		    fprintf(Prog_fp, "%c", *cp);
		    cp ++;
		} /* endif */

	    } /* endwhile */

	    fprintf(Prog_fp, ")\n");

	    fprintf(Prog_fp, "\tprintf(\"passed\\n\");\n");
	    fprintf(Prog_fp, "    else\n");
	    fprintf(Prog_fp, "    {\n");
	    fprintf(Prog_fp, "\tprintf(\"failed\\n\");\n");
	    switch (pp->get_directive_on_fail())
	    {
		case QUERYONFAIL:
		    fprintf(Prog_fp, "\tdone = FALSE;\n");
		    fprintf(Prog_fp, "\twhile (!done)\n");
		    fprintf(Prog_fp, "\t{\n");
		    fprintf(Prog_fp, 
			"\t    printf(\"Continue ? enter 'y' or 'n' \");\n");
		    fprintf(Prog_fp, 
		"\t    if (gets(buff) && (*buff == 'y' || *buff == 'n'))\n");
		    fprintf(Prog_fp, "\t\tdone = TRUE;\n");
		    fprintf(Prog_fp, "\t} /* endwhile */\n");
		    fprintf(Prog_fp, "\tif (*buff == 'n')\n");
		    fprintf(Prog_fp, "\t    exit(1);\n");
		    fprintf(Prog_fp, "\tprintf(\"\\n\");\n");
		    break;
		case HALTONFAIL:
		    fprintf(Prog_fp, 
			"\tprintf(\"Halting on failure\\n\\n\");\n");
		    fprintf(Prog_fp, "\texit(1);\n");
		    break;
		case CONTONFAIL:
		    fprintf(Prog_fp, 
			"\tprintf(\"Continuing on failure\\n\\n\");\n");
		    break;
		default:
		    break;
	    } /* end switch() */
	    fprintf(Prog_fp, "    } /* endif */\n\n");
	} /* endif */

	fflush(Prog_fp);
	predcount ++;

	delete icp;
	delete scp;
    } /* endfor */
    delete pcp;

    fprintf(Prog_fp, "\n");
    fflush(Prog_fp);

} /* end write_predicates() */

/****************************************************************
*
*
****************************************************************/

static void write_program(Predicatelist *predicates)
{

    if (!predicates)
    {
	fprintf(Log_fp, "ERROR: write_program() got NULL predicates\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if (Verbose)
    {
	fprintf(Log_fp, "writing program ... ");
	fflush(Log_fp);
    } /* endif */

    if (!BurtCreateUniqueFilename(PUBLICWRITEDIRECTORY, ".c", SourceFilename))
    {
	fprintf(Log_fp,
	    "ERROR: write_program() could not BurtCreateUniqueFilename()\n");
	fflush(Log_fp);
	BurtExit(1);
    } /* endif */

    if ((Prog_fp = fopen(SourceFilename, "w")) == NULL)
    {
	fprintf(Log_fp, 
	    "ERROR: write_program() could not open program file >%s<\n",
		SourceFilename);
	fflush(Log_fp);
	BurtExit(1);
    } 
    else
    {
	write_prologue();
	write_predicates(predicates);
	write_epilogue();

	fclose(Prog_fp);

	if (Verbose)
	{
	    fprintf(Log_fp, "done\n");
	    fflush(Log_fp);
	} /* endif */

    } /* endif */

} /* end write_program() */

/****************************************************************
*
*
****************************************************************/

static void write_prologue()
{

    fprintf(Prog_fp, "#define BOOL  char\n");
    fprintf(Prog_fp, "#define FALSE 0\n");
    fprintf(Prog_fp, "#define TRUE  1\n\n");
 
    fprintf(Prog_fp, "main()\n");
    fprintf(Prog_fp, "{\n\n");

    fprintf(Prog_fp, "char buff[100];\n");
    fprintf(Prog_fp, "BOOL done;\n\n");

    fflush(Prog_fp);

} /* end write_prologue() */

