/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#include <stdlib.h>
#include "burtgooey.h"
#include "burtgooeyfuncdecls.h"

/* Externally visible local functions */
void main_definc_button_cb();

/* Private local functions */
static void setup_definc();
/* callbacks */
static void definc_apply_button_cb();
static void definc_reset_button_cb();

/* Local Global Variables */
static char *Definc_labels_strings[] = 
    {"Defines: -Dname{=def} ", "Includes: -Ipathname"};
static Widget Definc_shell;
static Widget Definc_text[XtNumber(Definc_labels_strings)];
static char Defines_backup[CMDSIZE];
static char Includes_backup[CMDSIZE];

/****************************************************************
*
*
****************************************************************/

void main_definc_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

    if (!Definc_setup)
    {
	setup_definc((Widget) client_data);
	Definc_setup = TRUE;
    } /* endif */

    strcpy(Defines_backup, Defines);
    XmTextFieldSetString(Definc_text[0], Defines);

    strcpy(Includes_backup, Includes);
    XmTextFieldSetString(Definc_text[1], Includes);

    XtMapWidget(Definc_shell);

} /* end main_definc_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void setup_definc(shell)
Widget shell;
{

Arg args[NARGS];
int nargs;
XmString xs;
int i;
Widget definc_form, definc_rc, definc_parms_rc;
Widget definc_labels[XtNumber(Definc_labels_strings)];
Widget definc_buttons_rc;
Widget definc_apply_button, definc_reset_button; 
Widget definc_help_button, definc_done_button; 

    /****************/
    /*              */
    /* definc shell */
    /*              */
    /****************/

    Definc_shell = XtVaCreateWidget("Defines/Includes", 
		    topLevelShellWidgetClass, shell,
		    XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
		    NULL);


    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 150); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    definc_form = XtCreateWidget("definc_form", xmFormWidgetClass, Definc_shell,
		    args, nargs);

    /* main definc rowcolumn */
    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    definc_rc = XmCreateRowColumn(definc_form, "definc_rc", args, nargs);

	definc_parms_rc = XmCreateRowColumn(definc_rc, "definc_parms_rc", 
			    NULL, 0);

	    for (i = 0; i < XtNumber(Definc_labels_strings); i ++)
	    {
		definc_labels[i] = XmCreateLabel(definc_parms_rc, 
				    Definc_labels_strings[i], NULL, 0);
		Definc_text[i] = XmCreateTextField(definc_parms_rc, 
				    "", NULL, 0);
	    } /* endfor */
	    XtManageChildren(definc_labels, XtNumber(Definc_labels_strings));
	    XtManageChildren(Definc_text, XtNumber(Definc_labels_strings));

	XtManageChild(definc_parms_rc);

	nargs = 0;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	definc_buttons_rc = XmCreateRowColumn(definc_rc, "definc_parms_rc", 
			    args, nargs);

	    definc_apply_button = XmCreatePushButton(definc_buttons_rc, 
				"definc_apply_button", NULL, 0);
	    XtManageChild(definc_apply_button);

	    definc_reset_button = XmCreatePushButton(definc_buttons_rc, 
				"definc_reset_button", NULL, 0);
	    XtManageChild(definc_reset_button);

	    definc_help_button = XmCreatePushButton(definc_buttons_rc,
		"definc_help_button", NULL, 0);
	    XtManageChild(definc_help_button);

	    definc_done_button = XmCreatePushButton(definc_buttons_rc, 
				    "definc_done_button", NULL, 0);
	    XtManageChild(definc_done_button);

	XtManageChild(definc_buttons_rc);

    XtManageChild(definc_rc);

    XtManageChild(definc_form);
    XtManageChild(Definc_shell);

    /*************/
    /*           */
    /* Callbacks */
    /*           */
    /*************/

    XtAddCallback(definc_apply_button, XmNactivateCallback,
        definc_apply_button_cb, NULL);
    XtAddCallback(definc_reset_button, XmNactivateCallback,
        definc_reset_button_cb, NULL);
    XtAddCallback(definc_help_button, XmNactivateCallback,
	shared_temp_help_button_cb, NULL);
    XtAddCallback(definc_done_button, XmNactivateCallback, 
	shared_done_button_cb, (Widget) Definc_shell);

} /* end setup_definc() */

/**************************/
/*                        */
/* definc shell callbacks */
/*                        */
/**************************/

/****************************************************************
*
*
****************************************************************/

static void definc_apply_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

char *buff;

    /* Defines */
    buff = XmTextFieldGetString(Definc_text[0]);
    if (strlen(buff) < sizeof(Defines))
	strcpy(Defines, buff);
    else
popup_error("ERROR: Defines specification too long.  No change applied to defines.");

    XtFree(buff);

    /* Includes */
    buff = XmTextFieldGetString(Definc_text[1]);
    if (strlen(buff) < sizeof(Includes))
	strcpy(Includes, buff);
    else
popup_error("ERROR: Includes specification too long.  No change applied to includes.");
    XtFree(buff);

} /* end definc_commit_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void definc_reset_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

    XmTextFieldSetString(Definc_text[0], Defines_backup);
    XmTextFieldSetString(Definc_text[1], Includes_backup);

} /* end definc_commit_button_cb() */

