#include "burtcommon.h"
#include "burtobjs.h"

#include "list.h"
#include "node.h"
#include "cursor.h"

/* Note: all lists are ORDERED sets ... i.e., it is possible to add an */
/*       element to the list at the beginning or end and it is also    */
/*       possible to reference its elements via an index.              */

/********/
/*      */
/* List */
/*      */
/********/

/****************************************************************
*
*
****************************************************************/

int List::nelem()
{

int rc;
Node *np;

    for (rc = 0, np = gethead(); np; np = np->getnext(), rc ++)
    ;

    return rc;
}

/**************/
/*            */
/* Sspvarlist */
/*            */
/**************/

Sspvarlist::~Sspvarlist()
{

Sspvarnode *np;

    while (head)
    {
	np = (Sspvarnode *) head;
	head = np->getnext();

	if (is_owner())
	    delete np->thisone();
	delete np;

    } /* endwhile */
}

void Sspvarlist::prepend(Sspvar *fp)
{

Sspvarnode *fnp;

    if (fp)
    {
	fnp = new Sspvarnode(fp);
	fnp->setnext((Sspvarnode *) head);
	head = (Node *) fnp;
	if (!tail)
	    tail = head;
    } /* endif */

} 

void Sspvarlist::append(Sspvar *fp)
{

Sspvarnode *fnp;

    if (fp)
    {
	fnp = new Sspvarnode(fp);
	fnp->setnext((Sspvarnode *) NULL);
	if (tail)
	    ((Sspvarnode *) tail)->setnext(fnp);
	tail = (Node *) fnp;
	if (!head)
	    head = tail;
    } /* endif */

} 

Sspvar *Sspvarlist::remove(Sspvar *fp)
{

Sspvar *rc;
Sspvarnode *fnp;
Sspvarnode *temp;

    if (fnp = (Sspvarnode *) head)
    {
	/* non-empty list */
	if (fnp->thisone() == fp)
	{
	    /* one to remove is at the head of the list */
	    if (!(head = fnp->getnext()))
		/* emptied list */
		tail = (Node *) NULL;

	    rc = fnp->thisone();
	    delete fnp;
	}
	else
	{
	    /* one to remove is not at the head of the list */
	    while (fnp->getnext() 
		&& ((Sspvarnode *)(fnp->getnext()))->thisone() != fp)
		    fnp = (Sspvarnode *) fnp->getnext();
	    ;

	    if (temp = (Sspvarnode *) fnp->getnext())
	    {
		/* we found our guy, temp points to it and fnp precedes it */
		fnp->setnext((Sspvarnode *) temp->getnext());

		if (!fnp->getnext())
		    /* deleting last off list ... known not to be head */
		    tail = (Node *) fnp;

		rc = temp->thisone();
		delete temp;
	    }
	    else
		rc = (Sspvar *) NULL;
	} /* endif */
    } 
    else
	rc = (Sspvar *) NULL;

    return rc;

}

Sspvar* Sspvarlist::operator[](int index)
{
Sspvarnode* fn;
int i;

    for(i=0,fn=(Sspvarnode*)gethead();
	fn && i<index;
	    fn=(Sspvarnode*)fn->getnext(),i++)
    ;

    if(!fn) 
	return (Sspvar*)NULL;

    return fn->thisone();
}

/***************/
/*             */
/* Burtobjlist */
/*             */
/***************/

Burtobjlist::~Burtobjlist()
{

Burtobjnode *np;

    while (head)
    {
	np = (Burtobjnode *) head;
	head = np->getnext();

	if (is_owner())
	    delete np->thisone();
	delete np;

    } /* endwhile */
}

void Burtobjlist::prepend(Burtobj *fp)
{

Burtobjnode *fnp;

    if (fp)
    {
	fnp = new Burtobjnode(fp);
	fnp->setnext((Burtobjnode *) head);
	head = (Node *) fnp;
	if (!tail)
	    tail = head;
    } /* endif */

} 

void Burtobjlist::append(Burtobj *fp)
{

Burtobjnode *fnp;

    if (fp)
    {
	fnp = new Burtobjnode(fp);
	fnp->setnext((Burtobjnode *) NULL);
	if (tail)
	    ((Burtobjnode *) tail)->setnext(fnp);
	tail = (Node *) fnp;
	if (!head)
	    head = tail;
    } /* endif */

} 

Burtobj *Burtobjlist::remove(Burtobj *fp)
{

Burtobj *rc;
Burtobjnode *fnp;
Burtobjnode *temp;

    if (fnp = (Burtobjnode *) head)
    {
	/* non-empty list */
	if (fnp->thisone() == fp)
	{
	    /* one to remove is at the head of the list */
	    if (!(head = fnp->getnext()))
		/* emptied list */
		tail = (Node *) NULL;

	    rc = fnp->thisone();
	    delete fnp;
	}
	else
	{
	    /* one to remove is not at the head of the list */
	    while (fnp->getnext() 
		&& ((Burtobjnode *)(fnp->getnext()))->thisone() != fp)
		    fnp = (Burtobjnode *) fnp->getnext();
	    ;

	    if (temp = (Burtobjnode *) fnp->getnext())
	    {
		/* we found our guy, temp points */
		/* to it and fnp precedes it */
		fnp->setnext((Burtobjnode *) temp->getnext());

		if (!fnp->getnext())
		    /* deleting last off list ... known not to be head */
		    tail = (Node *) fnp;

		rc = temp->thisone();
		delete temp;
	    }
	    else
		rc = (Burtobj *) NULL;
	} /* endif */
    } 
    else
	rc = (Burtobj *) NULL;

    return rc;

}

Burtobj* Burtobjlist::operator[](int index)
{
Burtobjnode* fn;
int i;

    for(i=0,fn=(Burtobjnode*)gethead();
	fn && i<index;
	    fn=(Burtobjnode*)fn->getnext(),i++)
    ;

    if(!fn) 
	return (Burtobj *) NULL;

    return fn->thisone();
}

/*******************/
/*                 */
/* Burtobjlistlist */
/*                 */
/*******************/

Burtobjlistlist::~Burtobjlistlist()
{

Burtobjlistnode *np;

    while (head)
    {
	np = (Burtobjlistnode *) head;
	head = np->getnext();

	if (is_owner())
	    delete np->thisone();
	delete np;

    } /* endwhile */
}

void Burtobjlistlist::prepend(Burtobjlist *fp)
{

Burtobjlistnode *fnp;

    if (fp)
    {
	fnp = new Burtobjlistnode(fp);
	fnp->setnext((Burtobjlistnode *) head);
	head = (Node *) fnp;
	if (!tail)
	    tail = head;
    } /* endif */

} 

void Burtobjlistlist::append(Burtobjlist *fp)
{

Burtobjlistnode *fnp;

    if (fp)
    {
	fnp = new Burtobjlistnode(fp);
	fnp->setnext((Burtobjlistnode *) NULL);
	if (tail)
	    ((Burtobjlistnode *) tail)->setnext(fnp);
	tail = (Node *) fnp;
	if (!head)
	    head = tail;
    } /* endif */

} 

Burtobjlist *Burtobjlistlist::remove(Burtobjlist *fp)
{

Burtobjlist *rc;
Burtobjlistnode *fnp;
Burtobjlistnode *temp;

    if (fnp = (Burtobjlistnode *) head)
    {
	/* non-empty list */
	if (fnp->thisone() == fp)
	{
	    /* one to remove is at the head of the list */
	    if (!(head = fnp->getnext()))
		/* emptied list */
		tail = (Node *) NULL;

	    rc = fnp->thisone();
	    delete fnp;
	}
	else
	{
	    /* one to remove is not at the head of the list */
	    while (fnp->getnext() 
		&& ((Burtobjlistnode *)(fnp->getnext()))->thisone() != fp)
		    fnp = (Burtobjlistnode *) fnp->getnext();
	    ;

	    if (temp = (Burtobjlistnode *) fnp->getnext())
	    {
		/* we found our guy, now temp points */
		/* to it and fnp precedes it */
		fnp->setnext((Burtobjlistnode *) temp->getnext());

		if (!fnp->getnext())
		    /* deleting last off list ... known not to be head */
		    tail = (Node *) fnp;

		rc = temp->thisone();
		delete temp;
	    }
	    else
		rc = (Burtobjlist *) NULL;
	} /* endif */
    } 
    else
	rc = (Burtobjlist *) NULL;

    return rc;

}

Burtobjlist* Burtobjlistlist::operator[](int index)
{
Burtobjlistnode* fn;
int i;

    for(i=0,fn=(Burtobjlistnode*)gethead();
	fn && i<index;
	    fn=(Burtobjlistnode*)fn->getnext(),i++)
    ;

    if(!fn) 
	return (Burtobjlist *) NULL;

    return fn->thisone();
}

/*****************/
/*               */
/* Predicatelist */
/*               */
/*****************/

Predicatelist::~Predicatelist()
{

Predicatenode *np;

    while (head)
    {
	np = (Predicatenode *) head;
	head = np->getnext();

	if (is_owner())
	    delete np->thisone();
	delete np;

    } /* endwhile */
}

void Predicatelist::prepend(Predicate *fp)
{

Predicatenode *fnp;

    if (fp)
    {
	fnp = new Predicatenode(fp);
	fnp->setnext((Predicatenode *) head);
	head = (Node *) fnp;
	if (!tail)
	    tail = head;
    } /* endif */

} 

void Predicatelist::append(Predicate *fp)
{

Predicatenode *fnp;

    if (fp)
    {
	fnp = new Predicatenode(fp);
	fnp->setnext((Predicatenode *) NULL);
	if (tail)
	    ((Predicatenode *) tail)->setnext(fnp);
	tail = (Node *) fnp;
	if (!head)
	    head = tail;
    } /* endif */

} 

Predicate *Predicatelist::remove(Predicate *fp)
{

Predicate *rc;
Predicatenode *fnp;
Predicatenode *temp;

    if (fnp = (Predicatenode *) head)
    {
	/* non-empty list */
	if (fnp->thisone() == fp)
	{
	    /* one to remove is at the head of the list */
	    if (!(head = fnp->getnext()))
		/* emptied list */
		tail = (Node *) NULL;

	    rc = fnp->thisone();
	    delete fnp;
	}
	else
	{
	    /* one to remove is not at the head of the list */
	    while (fnp->getnext() 
		&& ((Predicatenode *)(fnp->getnext()))->thisone() != fp)
		    fnp = (Predicatenode *) fnp->getnext();
	    ;

	    if (temp = (Predicatenode *) fnp->getnext())
	    {
		/* we found our guy, temp points to it and fnp precedes it */
		fnp->setnext((Predicatenode *) temp->getnext());

		if (!fnp->getnext())
		    /* deleting last off list ... known not to be head */
		    tail = (Node *) fnp;

		rc = temp->thisone();
		delete temp;
	    }
	    else
		rc = (Predicate *) NULL;
	} /* endif */
    } 
    else
	rc = (Predicate *) NULL;

    return rc;

}

Predicate* Predicatelist::operator[](int index)
{
Predicatenode* fn;
int i;

    for(i=0,fn=(Predicatenode*)gethead();
	fn && i<index;
	    fn=(Predicatenode*)fn->getnext(),i++)
    ;

    if(!fn) 
	return (Predicate*)NULL;

    return fn->thisone();
}

/***************/
/*             */
/* Iocpvarlist */
/*             */
/***************/

Iocpvarlist::~Iocpvarlist()
{

Iocpvarnode *np;

    while (head)
    {
	np = (Iocpvarnode *) head;
	head = np->getnext();

	if (is_owner())
	    delete np->thisone();
	delete np;

    } /* endwhile */
}

void Iocpvarlist::prepend(Iocpvar *fp)
{

Iocpvarnode *fnp;

    if (fp)
    {
	fnp = new Iocpvarnode(fp);
	fnp->setnext((Iocpvarnode *) head);
	head = (Node *) fnp;
	if (!tail)
	    tail = head;
    } /* endif */

} 

void Iocpvarlist::append(Iocpvar *fp)
{

Iocpvarnode *fnp;

    if (fp)
    {
	fnp = new Iocpvarnode(fp);
	fnp->setnext((Iocpvarnode *) NULL);
	if (tail)
	    ((Iocpvarnode *) tail)->setnext(fnp);
	tail = (Node *) fnp;
	if (!head)
	    head = tail;
    } /* endif */

} 

Iocpvar *Iocpvarlist::remove(Iocpvar *fp)
{

Iocpvar *rc;
Iocpvarnode *fnp;
Iocpvarnode *temp;

    if (fnp = (Iocpvarnode *) head)
    {
	/* non-empty list */
	if (fnp->thisone() == fp)
	{
	    /* one to remove is at the head of the list */
	    if (!(head = fnp->getnext()))
		/* emptied list */
		tail = (Node *) NULL;

	    rc = fnp->thisone();
	    delete fnp;
	}
	else
	{
	    /* one to remove is not at the head of the list */
	    while (fnp->getnext() 
		&& ((Iocpvarnode *)(fnp->getnext()))->thisone() != fp)
		    fnp = (Iocpvarnode *) fnp->getnext();
	    ;

	    if (temp = (Iocpvarnode *) fnp->getnext())
	    {
		/* we found our guy, temp points to it and fnp precedes it */
		fnp->setnext((Iocpvarnode *) temp->getnext());

		if (!fnp->getnext())
		    /* deleting last off list ... known not to be head */
		    tail = (Node *) fnp;

		rc = temp->thisone();
		delete temp;
	    }
	    else
		rc = (Iocpvar *) NULL;
	} /* endif */
    } 
    else
	rc = (Iocpvar *) NULL;

    return rc;

}

Iocpvar* Iocpvarlist::operator[](int index)
{
Iocpvarnode* fn;
int i;

    for(i=0,fn=(Iocpvarnode*)gethead();
	fn && i<index;
	    fn=(Iocpvarnode*)fn->getnext(),i++)
    ;

    if(!fn) 
	return (Iocpvar*)NULL;

    return fn->thisone();
}

