/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#include "burtgooey.h"
#include "burtgooeyfuncdecls.h"
#include <unistd.h>

#define MATHOP3VIEWSNAPBUTTONNAME "math_op3_view_snap_button"

#define MATHOP3PRINTSNAPBUTTONNAME "math_op3_print_snap_button"

/* Externally visible functions */
void main_mult_button_cb();
void cleanup_mult();

/* Private local functions */
static void setup_math();
/* callbacks */
static void math_op3_snap_file_sel_button_cb();
static void math_view_snap_button_cb();
static void math_print_snap_button_cb();
static void math_go_button2_cb();
static void math_view_log_button2_cb();
static void math_file_sel_ok_cb();
static void math_operation_toggle_cb();
static void math_sdds_toggle2_cb();

/* Local Global Variables */
static char *Math_labels_strings[] =
    {"Snapshot Filename:", "Comments:", "Keywords:"};
static Widget Math_shell;
static Widget Math_log2_shell;
static Widget Math_op3_text;
static Widget Math_op3_view_snap_scroll_txt;
static Widget Math_op3_view_snap_shell;
static Widget Math_log2_scroll_txt;
static Widget Math_scalar_multiplier_text;
static Widget Math2_text[XtNumber(Math_labels_strings)];
static Widget Math_status2_label; 
static char Math_logfilename[MAXPATHLEN];
static char Math_sdds_output_spec2;


/****************************************************************
*
*
****************************************************************/

void main_mult_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

static char default_snapshot_filename[MAXPATHLEN];

    if (!Mult_setup)
    {
	setup_math((Widget) client_data);
	Mult_setup = TRUE;
    } /* endif */

    if (BurtCreateUniqueFilename((char *) NULL, ".snap", 
	default_snapshot_filename))
    {
	XmTextFieldSetString(Math2_text[0], default_snapshot_filename);
	XtMapWidget(Math_shell);
    }
    else
	popup_message("SYS FAULT: could not create default_snapshot_filename.");

} /* end main_math_button_cb() */

/****************************************************************
*
*
****************************************************************/

void cleanup_mult()
{
    unlink(Math_logfilename);
} /* end cleanup_math() */

/****************************************************************
*
*
****************************************************************/

static void setup_math(shell)
Widget shell;
{

Arg args[NARGS];
int nargs;
XmString xs;
int i;
Widget math_form, math_rc;
Widget math_equal2_label;

Widget math_scalar_multiplication_label;

Widget math_op3_rc;
Widget math_op3_snap_files_buttons_rc;
Widget math_op3_snap_file_sel_button, math_op3_view_snap_button;
Widget math_op3_print_snap_button;
Widget math_op3_view_snap_form, math_op3_view_snap_rc, math_op3_view_snap_done_button;
Widget math_scalar_rc, math_times_label, math_scalar_multiplier_label;

Widget math_log2_form, math_log2_rc, math_log_done2_button;

Widget math_sdds2_rc, math_sdds2_toggle[XtNumber(Sdds_labels_strings)];
Widget math2_labels[XtNumber(Math_labels_strings)];
Widget math_op3_snap_file_sel, math_buttons2_rc, math_go2_button;
Widget math_view_log2_button, math_print_log2_button, math_help2_button;
Widget math_done2_button;

    if (!BurtCreateUniqueFilename(PUBLICWRITEDIRECTORY, ".mult.log",
            Math_logfilename))
    {
        fprintf(Log_fp,
            "ERROR: unable to create a unique filename for mult log\n");
        exit(1);
    } /* endif */

    /**************/
    /*            */
    /* math shell */
    /*            */
    /**************/

    Math_shell = XtVaCreateWidget("Scalar Multiplication of Snapshot Files", 
		    topLevelShellWidgetClass, shell,
		    XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
		    NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 400); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 800); nargs ++;
    math_form = XtCreateWidget("math_form", xmFormWidgetClass, Math_shell, 
		    args, nargs);

    /* math popup snapshot file selection boxes */
    xs = XmStringCreateSimple("Snapshot File Selector");
    nargs = 0;
    XtSetArg(args[nargs], XmNdialogTitle, xs); nargs ++;
    math_op3_snap_file_sel = XmCreateFileSelectionDialog(math_form, 
				"math_op3_snap_file_sel", args, nargs);
    XtUnmanageChild(XmFileSelectionBoxGetChild(math_op3_snap_file_sel, 
	XmDIALOG_HELP_BUTTON)); /* removing HELP button */
    XmStringFree(xs);

    /* main set rowcolumn */
    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    math_rc = XmCreateRowColumn(math_form, "math_rc", args, nargs);

	/* op3 */
	math_op3_rc = XmCreateRowColumn(math_rc, "math_op3_rc", NULL, 0);
	    Math_op3_text = XmCreateTextField(math_op3_rc, 
			    "Math_op3_text", 
			    NULL, 0);
	    XtManageChild(Math_op3_text);

	    /* math op3 snapshot file buttons rowcolumn */
	    nargs = 0;
	    XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	    math_op3_snap_files_buttons_rc = XmCreateRowColumn(math_op3_rc, 
					"math_op3_snap_files_buttons_rc", 
					args, nargs);

		math_op3_snap_file_sel_button =  
		    XmCreatePushButton(math_op3_snap_files_buttons_rc, 
			"math_op3_snap_file_sel_button", 
			NULL, 0);
		XtManageChild(math_op3_snap_file_sel_button);

		math_op3_view_snap_button =  
		    XmCreatePushButton(math_op3_snap_files_buttons_rc, 
			MATHOP3VIEWSNAPBUTTONNAME, 
			NULL, 0);
		XtManageChild(math_op3_view_snap_button);

		math_op3_print_snap_button =  
		    XmCreatePushButton(math_op3_snap_files_buttons_rc, 
			MATHOP3PRINTSNAPBUTTONNAME, 
			NULL, 0);
		XtManageChild(math_op3_print_snap_button);

	    XtManageChild(math_op3_snap_files_buttons_rc);
	XtManageChild(math_op3_rc);

	nargs = 0;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	math_scalar_rc = XmCreateRowColumn(math_rc, 
			    "math_scalar_rc", 
			    args, nargs);

	    /* times label */
	    math_times_label = XmCreateLabel(math_scalar_rc, 
				"math_times_label", NULL, 0);
	    xs = XmStringCreateSimple("*");
	    XtVaSetValues(math_times_label,
		XmNlabelString, xs,
		NULL);
	    XmStringFree(xs);
	    XtManageChild(math_times_label);

	    Math_scalar_multiplier_text = XmCreateTextField(math_scalar_rc, 
			    "Math_scalar_multiplier_text", 
			    NULL, 0);
	    XtManageChild(Math_scalar_multiplier_text);
	    XmTextFieldSetString(Math_scalar_multiplier_text, "1.0");

	    /* times label */
	    math_scalar_multiplier_label = XmCreateLabel(math_scalar_rc, 
				"math_scalar_multiplier_label", NULL, 0);
	    xs = XmStringCreateSimple("Scalar Multiplier");
	    XtVaSetValues(math_scalar_multiplier_label,
		XmNlabelString, xs,
		NULL);
	    XmStringFree(xs);
	    XtManageChild(math_scalar_multiplier_label);

	XtManageChild(math_scalar_rc);

	/* = label */
	math_equal2_label = XmCreateLabel(math_rc, "math_equal2_label", 
		    NULL, 0);
	xs = XmStringCreateSimple("=");
	XtVaSetValues(math_equal2_label,
	    XmNlabelString, xs,
	    NULL);
	XmStringFree(xs);
	XtManageChild(math_equal2_label);

	for (i = 0; i < XtNumber(Math_labels_strings); i ++)
	{
	    math2_labels[i] = XmCreateLabel(math_rc, Math_labels_strings[i], 
				NULL, 0);
	    Math2_text[i] = XmCreateTextField(math_rc, "", NULL, 0);
	} /* endfor */
	XtManageChildren(math2_labels, XtNumber(Math_labels_strings));
	XtManageChildren(Math2_text, XtNumber(Math_labels_strings));

	/* math sdds spec */
	/* from page 414 AT&T Motif Course Manual */
	nargs = 0;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	math_sdds2_rc = XmCreateRadioBox(math_rc, "math_sdds2_rc", args, nargs);

	    for (i = 0; i < XtNumber(Sdds_labels_strings); i ++)
		math_sdds2_toggle[i] = XmCreateToggleButton(math_sdds2_rc, 
				    Sdds_labels_strings[i], NULL, 0);

	    /* setting initial radio box toggle state */
	    XmToggleButtonSetState(math_sdds2_toggle[NOREQ], True, False);
	    Math_sdds_output_spec2 = NOREQ;

	    XtManageChildren(math_sdds2_toggle, XtNumber(Sdds_labels_strings));

	XtManageChild(math_sdds2_rc);

	/* math buttons rowcolumn */
	nargs = 0;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	math_buttons2_rc = XmCreateRowColumn(math_rc, 
		"math_buttons2_rc", 
		args, nargs);

	    math_go2_button = XmCreatePushButton(math_buttons2_rc, 
				"math_go2_button", NULL, 0);
	    XtManageChild(math_go2_button);

	    Math_status2_label = 
		XmCreateLabel(math_buttons2_rc, "Math_status2_label", NULL, 0);

	    set_ok(Math_status2_label);
	    XtManageChild(Math_status2_label);

	    math_view_log2_button = XmCreatePushButton(math_buttons2_rc, 
				    "math_view_log2_button", NULL, 0);
	    XtManageChild(math_view_log2_button);

	    math_print_log2_button = XmCreatePushButton(math_buttons2_rc, 
				    "math_print_log2_button", NULL, 0);
	    XtManageChild(math_print_log2_button);

	    math_help2_button = XmCreatePushButton(math_buttons2_rc, 
				"math_help2_button", NULL, 0);
	    XtManageChild(math_help2_button);

	    math_done2_button = XmCreatePushButton(math_buttons2_rc, 
				"math_done2_button", NULL, 0);
	    XtManageChild(math_done2_button);

	XtManageChild(math_buttons2_rc);

    XtManageChild(math_rc);

    /* math view op3 snap shell */
    Math_op3_view_snap_shell = XtVaCreateWidget("Snapshot File Viewer", 
	    topLevelShellWidgetClass, Math_shell,
	    XmNmappedWhenManaged, False,
	    XmNmwmDecorations, 
		(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
	    NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 350); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    math_op3_view_snap_form = XtCreateWidget("math_op3_view_snap_form", 
				xmFormWidgetClass, Math_op3_view_snap_shell,
				args, nargs);

    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    math_op3_view_snap_rc = XmCreateRowColumn(math_op3_view_snap_form, 
			    "math_op3_view_snap_rc", args, nargs);

	nargs = 0;
	XtSetArg(args[nargs], XmNeditMode, XmMULTI_LINE_EDIT); nargs ++;
	XtSetArg(args[nargs], XmNeditable, False); nargs ++;
	XtSetArg(args[nargs], XmNheight, 300); nargs ++;
	Math_op3_view_snap_scroll_txt = 
	    XmCreateScrolledText(math_op3_view_snap_rc, 
		"math_op3_view_math_scroll_txt", args, nargs);
	XtManageChild(Math_op3_view_snap_scroll_txt);

	math_op3_view_snap_done_button = 
	    XmCreatePushButton(math_op3_view_snap_rc, 
		"math_op3_view_snap_done_button", NULL, 0);
	XtManageChild(math_op3_view_snap_done_button);

    XtManageChild(math_op3_view_snap_rc);
    XtManageChild(math_op3_view_snap_form);
    XtManageChild(Math_op3_view_snap_shell);

    /* math log2 shell */
    Math_log2_shell = XtVaCreateWidget("Scalar Multiplication Log Viewer", 
			topLevelShellWidgetClass, Math_shell,
			XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
			NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 350); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    math_log2_form = XtCreateWidget("math_log2_form", xmFormWidgetClass, 
		    Math_log2_shell, 
		    args, nargs);

    /* main math log rowcolumn */
    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    math_log2_rc = XmCreateRowColumn(math_log2_form, "math_log2_rc", 
		    args, nargs);

	nargs = 0;
	XtSetArg(args[nargs], XmNeditMode, XmMULTI_LINE_EDIT); nargs ++;
	XtSetArg(args[nargs], XmNeditable, False); nargs ++;
	XtSetArg(args[nargs], XmNheight, 300); nargs ++;
	Math_log2_scroll_txt = XmCreateScrolledText(math_log2_rc, 
				"Math_log2_scroll_txt", args, nargs);
	XtManageChild(Math_log2_scroll_txt);

	math_log_done2_button = XmCreatePushButton(math_log2_rc, 
				"math_log_done2_button", 
				NULL, 0);
	XtManageChild(math_log_done2_button);

    XtManageChild(math_log2_rc);
    XtManageChild(math_log2_form);
    XtManageChild(Math_log2_shell);

    XtManageChild(math_form);
    XtManageChild(Math_shell);

    /*************/
    /*           */
    /* Callbacks */
    /*           */
    /*************/

    XtAddCallback(math_go2_button, XmNactivateCallback, math_go_button2_cb, 
	NULL);
    XtAddCallback(math_view_log2_button, XmNactivateCallback, 
	math_view_log_button2_cb, NULL);
    XtAddCallback(math_print_log2_button, XmNactivateCallback, 
	shared_print_log_button_cb, (char *) Math_logfilename);
    XtAddCallback(math_help2_button, XmNactivateCallback, 
	shared_temp_help_button_cb, NULL);
    XtAddCallback(math_done2_button, XmNactivateCallback, 
	shared_done_button_cb, (Widget) Math_shell);
    XtAddCallback(math_op3_snap_file_sel_button, XmNactivateCallback, 
	math_op3_snap_file_sel_button_cb, (Widget) math_op3_snap_file_sel);
    XtAddCallback(math_op3_view_snap_button, XmNactivateCallback, 
	math_view_snap_button_cb, NULL);
    XtAddCallback(math_op3_print_snap_button, XmNactivateCallback, 
	math_print_snap_button_cb, NULL);
    XtAddCallback(math_op3_snap_file_sel, XmNcancelCallback, 
	XtUnmanageChild, NULL);
    XtAddCallback(math_op3_snap_file_sel, XmNokCallback, math_file_sel_ok_cb, 
	(Widget) Math_op3_text);
    XtAddCallback(math_op3_view_snap_done_button, XmNactivateCallback, 
	shared_done_button_cb, (Widget) Math_op3_view_snap_shell);
    /* math log2 shell */
    XtAddCallback(math_log_done2_button, XmNactivateCallback, 
	shared_done_button_cb, (Widget) Math_log2_shell);
    for (i = 0; i < XtNumber(Sdds_labels_strings); i ++)
	XtAddCallback(math_sdds2_toggle[i], XmNvalueChangedCallback,
	    math_sdds_toggle2_cb, (int) i);

} /* end setup_math() */

/************************/
/*                      */
/* math shell callbacks */
/*                      */
/************************/

/****************************************************************
*
*
****************************************************************/

static void math_op3_snap_file_sel_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

    XtManageChild((Widget) client_data);
    XtPopup(XtParent((Widget) client_data), XtGrabNone);

} /* end math_op3_snap_file_sel_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void math_view_snap_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

XmStringTable choices;
int n_choices;
int i, j;
int total_bytes;
char *total_buff;
int fd;
char mbuff[1000];
struct filerep fbuff;
char *cp;
char *fname;
struct stat stat_buff;
BOOL all_ascii;

    XtVaSetValues(Math_op3_view_snap_scroll_txt,
	XmNvalue, "",
	NULL);
    fname = XmTextFieldGetString(Math_op3_text);

    if (*fname)
    {
	if ((fd = open(fname, O_RDONLY)) != -1)
	{
	    if (!fstat(fd, &stat_buff))
	    {
		if (S_ISREG(stat_buff.st_mode))
		{
		    if (fbuff.buff = (char *) calloc (stat_buff.st_size,1))
		    {
			if (read(fd, fbuff.buff, stat_buff.st_size) != -1)
			{
			    /* loop to check file is printable */
			    for (j = 0, all_ascii = TRUE; 
				all_ascii && j < stat_buff.st_size; 
				    j ++)
					all_ascii = isascii(fbuff.buff[j]);

			    if (all_ascii)
			    {
				fbuff.nbytes = stat_buff.st_size;
				fbuff.fname = strdup(fname);

				/* now filling scrolled text */
				total_bytes = 
			(fbuff.nbytes+strlen(fbuff.fname)+1+strlen("\n\n"));
				/* for null char at end */
				total_bytes ++;

				if (total_bytes > 1)
				{
				    /* there are files to view */
				    if (total_buff = 
					    (char *) calloc(total_bytes, 1))
				    {
					cp = total_buff;

					/* appending (via copy) file to */
					/* total_buff and then free mem */
					/* from fbuff */
					memcpy(cp, fbuff.fname,
					    strlen(fbuff.fname));
					cp = cp + strlen(fbuff.fname);

					*cp = '\n';
					cp ++;

					memcpy(cp, fbuff.buff, fbuff.nbytes);
					cp = cp + fbuff.nbytes;

					memcpy(cp, "\n\n", strlen("\n\n"));
					cp = cp + strlen("\n\n");

					free(fbuff.buff);
					free(fbuff.fname);

					/* although buffer was allocated with */
					/* calloc which presumably allocates */
					/* a block with all zeros, still     */ 
					/* need to explicitly place the null */
					/* char at the end  so as not to get */
					/* extra garbage characters in the   */
					/* scroll text ... can't explain why */
					/* this works ... just happy that it */
					/* does.                             */

					*cp = '\0';

				    XtVaSetValues(Math_op3_view_snap_scroll_txt,
						XmNvalue, total_buff,
						NULL);

					free(total_buff);
				    }
				    else
				    {
					sprintf(mbuff, 
					"SYS FAULT: failed calloc() %d bytes.", 
					    total_bytes);
					popup_message(mbuff);
				    } /* endif */
				} /* endif */
			    }
			    else
			    {
				free(fbuff.buff);

				sprintf(mbuff, "ERROR: file %s not viewable.",
				    fname);
				popup_error(mbuff);
			    } /* endif */
			}
			else
			{
			    free(fbuff.buff);

			    sprintf(mbuff, 
				"SYS FAULT: could not read() file %s.", fname);
			    popup_message(mbuff);
			} /* endif */

		    }
		    else
		    {
			sprintf(mbuff, "SYS FAULT: failed calloc() %d bytes.", 
			    stat_buff.st_size);
			popup_message(mbuff);
		    } /* endif */
		}
		else
		{
		    sprintf(mbuff, 
		"NOTE: cannot view file %s because it is not a regular file.", 
			fname);
		popup_message(mbuff);
		} /* endif */
	    }
	    else
		popup_message("SYS FAULT: error in fstat().");

	    close(fd);
	}
	else
	{
	    sprintf(mbuff, "SYS FAULT: cannot open() file %s.", fname);
	    popup_message(mbuff);
	} /* endif */
    } /* endif */

    XtFree(fname);

    XtMapWidget(Math_op3_view_snap_shell);

} /* end math_view_snap_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void math_print_snap_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

char *fname;

    fname = XmTextFieldGetString(Math_op3_text);

    print_file(fname);
    XtFree(fname);

} /* end math_print_snap_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void math_go_button2_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{


int rc;
char *cp;
char command[CMDSIZE];
char *op1_fname, *mult_factor, *result_fname;
char *ret_cp;
BOOL ok;

    ok = TRUE;

    op1_fname = XmTextFieldGetString(Math_op3_text);
    mult_factor = XmTextFieldGetString(Math_scalar_multiplier_text);
    result_fname = XmTextFieldGetString(Math2_text[0]);

    /* checking if a number can be extracted from mult_factor */

    /* NOTE: this is the same method used in burtmath.  it also    */
    /*       converts as far as it can into the string and ignores */
    /*       extraneous characters at the end.                     */

    /* strtod() sets ret_cp = mult_factor upon failure */
    strtod(mult_factor, &ret_cp);

    if (*op1_fname && *result_fname && ret_cp != mult_factor)
    {
	if (strlen("burtmath") < CMDSIZE)
	    strcpy(command, "burtmath");
	else
	    ok = FALSE;

	if (ok &&
	    (((strlen(command) + 
		1+strlen(op1_fname) 
		+strlen(" -mult")+ 1+strlen(mult_factor)
		+strlen(" -o")+ 1+strlen(result_fname)
		+strlen(" -v")
		+strlen(" -l" + 1+strlen(Math_logfilename)))+1) < CMDSIZE))
	{
	    strcat(command, " ");
	    strcat(command, op1_fname);

	    strcat(command, " -mult");
	    strcat(command, " ");
	    strcat(command, mult_factor);

	    strcat(command, " -o");
	    strcat(command, " ");
	    strcat(command, result_fname);

	    strcat(command, " -v");

	    unlink(Math_logfilename);
	    strcat(command, " -l");
	    strcat(command, " ");
	    strcat(command, Math_logfilename);

	} 
	else
	    ok = FALSE;

        cp = XmTextFieldGetString(Math2_text[1]);
        if (*cp)
        {
            if (ok && (strlen(command) + 4+strlen(cp)) < CMDSIZE)
            {
                strcat(command, " -c \"");
                strcat(command, cp);
                strcat(command, "\"");
            }
            else
                ok = FALSE;
        } /* endif */
        XtFree(cp);

        cp = XmTextFieldGetString(Math2_text[2]);
        if (*cp)
        {
            if (ok && (strlen(command) + 4+strlen(cp)) < CMDSIZE)
            {
                strcat(command, " -k \"");
                strcat(command, cp);
                strcat(command, "\"");
            }
            else
                ok = FALSE;
        } /* endif */
        XtFree(cp);

        if (ok)
        {
            if (Math_sdds_output_spec2 == REQSDDS)
            {
                if (((strlen(command) + strlen(" -sdds"))+1) < CMDSIZE)
                    strcat(command, " -sdds");
                else
                    ok = FALSE;
            }
            else if (Math_sdds_output_spec2 == REQNOSDDS)
            {
                if (((strlen(command) + strlen(" -nosdds"))+1) < CMDSIZE)
                    strcat(command, " -nosdds");
                else
                    ok = FALSE;
            } /* endif */
        } /* endif */

	if (ok)
	{
	    set_running(Math_status2_label);
	    clear_scrolled_text(Math_log2_scroll_txt);
	    
printf(">%s<\n", command);
	    /* calling burtmath */
	    rc = system(command);

	    /* dumping logfile into scrolled text */
	    file_to_scrolled_text(Math_logfilename, Math_log2_scroll_txt);

	    if (WIFEXITED(rc))
		if (WEXITSTATUS(rc))
		    set_not_ok(Math_status2_label);
		else
		    set_ok(Math_status2_label);
	    else
	    popup_message("SYS FAULT: system() returned but cmd not exited.");

	}
	else
popup_message("SYS FAULT: could not construct command ... not long enough.");
    } 
    else
    {
	if (!(*op1_fname))
	    popup_error("ERROR: must specify a snapshot file as an operand.");

	if (!(*result_fname))
	    popup_error("ERROR: must specify a result file.");

	if (ret_cp == mult_factor)
	    popup_error("ERROR: multiplication factor is not a number.");

    } /* endif */

    XtFree(op1_fname);
    XtFree(mult_factor);
    XtFree(result_fname);

} /* end math_go_button2_cb() */

/****************************************************************
*
*
****************************************************************/

static void math_view_log_button2_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    XtMapWidget(Math_log2_shell);
} /* end math_view_log_button2_cb() */

/****************************************************************
*
* called for BOTH operands of the math operation.
*
****************************************************************/

static void math_file_sel_ok_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

char *cp;
Widget text_field;

    text_field = (Widget) client_data;

    cp = XmTextFieldGetString(XmFileSelectionBoxGetChild(w, XmDIALOG_TEXT));

    XtVaSetValues(text_field,
	XmNvalue, cp,
	NULL);

    XtFree(cp);

} /* end math_file_sel_ok_cb() */

/****************************************************************
*
*
****************************************************************/

static void math_sdds_toggle2_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

    if (XmToggleButtonGetState(w))
	Math_sdds_output_spec2 = (int) client_data;

} /* end math_sdds_toggle2_cb() */

