/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#include "burtgooey.h"
#include "burtgooeyfuncdecls.h"
#include <unistd.h>

/* Externally visible local functions */
void main_read_button_cb();
void cleanup_read();

/* Private local functions */
static void setup_read();
/* callbacks */
static void read_clear_button_cb();
static void read_print_button_cb();
static void read_read_button_cb();
static void read_view_log_button_cb();
static void read_file_sel_ok_cb();
static void read_remove_button_cb();
static void read_req_file_sel_button_cb();
static void read_view_req_button_cb();
static void read_sdds_toggle_cb();

/* Local Global Variables */
static char *Read_labels_strings[] =
    {"Snapshot Filename:", "Comments:", "Keywords:"};
static Widget Read_shell;
static Widget Read_log_shell;
static Widget Read_text[XtNumber(Read_labels_strings)];
static char Read_logfilename[MAXPATHLEN];
static Widget Read_status_label; 
static char Read_sdds_output_spec;
static Widget Read_log_scroll_txt;
static Widget Read_view_req_scroll_txt;
static Widget Read_view_req_shell;
static Widget Read_req_scr_lst;

/****************************************************************
*
*
****************************************************************/

void main_read_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

static char default_snapshot_filename[MAXPATHLEN];

    if (!Read_setup)
    {
	setup_read((Widget) client_data);
	Read_setup = TRUE;
    } /* endif */

    if (BurtCreateUniqueFilename((char *) NULL, ".snap", 
	default_snapshot_filename))
    {
	XmTextFieldSetString(Read_text[0], default_snapshot_filename);
	XtMapWidget(Read_shell);
    }
    else
	popup_message("SYS FAULT: could not create default_snapshot_filename.");

} /* end main_read_button_cb() */

/****************************************************************
*
*
****************************************************************/

void cleanup_read()
{
    unlink(Read_logfilename);
} /* end cleanup_read() */

/****************************************************************
*
*
****************************************************************/

static void setup_read(shell)
Widget shell;
{

Arg args[NARGS];
int nargs;
XmString xs;
int i;
Widget read_form, read_req_file_sel;
Widget read_req_file_sel_button, read_view_req_button;
Widget read_remove_button, read_clear_button, read_print_button;
Widget read_labels[XtNumber(Read_labels_strings)];
Widget read_read_button, read_view_log_button, read_help_button; 
Widget read_done_button, read_print_log_button;
Widget read_rc, read_req_files_list_rc, read_req_files_buttons_rc;
Widget read_parms_rc, read_buttons_rc;
/* read log shell */
Widget read_log_form, read_log_rc;
Widget read_log_done_button;
/* read view req shell */
Widget read_view_req_form, read_view_req_rc;
Widget read_view_req_done_button;
Widget read_sdds_rc, read_sdds_toggle[XtNumber(Sdds_labels_strings)];

    if (!BurtCreateUniqueFilename(PUBLICWRITEDIRECTORY, ".read.log",
            Read_logfilename))
    {
        fprintf(Log_fp,
            "ERROR: unable to create a unique filename for read log\n");
        exit(1);
    } /* endif */

    /**************/
    /*            */
    /* read shell */
    /*            */
    /**************/

    Read_shell = XtVaCreateWidget("Backup", 
		    topLevelShellWidgetClass, shell,
		    XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
		    NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 400); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    read_form = XtCreateWidget("read_form", xmFormWidgetClass, Read_shell, 
		    args, nargs);

    /* read popup request file selection box */
    xs = XmStringCreateSimple("Request File Selector");
    nargs = 0;
    XtSetArg(args[nargs], XmNdialogTitle, xs); nargs ++;
    read_req_file_sel = XmCreateFileSelectionDialog(read_form, 
			"read_req_file_sel", args, nargs);
    XtUnmanageChild(XmFileSelectionBoxGetChild(read_req_file_sel, 
	XmDIALOG_HELP_BUTTON)); /* removing HELP button */
    XmStringFree(xs);

    /* main read rowcolumn */
    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    read_rc = XmCreateRowColumn(read_form, "read_rc", args, nargs);

	/* read request file rowcolumn */
	read_req_files_list_rc = XmCreateRowColumn(read_rc, 
		"read_req_files_list_rc", 
		NULL, 0);

	    nargs = 0;
	    XtSetArg(args[nargs], XmNselectionPolicy, XmEXTENDED_SELECT); nargs ++;
	    XtSetArg(args[nargs], XmNheight, 100); nargs ++;
	    Read_req_scr_lst = XmCreateScrolledList(read_req_files_list_rc, 
			    "Read_req_scr_lst", args, nargs);
	    XtManageChild(Read_req_scr_lst);

	XtManageChild(read_req_files_list_rc);

	/* read request file buttons rowcolumn */
	nargs = 0;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	read_req_files_buttons_rc = XmCreateRowColumn(read_rc, 
				    "read_req_files_buttons_rc", args, nargs);

	    read_req_file_sel_button =  
    XmCreatePushButton(read_req_files_buttons_rc, "read_req_file_sel_button", 
		    NULL, 0);
	    XtManageChild(read_req_file_sel_button);

	    read_view_req_button =  
    XmCreatePushButton(read_req_files_buttons_rc, "read_view_req_button", 
		    NULL, 0);
	    XtManageChild(read_view_req_button);

	    read_remove_button = 
	    XmCreatePushButton(read_req_files_buttons_rc, "read_remove_button", 
		    NULL, 0);
	    XtManageChild(read_remove_button);

	    read_clear_button = 
	    XmCreatePushButton(read_req_files_buttons_rc, "read_clear_button", 
		    NULL, 0);
	    XtManageChild(read_clear_button);

	    read_print_button = 
	    XmCreatePushButton(read_req_files_buttons_rc, "read_print_button", 
		    NULL, 0);
	    XtManageChild(read_print_button);

	XtManageChild(read_req_files_buttons_rc);

	read_parms_rc = XmCreateRowColumn(read_rc, "read_parms_rc", 
			    NULL, 0);

	    for (i = 0; i < XtNumber(Read_labels_strings); i ++)
	    {
		read_labels[i] = XmCreateLabel(read_parms_rc, 
				    Read_labels_strings[i], NULL, 0);
		Read_text[i] = XmCreateTextField(read_parms_rc, "", NULL, 0);
	    } /* endfor */
	    XtManageChildren(read_labels, XtNumber(Read_labels_strings));
	    XtManageChildren(Read_text, XtNumber(Read_labels_strings));

	XtManageChild(read_parms_rc);

	/* read sdds spec */
	/* from page 414 AT&T Motif Course Manual */
	nargs = 0;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	read_sdds_rc = XmCreateRadioBox(read_rc, "read_sdds_rc", args, nargs);

	    for (i = 0; i < XtNumber(Sdds_labels_strings); i ++)
		read_sdds_toggle[i] = XmCreateToggleButton(read_sdds_rc, 
				    Sdds_labels_strings[i], NULL, 0);

	    /* setting initial radio box toggle state */
	    XmToggleButtonSetState(read_sdds_toggle[NOREQ], True, False);
	    Read_sdds_output_spec = NOREQ;

	    XtManageChildren(read_sdds_toggle, XtNumber(Sdds_labels_strings));

	XtManageChild(read_sdds_rc);

	nargs = 0;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	read_buttons_rc = XmCreateRowColumn(read_rc, "read_parms_rc", 
			    args, nargs);

	    read_read_button = XmCreatePushButton(read_buttons_rc, 
				"read_read_button", NULL, 0);
	    XtManageChild(read_read_button);

	    Read_status_label = XmCreateLabel(read_buttons_rc, 
				"Read_status_label", NULL, 0);

	    XmChangeColor(Read_status_label, Green_pixel);
	    xs = XmStringCreateSimple(OKSTRING);
	    XtVaSetValues(Read_status_label,
		XmNlabelString, xs,
		NULL);
	    XmStringFree(xs);
	    XtManageChild(Read_status_label);

	    read_view_log_button = XmCreatePushButton(read_buttons_rc, 
				    "read_view_log_button", NULL, 0);
	    XtManageChild(read_view_log_button);

	    read_print_log_button = XmCreatePushButton(read_buttons_rc, 
				    "read_print_log_button", NULL, 0);
	    XtManageChild(read_print_log_button);

	    read_help_button = XmCreatePushButton(read_buttons_rc, 
				"read_help_button", NULL, 0);
	    XtManageChild(read_help_button);

	    read_done_button = XmCreatePushButton(read_buttons_rc, 
				"read_done_button", NULL, 0);
	    XtManageChild(read_done_button);

	XtManageChild(read_buttons_rc);

    XtManageChild(read_rc);

    /* read log shell */
    Read_log_shell = XtVaCreateWidget("Backup Log Viewer", 
		    topLevelShellWidgetClass, Read_shell,
		    XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
		    NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 350); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    read_log_form = XtCreateWidget("read_log_form", xmFormWidgetClass, 
			Read_log_shell, args, nargs);

    /* read log rowcolumn */
    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    read_log_rc = XmCreateRowColumn(read_log_form, "read_log_rc", args, nargs);

	nargs = 0;
	XtSetArg(args[nargs], XmNeditMode, XmMULTI_LINE_EDIT); nargs ++;
	XtSetArg(args[nargs], XmNeditable, False); nargs ++;
	XtSetArg(args[nargs], XmNheight, 300); nargs ++;
	Read_log_scroll_txt = XmCreateScrolledText(read_log_rc, 
				"Read_log_scroll_txt", args, nargs);
	XtManageChild(Read_log_scroll_txt);

	read_log_done_button = XmCreatePushButton(read_log_rc, 
				    "read_log_done_button", NULL, 0);
	XtManageChild(read_log_done_button);

    XtManageChild(read_log_rc);

    XtManageChild(read_log_form);
    XtManageChild(Read_log_shell);

    /* read view req shell */
    Read_view_req_shell = XtVaCreateWidget("Request File Viewer", 
		    topLevelShellWidgetClass, Read_shell,
		    XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
		    NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 350); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    read_view_req_form = XtCreateWidget("read_view_req_form", xmFormWidgetClass,
			    Read_view_req_shell, args, nargs);

    /* main read rowcolumn */
    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    read_view_req_rc = XmCreateRowColumn(read_view_req_form, "read_view_req_rc",
			args, nargs);

	nargs = 0;
	XtSetArg(args[nargs], XmNeditMode, XmMULTI_LINE_EDIT); nargs ++;
	XtSetArg(args[nargs], XmNeditable, False); nargs ++;
	XtSetArg(args[nargs], XmNheight, 300); nargs ++;
	Read_view_req_scroll_txt = XmCreateScrolledText(read_view_req_rc, 
				    "Read_view_req_scroll_txt", args, nargs);
	XtManageChild(Read_view_req_scroll_txt);

	read_view_req_done_button = XmCreatePushButton(read_view_req_rc, 
				    "read_view_req_done_button", NULL, 0);
	XtManageChild(read_view_req_done_button);

    XtManageChild(read_view_req_rc);

    XtManageChild(read_view_req_form);
    XtManageChild(Read_view_req_shell);

    XtManageChild(read_form);
    XtManageChild(Read_shell);

    /*************/
    /*           */
    /* Callbacks */
    /*           */
    /*************/

    XtAddCallback(read_read_button, XmNactivateCallback, read_read_button_cb,
	NULL);
    XtAddCallback(read_view_log_button, XmNactivateCallback,
        read_view_log_button_cb, NULL);
    XtAddCallback(read_print_log_button, XmNactivateCallback,
        shared_print_log_button_cb, (char *) Read_logfilename);
    XtAddCallback(read_help_button, XmNactivateCallback,
        shared_temp_help_button_cb, NULL);
    XtAddCallback(read_done_button, XmNactivateCallback,
        shared_done_button_cb, (Widget) Read_shell);
    XtAddCallback(read_remove_button, XmNactivateCallback,
        read_remove_button_cb, NULL);
    XtAddCallback(read_req_file_sel_button, XmNactivateCallback,
        read_req_file_sel_button_cb, (Widget) read_req_file_sel);
    XtAddCallback(read_view_req_button, XmNactivateCallback,
        read_view_req_button_cb, NULL);
    XtAddCallback(read_clear_button, XmNactivateCallback, read_clear_button_cb,
        NULL);
    XtAddCallback(read_print_button, XmNactivateCallback, read_print_button_cb,
        NULL);
    XtAddCallback(read_req_file_sel, XmNcancelCallback, XtUnmanageChild, NULL);
    XtAddCallback(read_req_file_sel, XmNokCallback, read_file_sel_ok_cb, NULL);
    /* read log shell */
    XtAddCallback(read_log_done_button, XmNactivateCallback,
        shared_done_button_cb, (Widget) Read_log_shell);
    /* read view req shell */
    XtAddCallback(read_view_req_done_button, XmNactivateCallback,
        shared_done_button_cb, (Widget) Read_view_req_shell);
    for (i = 0; i < XtNumber(Sdds_labels_strings); i ++)
        XtAddCallback(read_sdds_toggle[i], XmNvalueChangedCallback,
            read_sdds_toggle_cb, (int) i);

} /* end setup_read() */

/************************/
/*                      */
/* read shell callbacks */
/*                      */
/************************/

/****************************************************************
*
*
****************************************************************/

static void read_clear_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    XmListDeleteAllItems(Read_req_scr_lst);
} /* end read_clear_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void read_print_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

XmStringTable choices;
int n_choices;
int i, j;
char mbuff[1000];
char *cp;

    /* returns the items themselves, not a copy ... do not free */
    XtVaGetValues(Read_req_scr_lst,
	XmNselectedItemCount, &n_choices,
	XmNselectedItems, &choices,
	NULL);

    if (n_choices)
    {
	/* loop to discover legal printable files from those selected */
	for (i = 0; i < n_choices; i ++)
	{
	    if (XmStringGetLtoR(choices[i], XmFONTLIST_DEFAULT_TAG, &cp))
	    {
		print_file(cp);
		XtFree(cp);
	    } 
	    else
popup_message("SYS FAULT: unable to convert one of the files to a string.");
	} /* endfor */
    } /* endif */

} /* end read_print_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void read_file_sel_ok_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

char *cp;
XmString xm_str;

    cp = XmTextFieldGetString(XmFileSelectionBoxGetChild(w, XmDIALOG_TEXT));

    xm_str = XmStringCreateSimple(cp);

    if (!XmListItemExists(Read_req_scr_lst, xm_str))
	/* placing it in position 0 ... meaning at the end */
	XmListAddItem(Read_req_scr_lst, xm_str, 0); 

    XtFree(cp);
    XmStringFree(xm_str);

} /* end read_file_sel_ok_cb() */

/****************************************************************
*
*
****************************************************************/

static void read_read_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

XmStringTable choices;
XmString xs;
int n_choices;
int i;
int buffsize;
int rc;
char *cp;
char *snapshotfile;
char command[CMDSIZE];
struct stat stat_buff;
int fd;
char *logbuff;
BOOL ok;

    ok = TRUE;

    /* returns the items themselves, not a copy ... do not free */
    XtVaGetValues(Read_req_scr_lst,
	XmNitemCount, &n_choices,
	XmNitems, &choices,
	NULL);

    snapshotfile = XmTextFieldGetString(Read_text[0]);

    if (n_choices && *snapshotfile)
    {
	strcpy(command, "burtrb -f");
	for (i = 0; ok && i < n_choices; i ++)
	{
	    if (XmStringGetLtoR(choices[i], XmFONTLIST_DEFAULT_TAG, &cp))
	    {
		if ((strlen(command) +strlen(cp) + 2) < CMDSIZE)
		{
		    strcat(command, " ");
		    strcat(command, cp);
		    XtFree(cp);
		}
		else
		    ok = FALSE;
	    } /* endif */
	} /* endfor */

	if (*Defines)
	{
	    if (ok && ((strlen(command) + 2+strlen(Defines)) < CMDSIZE))
	    {
    printf("adding Defines >%s< to command\n", Defines);
		strcat(command, " ");
		strcat(command, Defines);
		strcat(command, " ");
	    }
	    else
		ok = FALSE;
	} /* endif */

	if (*Includes)
	{
	    if (ok && ((strlen(command) + 2+strlen(Includes)) < CMDSIZE))
	    {
		strcat(command, " ");
		strcat(command, Includes);
		strcat(command, " ");
	    }
	    else
		ok = FALSE;
	} /* endif */

	if (ok &&
	    ((strlen(command) + 4+strlen(snapshotfile) + 
		4+strlen(Read_logfilename) + 4 + 1) < CMDSIZE))
	{
	    strcat(command, " -o ");
	    strcat(command, snapshotfile);

	    unlink(Read_logfilename);
	    strcat(command, " -l ");
	    strcat(command, Read_logfilename);

	    strcat(command, " -v ");
	} 
	else
	    ok = FALSE;

	cp = XmTextFieldGetString(Read_text[1]);
	if (*cp)
	{
	    if (ok && (strlen(command) + 4+strlen(cp)) < CMDSIZE)
	    {
		strcat(command, " -c \"");
		strcat(command, cp);
		strcat(command, "\"");
	    }
	    else
		ok = FALSE;
	} /* endif */
	XtFree(cp);

	cp = XmTextFieldGetString(Read_text[2]);
	if (*cp)
	{
	    if (ok && (strlen(command) + 4+strlen(cp)) < CMDSIZE)
	    {
		strcat(command, " -k \"");
		strcat(command, cp);
		strcat(command, "\"");
	    }
	    else
		ok = FALSE;
	} /* endif */
	XtFree(cp);

	if (ok)
	{
	    if (Read_sdds_output_spec == REQSDDS)
	    {
		if (((strlen(command) + strlen(" -sdds"))+1) < CMDSIZE)
		    strcat(command, " -sdds");
		else
		    ok = FALSE;
	    }
	    else if (Read_sdds_output_spec == REQNOSDDS)
	    {
		if (((strlen(command) + strlen(" -nosdds"))+1) < CMDSIZE)
		    strcat(command, " -nosdds");
		else
		    ok = FALSE;
	    } /* endif */
	} /* endif */

	if (ok)
	{
	    set_running(Read_status_label);
	    clear_scrolled_text(Read_log_scroll_txt);
	    
    printf(">%s<\n", command);
	    /* calling burtrb */
	    rc = system(command);

	    /* dumping logfile into scrolled text */
	    file_to_scrolled_text(Read_logfilename, Read_log_scroll_txt);

	    if (WIFEXITED(rc))
		if (WEXITSTATUS(rc))
		    set_not_ok(Read_status_label);
		else
		    set_ok(Read_status_label);
	    else
	    popup_message("SYS FAULT: system() returned but cmd not exited.");

	}
	else
popup_message("SYS FAULT: could not construct command ... not long enough.");
    } 
    else
    {
	if (!n_choices)
	    popup_error("ERROR: Need at least one request file.");

	if (!(*snapshotfile))
	    popup_error("ERROR: must specify an snapshot file.");

    } /* endif */

    XtFree(snapshotfile);

} /* end read_read_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void read_remove_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

int *pos_list, pos_cnt;
int fix, i, j;

    pos_list = (int *) NULL;

    if (XmListGetSelectedPos(Read_req_scr_lst, &pos_list, &pos_cnt))
    {
	/* at least one selected in scrolled list */
	for (i = 0; i < pos_cnt; i ++)
	{
	    /* loop to find all those that have been removed */
	    /* so far that were ABOVE the item being removed */
	    /* now.  Each one we find that was removed ABOVE */
	    /* the one we're removing now decrements the pos */
	    /* of the one we're removing now by one.         */
	    for (j = 0, fix = 0; j < i; j ++)
		if (pos_list[j] < pos_list[i])
		    fix ++;
	    XmListDeletePos(Read_req_scr_lst, pos_list[i]-fix);
	} /* endfor */
    } /* endif */

    XtFree(pos_list);

} /* end read_remove_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void read_req_file_sel_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

    XtManageChild((Widget) client_data);
    XtPopup(XtParent((Widget) client_data), XtGrabNone);

} /* end read_req_file_sel_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void read_view_log_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

    XtMapWidget(Read_log_shell);

} /* end read_view_log_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void read_view_req_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

XmStringTable choices;
int n_choices;
int i, j;
int next_file_idx;
int total_bytes;
char *total_buff;
int fd;
char mbuff[1000];
struct filerep *fbuff;
char *filecp;
char *cp;
struct stat stat_buff;
BOOL all_ascii;

    /* clearing the view req scroll text */
    XtVaSetValues(Read_view_req_scroll_txt,
	XmNvalue, "",
	NULL);

    /* returns the items themselves, not a copy ... do not free */
    XtVaGetValues(Read_req_scr_lst,
	XmNselectedItemCount, &n_choices,
	XmNselectedItems, &choices,
	NULL);

    if (n_choices)
    {
	if (fbuff = (struct filerep *) calloc(n_choices,sizeof(struct filerep)))
	{
	    /* loop to discover legal printable files from those selected */
	    for (i = 0, next_file_idx = 0; i < n_choices; i ++)
	    {
		if (XmStringGetLtoR(choices[i], XmFONTLIST_DEFAULT_TAG, &cp))
		{
		    if ((fd = open(cp, O_RDONLY)) != -1)
		    {
			if (!fstat(fd, &stat_buff))
			{
			    if (S_ISREG(stat_buff.st_mode))
			    {
				if (fbuff[next_file_idx].buff = 
				    (char *) calloc (stat_buff.st_size,1))
				{
				    if (read(fd, fbuff[next_file_idx].buff, 
					stat_buff.st_size) != -1)
				    {
					/* loop to check file is printable */
					for (j = 0, all_ascii = TRUE; 
					    all_ascii && j < stat_buff.st_size; 
						j ++)
						    all_ascii = 
					isascii(fbuff[next_file_idx].buff[j]);

					if (all_ascii)
					{
					    fbuff[next_file_idx].nbytes 
						= stat_buff.st_size;
					    fbuff[next_file_idx].fname =
						strdup(cp);
					    next_file_idx ++;
					}
					else
					{
					    free(fbuff[next_file_idx].buff);

					    sprintf(mbuff, 
						"ERROR: file %s not viewable.",
						cp);
					    popup_error(mbuff);
					} /* endif */
				    }
				    else
				    {
					free(fbuff[next_file_idx].buff);

					sprintf(mbuff, 
					"SYS FAULT: could not read() file %s.",
					    cp);
					popup_message(mbuff);
				    } /* endif */

				}
				else
				{
				    sprintf(mbuff, 
					"SYS FAULT: failed calloc() %d bytes.", 
					stat_buff.st_size);
				    popup_message(mbuff);
				} /* endif */
			    }
			    else
			    {
				sprintf(mbuff, 
		"NOTE: cannot view file %s because it is not a regular file.", 
				    cp);
			    popup_message(mbuff);
			    } /* endif */
			}
			else
			    popup_message("SYS FAULT: error in fstat().");

			close(fd);
		    }
		    else
		    {
			sprintf(mbuff, "SYS FAULT: cannot open() file %s.", cp);
			popup_message(mbuff);
		    } /* endif */

		    XtFree(cp);
		} 
		else
    popup_message("SYS FAULT: unable to convert one of the files to a string.");
	    } /* endfor */

	    /* now fbuff has all the valid files */
	    for (i = 0, total_bytes = 0; i < next_file_idx; i ++)
		total_bytes += 
		    (fbuff[i].nbytes+strlen(fbuff[i].fname)+1+strlen("\n\n"));
	    /* for null char at end */
	    total_bytes ++;


	    if (total_bytes > 1)
	    {
		/* there are files to view */
		if (total_buff = (char *) calloc(total_bytes, 1))
		{
		    /* loop to append (via copy) each file to total_buff */
		    /* and then free mem from fbuff */
		    for (i = 0, cp = total_buff; i < next_file_idx; i ++)
		    {
			memcpy(cp, fbuff[i].fname, strlen(fbuff[i].fname));
			cp = cp + strlen(fbuff[i].fname);

			*cp = '\n';
			cp ++;

			memcpy(cp, fbuff[i].buff, fbuff[i].nbytes);
			cp = cp + fbuff[i].nbytes;

			memcpy(cp, "\n\n", strlen("\n\n"));
			cp = cp + strlen("\n\n");

			free(fbuff[i].buff);
			free(fbuff[i].fname);
		    } /* endfor */

		    /* although buffer was allocated with calloc(), which */
		    /* presumably allocates a block with all zeros, still */
		    /* need to explicitly place the null char at the end  */
		    /* so as not to get extra garbage characters in the   */
		    /* scroll text ... can't explain why this works ...   */
		    /* just happy that it does.                           */

		    *cp = '\0';

		    XtVaSetValues(Read_view_req_scroll_txt,
			XmNvalue, total_buff,
			NULL);

		    free(total_buff);
		}
		else
		{
		    sprintf(mbuff, "SYS FAULT: failed calloc() %d bytes.", 
			total_bytes);
		    popup_message(mbuff);
		} /* endif */
	    } /* endif */

	    free(fbuff);
	}
	else
	{
	    sprintf(mbuff, "SYS FAULT: failed to calloc() %d bytes.", 
		n_choices*sizeof(struct filerep));
	    popup_message(mbuff);
	} /* endif */
    } /* endif */

    XtMapWidget(Read_view_req_shell);

} /* end read_view_req_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void read_sdds_toggle_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

    if (XmToggleButtonGetState(w))
	Read_sdds_output_spec = (int) client_data;

} /* end read_sdds_toggle_cb() */

