/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#include "burtgooey.h"
#include "burtgooeyfuncdecls.h"
#include <unistd.h>

#define SETOP1VIEWREQBUTTONNAME "set_op1_view_req_button"
#define SETOP2VIEWREQBUTTONNAME "set_op2_view_req_button"

#define SETOP1PRINTREQBUTTONNAME "set_op1_print_req_button"
#define SETOP2PRINTREQBUTTONNAME "set_op2_print_req_button"

/* Externally visible functions */
void main_set_button_cb();
void cleanup_set();

/* Private local functions */
static void setup_set();
/* callbacks */
static void set_op1_req_file_sel_button_cb();
static void set_op2_req_file_sel_button_cb();
static void set_view_req_button_cb();
static void set_print_req_button_cb();
static void set_go_button_cb();
static void set_view_log_button_cb();
static void set_file_sel_ok_cb();
static void set_operation_toggle_cb();
static void set_sdds_toggle_cb();

/* Local Global Variables */
/* set operations */
#define UNION        0
#define INTERSECTION 1
#define DIFFERENCE   2

/* NOTE: Must be in the same order as the #defines */
static char *Set_labels_strings[] = {"Union", "Intersection", "Difference"};
static Widget Set_shell;
static Widget Set_result_text;
static Widget Set_log_shell;
static Widget Set_op1_text;
static Widget Set_op1_view_req_scroll_txt;
static Widget Set_op1_view_req_shell;
static Widget Set_op2_text;
static Widget Set_op2_view_req_scroll_txt;
static Widget Set_op2_view_req_shell;
static Widget Set_log_scroll_txt;
static Widget Set_status_label;
static char Set_logfilename[MAXPATHLEN];
static char Set_operation;
static char Set_sdds_output_spec;

/****************************************************************
*
*
****************************************************************/

void main_set_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

static char default_result_filename[MAXPATHLEN];

    if (!Set_setup)
    {
	setup_set((Widget) client_data);
	Set_setup = TRUE;
    } /* endif */

    if (BurtCreateUniqueFilename((char *) NULL, ".req", 
	default_result_filename))
    {
        XmTextFieldSetString(Set_result_text,
            default_result_filename);
        XtMapWidget(Set_shell);
    }
    else
        popup_message("SYS FAULT: could not create default_result_filename.");

} /* end main_set_button_cb() */

/****************************************************************
*
*
****************************************************************/

void cleanup_set()
{
    unlink(Set_logfilename);
} /* end cleanup_set() */

/****************************************************************
*
*
****************************************************************/

static void setup_set(shell)
Widget shell;
{

Arg args[NARGS];
int nargs;
XmString xs;
int i;

Widget set_form, set_rc; 
Widget set_op1_req_file_sel, set_op2_req_file_sel;
Widget set_buttons_rc;
Widget set_go_button, set_view_log_button, set_help_button, set_done_button;
Widget set_print_log_button;
Widget set_ops_rc;
Widget set_toggle[XtNumber(Set_labels_strings)];
Widget set_equal_label;

Widget set_op1_rc;
Widget set_op1_req_files_buttons_rc;
Widget set_op1_req_file_sel_button, set_op1_view_req_button, set_op1_print_req_button;
Widget set_op1_view_req_form, set_op1_view_req_rc, set_op1_view_req_done_button;

Widget set_op2_rc; 
Widget set_op2_req_files_buttons_rc;
Widget set_op2_req_file_sel_button, set_op2_view_req_button, set_op2_print_req_button;
Widget set_op2_view_req_form, set_op2_view_req_rc, set_op2_view_req_done_button;

Widget set_log_form, set_log_rc, set_log_done_button;

Widget set_sdds_rc, set_sdds_toggle[XtNumber(Sdds_labels_strings)];

Dimension height, width;

    if (!BurtCreateUniqueFilename(PUBLICWRITEDIRECTORY, ".set.log",
            Set_logfilename))
    {
        fprintf(Log_fp,
            "ERROR: unable to create a unique filename for set log\n");
        exit(1);
    } /* endif */

    /*************/
    /*           */
    /* set shell */
    /*           */
    /*************/

    Set_shell = XtVaCreateWidget("Set Operations on Request Files", 
		    topLevelShellWidgetClass, shell,
		    XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
		    NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 330); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    set_form = XtCreateWidget("set_form", xmFormWidgetClass, Set_shell, 
		    args, nargs);

    /* set popup request file selection boxes */
    xs = XmStringCreateSimple("Request File Selector (Operand 1)");
    nargs = 0;
    XtSetArg(args[nargs], XmNdialogTitle, xs); nargs ++;
    set_op1_req_file_sel = 
	XmCreateFileSelectionDialog(set_form, "set_op1_req_file_sel", 
	    args, nargs);
    XtUnmanageChild(XmFileSelectionBoxGetChild(set_op1_req_file_sel, 
	XmDIALOG_HELP_BUTTON)); /* removing HELP button */
    XmStringFree(xs);

    xs = XmStringCreateSimple("Request File Selector (Operand 2)");
    nargs = 0;
    XtSetArg(args[nargs], XmNdialogTitle, xs); nargs ++;
    set_op2_req_file_sel = 
	XmCreateFileSelectionDialog(set_form, "set_op2_req_file_sel", 
	    args, nargs);
    XtUnmanageChild(XmFileSelectionBoxGetChild(set_op2_req_file_sel, 
	XmDIALOG_HELP_BUTTON)); /* removing HELP button */
    XmStringFree(xs);

    /* main set rowcolumn */
    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    set_rc = XmCreateRowColumn(set_form, "set_rc", args, nargs);

	/* op1 */
	set_op1_rc = XmCreateRowColumn(set_rc, "set_op1_rc", NULL, 0);
	    Set_op1_text = XmCreateTextField(set_op1_rc, 
			    "Set_op1_text", 
			    NULL, 0);
	    XtManageChild(Set_op1_text);

	    /* set op1 request file buttons rowcolumn */
	    nargs = 0;
	    XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	    set_op1_req_files_buttons_rc = XmCreateRowColumn(set_op1_rc, 
					"set_op1_req_files_buttons_rc", 
					args, nargs);

		set_op1_req_file_sel_button =  
		    XmCreatePushButton(set_op1_req_files_buttons_rc, 
			"set_op1_req_file_sel_button", 
			NULL, 0);
		XtManageChild(set_op1_req_file_sel_button);

		set_op1_view_req_button =  
		    XmCreatePushButton(set_op1_req_files_buttons_rc, 
			SETOP1VIEWREQBUTTONNAME, 
			NULL, 0);
		XtManageChild(set_op1_view_req_button);

		set_op1_print_req_button =  
		    XmCreatePushButton(set_op1_req_files_buttons_rc, 
			SETOP1PRINTREQBUTTONNAME, 
			NULL, 0);
		XtManageChild(set_op1_print_req_button);

	    XtManageChild(set_op1_req_files_buttons_rc);
	XtManageChild(set_op1_rc);

	/* set operation */
	/* from page 414 AT&T Motif Course Manual */
	nargs = 0;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	set_ops_rc = XmCreateRadioBox(set_rc, "set_ops_rc", 
			args, nargs);

	    for (i = 0; i < XtNumber(Set_labels_strings); i ++)
		set_toggle[i] = XmCreateToggleButton(set_ops_rc, 
				    Set_labels_strings[i], NULL, 0);

	    /* setting initial radio box toggle state */
	    XmToggleButtonSetState(set_toggle[UNION], True, False);
	    Set_operation = UNION;

	    XtManageChildren(set_toggle, XtNumber(Set_labels_strings));

	XtManageChild(set_ops_rc);

	/* op2 */
	set_op2_rc = XmCreateRowColumn(set_rc, "set_op2_rc", NULL, 0);
	    Set_op2_text = XmCreateTextField(set_op2_rc, 
				"Set_op2_text", NULL, 0);
	    XtManageChild(Set_op2_text);

	    /* set op2 request file buttons rowcolumn */
	    nargs = 0;
	    XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	    set_op2_req_files_buttons_rc = XmCreateRowColumn(set_op2_rc, 
					"set_op2_req_files_buttons_rc", 
					args, nargs);

		set_op2_req_file_sel_button =  
		    XmCreatePushButton(set_op2_req_files_buttons_rc, 
			"set_op2_req_file_sel_button", 
			NULL, 0);
		XtManageChild(set_op2_req_file_sel_button);

		set_op2_view_req_button =  
		    XmCreatePushButton(set_op2_req_files_buttons_rc, 
			SETOP2VIEWREQBUTTONNAME, 
			NULL, 0);
		XtManageChild(set_op2_view_req_button);

		set_op2_print_req_button =  
		    XmCreatePushButton(set_op2_req_files_buttons_rc, 
			SETOP2PRINTREQBUTTONNAME, 
			NULL, 0);
		XtManageChild(set_op2_print_req_button);

	    XtManageChild(set_op2_req_files_buttons_rc);
	XtManageChild(set_op2_rc);

	/* = label */
	set_equal_label = XmCreateLabel(set_rc, "set_equal_label", 
		    NULL, 0);
	xs = XmStringCreateSimple("=");
	XtVaSetValues(set_equal_label,
	    XmNlabelString, xs,
	    NULL);
	XmStringFree(xs);
	XtManageChild(set_equal_label);

	/* result */
	Set_result_text = 
	    XmCreateTextField(set_rc, "Set_result_text", NULL, 0);
	XtManageChild(Set_result_text);

	/* set sdds spec */
	/* from page 414 AT&T Motif Course Manual */
	nargs = 0;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	set_sdds_rc = XmCreateRadioBox(set_rc, "set_sdds_rc", args, nargs);

	    for (i = 0; i < XtNumber(Sdds_labels_strings); i ++)
		set_sdds_toggle[i] = XmCreateToggleButton(set_sdds_rc, 
				    Sdds_labels_strings[i], NULL, 0);

	    /* setting initial radio box toggle state */
	    XmToggleButtonSetState(set_sdds_toggle[NOREQ], True, False);
	    Set_sdds_output_spec = NOREQ;

	    XtManageChildren(set_sdds_toggle, XtNumber(Sdds_labels_strings));

	XtManageChild(set_sdds_rc);

	/* set buttons rowcolumn */
	nargs = 0;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	set_buttons_rc = XmCreateRowColumn(set_rc, 
		"set_buttons_rc", 
		args, nargs);

	    set_go_button = XmCreatePushButton(set_buttons_rc, "set_go_button", 
				NULL, 0);
	    XtManageChild(set_go_button);

	    Set_status_label = XmCreateLabel(set_buttons_rc, "Set_status_label",
				NULL, 0);

	    set_ok(Set_status_label);

	    XtManageChild(Set_status_label);

	    set_view_log_button = XmCreatePushButton(set_buttons_rc, 
				    "set_view_log_button", NULL, 0);
	    XtManageChild(set_view_log_button);

	    set_print_log_button = XmCreatePushButton(set_buttons_rc, 
				    "set_print_log_button", NULL, 0);
	    XtManageChild(set_print_log_button);

	    set_help_button = XmCreatePushButton(set_buttons_rc, 
				"set_help_button", NULL, 0);
	    XtManageChild(set_help_button);

	    set_done_button = XmCreatePushButton(set_buttons_rc, 
				"set_done_button", NULL, 0);
	    XtManageChild(set_done_button);

	XtManageChild(set_buttons_rc);

    XtManageChild(set_rc);

    /* set view op1 req shell */
    Set_op1_view_req_shell = 
	XtVaCreateWidget("Request File Viewer (Operand 1)", 
		    topLevelShellWidgetClass, Set_shell,
		    XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
		    NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 350); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    set_op1_view_req_form = XtCreateWidget("set_op1_view_req_form", 
				xmFormWidgetClass, Set_op1_view_req_shell,
				args, nargs);

    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    set_op1_view_req_rc = XmCreateRowColumn(set_op1_view_req_form, 
			    "set_op1_view_req_rc", args, nargs);

	nargs = 0;
	XtSetArg(args[nargs], XmNeditMode, XmMULTI_LINE_EDIT); nargs ++;
	XtSetArg(args[nargs], XmNeditable, False); nargs ++;
	XtSetArg(args[nargs], XmNheight, 300); nargs ++;
	Set_op1_view_req_scroll_txt = XmCreateScrolledText(set_op1_view_req_rc, 
		"Set_op1_view_req_scroll_txt", args, nargs);
	XtManageChild(Set_op1_view_req_scroll_txt);

	set_op1_view_req_done_button = XmCreatePushButton(set_op1_view_req_rc, 
		"set_op1_view_req_done_button", NULL, 0);
	XtManageChild(set_op1_view_req_done_button);

    XtManageChild(set_op1_view_req_rc);
    XtManageChild(set_op1_view_req_form);
    XtManageChild(Set_op1_view_req_shell);

    /* set view op2 req shell */
    Set_op2_view_req_shell = XtVaCreateWidget("Request File Viewer (Operand 2)",
		    topLevelShellWidgetClass, Set_shell,
		    XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
		    NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 350); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    set_op2_view_req_form = XtCreateWidget("set_op2_view_req_form", 
				xmFormWidgetClass, Set_op2_view_req_shell,
				args, nargs);

    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    set_op2_view_req_rc = XmCreateRowColumn(set_op2_view_req_form, 
			    "set_op2_view_req_rc", args, nargs);

	nargs = 0;
	XtSetArg(args[nargs], XmNeditMode, XmMULTI_LINE_EDIT); nargs ++;
	XtSetArg(args[nargs], XmNeditable, False); nargs ++;
	XtSetArg(args[nargs], XmNheight, 300); nargs ++;
	Set_op2_view_req_scroll_txt = 
	    XmCreateScrolledText(set_op2_view_req_rc, 
		"Set_op2_view_req_scroll_txt", args, nargs);
	XtManageChild(Set_op2_view_req_scroll_txt);

	set_op2_view_req_done_button = 
	    XmCreatePushButton(set_op2_view_req_rc, 
		"set_op2_view_req_done_button", NULL, 0);
	XtManageChild(set_op2_view_req_done_button);

    XtManageChild(set_op2_view_req_rc);
    XtManageChild(set_op2_view_req_form);
    XtManageChild(Set_op2_view_req_shell);

    /* set log shell */
    Set_log_shell = XtVaCreateWidget("Set Log Viewer", 
			topLevelShellWidgetClass, Set_shell,
			XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
			NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 350); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    set_log_form = XtCreateWidget("set_log_form", xmFormWidgetClass, 
		    Set_log_shell, 
		    args, nargs);

    /* main set log rowcolumn */
    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    set_log_rc = XmCreateRowColumn(set_log_form, "set_log_rc", 
		    args, nargs);

	nargs = 0;
	XtSetArg(args[nargs], XmNeditMode, XmMULTI_LINE_EDIT); nargs ++;
	XtSetArg(args[nargs], XmNeditable, False); nargs ++;
	XtSetArg(args[nargs], XmNheight, 300); nargs ++;
	Set_log_scroll_txt = XmCreateScrolledText(set_log_rc, 
				"Set_log_scroll_txt", args, nargs);
	XtManageChild(Set_log_scroll_txt);

	set_log_done_button = XmCreatePushButton(set_log_rc, 
				"set_log_done_button", 
				NULL, 0);
	XtManageChild(set_log_done_button);

    XtManageChild(set_log_rc);
    XtManageChild(set_log_form);
    XtManageChild(Set_log_shell);

    XtManageChild(set_form);
    XtManageChild(Set_shell);

    /*************/
    /*           */
    /* Callbacks */
    /*           */
    /*************/

    XtAddCallback(set_go_button, XmNactivateCallback, set_go_button_cb, NULL);
    XtAddCallback(set_view_log_button, XmNactivateCallback, 
	set_view_log_button_cb, NULL);
    XtAddCallback(set_print_log_button, XmNactivateCallback, 
	shared_print_log_button_cb, (char *) Set_logfilename);
    XtAddCallback(set_help_button, XmNactivateCallback, 
	shared_temp_help_button_cb, NULL);
    XtAddCallback(set_done_button, XmNactivateCallback, 
	shared_done_button_cb, (Widget) Set_shell);
    XtAddCallback(set_op1_req_file_sel_button, XmNactivateCallback, 
	set_op1_req_file_sel_button_cb, (Widget) set_op1_req_file_sel);
    XtAddCallback(set_op2_req_file_sel_button, XmNactivateCallback, 
	set_op2_req_file_sel_button_cb, (Widget) set_op2_req_file_sel);
    XtAddCallback(set_op1_view_req_button, XmNactivateCallback, 
	set_view_req_button_cb, NULL);
    XtAddCallback(set_op2_view_req_button, XmNactivateCallback, 
	set_view_req_button_cb, NULL);
    XtAddCallback(set_op1_print_req_button, XmNactivateCallback, 
	set_print_req_button_cb, NULL);
    XtAddCallback(set_op2_print_req_button, XmNactivateCallback, 
	set_print_req_button_cb, NULL);
    XtAddCallback(set_op1_req_file_sel, XmNcancelCallback, 
	XtUnmanageChild, NULL);
    XtAddCallback(set_op2_req_file_sel, XmNcancelCallback, 
	XtUnmanageChild, NULL);
    XtAddCallback(set_op1_req_file_sel, XmNokCallback, set_file_sel_ok_cb, 
	(Widget) Set_op1_text);
    XtAddCallback(set_op2_req_file_sel, XmNokCallback, set_file_sel_ok_cb, 
	(Widget) Set_op2_text);
    XtAddCallback(set_op1_view_req_done_button, XmNactivateCallback, 
	shared_done_button_cb, (Widget) Set_op1_view_req_shell);
    XtAddCallback(set_op2_view_req_done_button, XmNactivateCallback, 
	shared_done_button_cb, (Widget) Set_op2_view_req_shell);
    for (i = 0; i < XtNumber(Set_labels_strings); i ++)
	XtAddCallback(set_toggle[i], XmNvalueChangedCallback,
	    set_operation_toggle_cb, (XtPointer) i);
    for (i = 0; i < XtNumber(Sdds_labels_strings); i ++)
	XtAddCallback(set_sdds_toggle[i], XmNvalueChangedCallback,
	    set_sdds_toggle_cb, (XtPointer) i);
    /* set log shell */
    XtAddCallback(set_log_done_button, XmNactivateCallback, 
	shared_done_button_cb, (Widget) Set_log_shell);

} /* end setup_set() */

/***********************/
/*                     */
/* set shell callbacks */
/*                     */
/***********************/

/****************************************************************
*
*
****************************************************************/

static void set_op1_req_file_sel_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

    XtManageChild((Widget) client_data);
    XtPopup(XtParent((Widget) client_data), XtGrabNone);

} /* end set_op1_req_file_sel_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void set_op2_req_file_sel_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

    XtManageChild((Widget) client_data);
    XtPopup(XtParent((Widget) client_data), XtGrabNone);

} /* end set_op2_req_file_sel_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void set_view_req_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

XmStringTable choices;
int n_choices;
int i, j;
int total_bytes;
char *total_buff;
int fd;
char mbuff[1000];
struct filerep fbuff;
char *cp;
char *fname;
struct stat stat_buff;
BOOL all_ascii;

Dimension height, width;

    /* clearing the view req scroll text */
    if (strcmp(XtName(w), SETOP2VIEWREQBUTTONNAME))
	XtVaSetValues(Set_op1_view_req_scroll_txt,
	    XmNvalue, "",
	    NULL);
    else
	XtVaSetValues(Set_op2_view_req_scroll_txt,
	    XmNvalue, "",
	    NULL);

    if (strcmp(XtName(w), SETOP2VIEWREQBUTTONNAME))
	fname = XmTextFieldGetString(Set_op1_text);
    else
	fname = XmTextFieldGetString(Set_op2_text);

    if (*fname)
    {
	if ((fd = open(fname, O_RDONLY)) != -1)
	{
	    if (!fstat(fd, &stat_buff))
	    {
		if (S_ISREG(stat_buff.st_mode))
		{
		    if (fbuff.buff = (char *) calloc (stat_buff.st_size,1))
		    {
			if (read(fd, fbuff.buff, stat_buff.st_size) != -1)
			{
			    /* loop to check file is printable */
			    for (j = 0, all_ascii = TRUE; 
				all_ascii && j < stat_buff.st_size; 
				    j ++)
					all_ascii = isascii(fbuff.buff[j]);

			    if (all_ascii)
			    {
				fbuff.nbytes = stat_buff.st_size;
				fbuff.fname = strdup(fname);

				/* now filling scrolled text */
				total_bytes = 
			(fbuff.nbytes+strlen(fbuff.fname)+1+strlen("\n\n"));
				/* for null char at end */
				total_bytes ++;

				if (total_bytes > 1)
				{
				    /* there are files to view */
				    if (total_buff = 
					    (char *) calloc(total_bytes, 1))
				    {
					cp = total_buff;

					/* appending (via copy) file to */
					/* total_buff and then free mem */
					/* from fbuff */
					memcpy(cp, fbuff.fname, 
					    strlen(fbuff.fname));
					cp = cp + strlen(fbuff.fname);

					*cp = '\n';
					cp ++;

					memcpy(cp, fbuff.buff, fbuff.nbytes);
					cp = cp + fbuff.nbytes;

					memcpy(cp, "\n\n", strlen("\n\n"));
					cp = cp + strlen("\n\n");

					free(fbuff.buff);
					free(fbuff.fname);

					/* although buffer was allocated with */
					/* calloc which presumably allocates */
					/* a block with all zeros, still     */ 
					/* need to explicitly place the null */
					/* char at the end  so as not to get */
					/* extra garbage characters in the   */
					/* scroll text ... can't explain why */
					/* this works ... just happy that it */
					/* does.                             */

					*cp = '\0';

					if (strcmp(XtName(w), 
					    SETOP2VIEWREQBUTTONNAME))
				    XtVaSetValues(Set_op1_view_req_scroll_txt,
						XmNvalue, total_buff,
						NULL);
					else
				    XtVaSetValues(Set_op2_view_req_scroll_txt,
						XmNvalue, total_buff,
						NULL);

					free(total_buff);
				    }
				    else
				    {
					sprintf(mbuff, 
					"SYS FAULT: failed calloc() %d bytes.", 
					    total_bytes);
					popup_message(mbuff);
				    } /* endif */
				} /* endif */
			    }
			    else
			    {
				free(fbuff.buff);

				sprintf(mbuff, "ERROR: file %s not viewable.",
				    fname);
				popup_error(mbuff);
			    } /* endif */
			}
			else
			{
			    free(fbuff.buff);

			    sprintf(mbuff, 
				"SYS FAULT: could not read() file %s.", fname);
			    popup_message(mbuff);
			} /* endif */

		    }
		    else
		    {
			sprintf(mbuff, "SYS FAULT: failed calloc() %d bytes.", 
			    stat_buff.st_size);
			popup_message(mbuff);
		    } /* endif */
		}
		else
		{
		    sprintf(mbuff, 
		"NOTE: cannot view file %s because it is not a regular file.", 
			fname);
		popup_message(mbuff);
		} /* endif */
	    }
	    else
		popup_message("SYS FAULT: error in fstat().");

	    close(fd);
	}
	else
	{
	    sprintf(mbuff, "SYS FAULT: cannot open() file %s.", fname);
	    popup_message(mbuff);
	} /* endif */
    } /* endif */

    XtFree(fname);

    if (strcmp(XtName(w), SETOP2VIEWREQBUTTONNAME))
/* { */
/* XtVaGetValues(Set_op1_view_req_shell, XmNheight, &height, XmNwidth, &width, NULL); */
/* printf("mapping op1 viewer h=%d (set 350) w=%d (set 600)\n", height, width); */
	XtMapWidget(Set_op1_view_req_shell);
/* } */
    else
/* { */
/* XtVaGetValues(Set_op2_view_req_shell, XmNheight, &height, XmNwidth, &width, NULL); */
/* printf("mapping op2 viewer h=%d (set 350) w=%d (set 600)\n", height, width); */
	XtMapWidget(Set_op2_view_req_shell);
/* }  */

} /* end set_view_req_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void set_print_req_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

char *fname;

    if (strcmp(XtName(w), SETOP2PRINTREQBUTTONNAME))
	fname = XmTextFieldGetString(Set_op1_text);
    else
	fname = XmTextFieldGetString(Set_op2_text);

    print_file(fname);
    XtFree(fname);

} /* end set_print_req_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void set_go_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

XmStringTable choices;
XmString xs;
int n_choices;
int i;
int buffsize;
int rc;
char *cp;
char *snapshotfile;
char command[CMDSIZE];
struct stat stat_buff;
int fd;
char *logbuff;
char *op1_fname, *op2_fname, *result_fname;
BOOL ok;

    ok = TRUE;

    op1_fname = XmTextFieldGetString(Set_op1_text);
    op2_fname = XmTextFieldGetString(Set_op2_text);
    result_fname = XmTextFieldGetString(Set_result_text);

    if (*op1_fname && *op2_fname && *result_fname)
    {
	if (strlen("burtset -f") < CMDSIZE)
	    strcpy(command, "burtset -f");
	else
	    ok = FALSE;

	if (ok &&
	    (((strlen(command) + 
		1+strlen(op1_fname) + 1+strlen(op2_fname)
		+strlen(" -o")+ 1+strlen(result_fname)
		+strlen(" -v")
		+strlen(" -l" + 1+strlen(Set_logfilename)))+1) < CMDSIZE))
	{
	    strcat(command, " ");
	    strcat(command, op1_fname);
	    strcat(command, " ");
	    strcat(command, op2_fname);

	    strcat(command, " -o");
	    strcat(command, " ");
	    strcat(command, result_fname);

	    strcat(command, " -v");

	    unlink(Set_logfilename);
	    strcat(command, " -l");
	    strcat(command, " ");
	    strcat(command, Set_logfilename);

	} 
	else
	    ok = FALSE;

        if (*Defines)
        {
            if (ok && ((strlen(command) + 2+strlen(Defines)) < CMDSIZE))
            {
    printf("adding Defines >%s< to command\n", Defines);
                strcat(command, " ");
                strcat(command, Defines);
                strcat(command, " ");
            }
            else
                ok = FALSE;
        } /* endif */

        if (*Includes)
        {
            if (ok && ((strlen(command) + 2+strlen(Includes)) < CMDSIZE))
            {
                strcat(command, " ");
                strcat(command, Includes);
                strcat(command, " ");
            }
            else
                ok = FALSE;
        } /* endif */

	if (ok)
	{
	    switch (Set_operation)
	    {
		case UNION:
		    if (((strlen(command) + strlen(" -union"))+1) < CMDSIZE)
			strcat(command, " -union");
		    else
			ok = FALSE;
		    break;
		case INTERSECTION:
		    if (((strlen(command) + strlen(" -inter"))+1) < CMDSIZE)
			strcat(command, " -inter");
		    else
			ok = FALSE;
		    break;
		case DIFFERENCE:
		    if (((strlen(command) + strlen(" -diff"))+1) < CMDSIZE)
			strcat(command, " -diff");
		    else
			ok = FALSE;
		    break;
		default:
		    /* unrecognizable command ... should never get here */
		popup_message("SYS FAULT: Unrecognizable operator specified.");
		    ok = FALSE;
		    break;
	    } /* end switch() */
	} /* endif */

        if (ok)
        {
            if (Set_sdds_output_spec == REQSDDS)
            {
                if (((strlen(command) + strlen(" -sdds"))+1) < CMDSIZE)
                    strcat(command, " -sdds");
                else
                    ok = FALSE;
            }
            else if (Set_sdds_output_spec == REQNOSDDS)
            {
                if (((strlen(command) + strlen(" -nosdds"))+1) < CMDSIZE)
                    strcat(command, " -nosdds");
                else
                    ok = FALSE;
            } /* endif */
        } /* endif */

	if (ok)
	{
	    set_running(Set_status_label);
	    clear_scrolled_text(Set_log_scroll_txt);
	    
printf(">%s<\n", command);
	    /* calling burtset */
	    rc = system(command);

	    /* dumping logfile into scrolled text */
	    file_to_scrolled_text(Set_logfilename, Set_log_scroll_txt);

	    if (WIFEXITED(rc))
		if (WEXITSTATUS(rc))
		    set_not_ok(Set_status_label);
		else
		    set_ok(Set_status_label);
	    else
	    popup_message("SYS FAULT: system() returned but cmd not exited.");

	}
	else
popup_message("SYS FAULT: could not construct command ... not long enough.");
    } 
    else
    {
	if (!(*op1_fname) || !(*op2_fname))
	    popup_error("ERROR: must specify two request files as operands.");

	if (!(*result_fname))
	    popup_error("ERROR: must specify a result file.");
    } /* endif */

    XtFree(op1_fname);
    XtFree(op2_fname);
    XtFree(result_fname);

} /* end set_go_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void set_view_log_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    XtMapWidget(Set_log_shell);
} /* end set_view_log_button_cb() */

/****************************************************************
*
* called for BOTH operands of the set operation.
*
****************************************************************/

static void set_file_sel_ok_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

char *cp;
Widget text_field;

    text_field = (Widget) client_data;

    cp = XmTextFieldGetString(XmFileSelectionBoxGetChild(w, XmDIALOG_TEXT));

    XtVaSetValues(text_field,
	XmNvalue, cp,
	NULL);

    XtFree(cp);

} /* end set_file_sel_ok_cb() */

/****************************************************************
*
*
****************************************************************/

static void set_operation_toggle_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

    if (XmToggleButtonGetState(w))
	Set_operation = (int) client_data;

} /* end set_operation_toggle_cb() */

/****************************************************************
*
*
****************************************************************/

static void set_sdds_toggle_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

    if (XmToggleButtonGetState(w))
	Set_sdds_output_spec = (int) client_data;

} /* end set_sdds_toggle_cb() */

