/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
#include "burtgooey.h"
#include "burtgooeyfuncdecls.h"
#include <unistd.h>

/* Externally visible functions */
void main_write_button_cb();
void cleanup_write();

/* Private local functions */
static void setup_write();
/* callbacks */
static void write_snap_file_sel_button_cb();
static void write_view_snap_button_cb();
static void write_remove_snap_button_cb();
static void write_clear_snap_button_cb();
static void write_print_snap_button_cb();
static void write_snap_file_sel_ok_cb();
static void write_write_button_cb();
static void write_view_log_button_cb();
static void write_sdds_toggle_cb();

/* Local Global Variables */

static char *Write_labels_strings[] =
    {"Write Absolute Snapshots as Additions",
    "Write Relative Snapshots as Replacements"};
static Widget Write_shell;
static Widget Write_log_shell;
static Widget Write_out_shell;
static Widget Write_snap_scr_lst;
static Widget Write_toggle[XtNumber(Write_labels_strings)];
static Widget Write_log_scroll_txt;
static Widget Write_out_scroll_txt;
static Widget Write_view_snap_scroll_txt;
static Widget Write_view_snap_shell;
static Widget Write_status_label;
static char Write_logfilename[MAXPATHLEN];
static char Write_outfilename[MAXPATHLEN];
static char Write_sdds_output_spec;

/****************************************************************
*
*
****************************************************************/

void main_write_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{
    if (!Write_setup)
    {
	setup_write((Widget) client_data);
	Write_setup= TRUE;
    } /* endif */

    XtMapWidget(Write_shell);

} /* end main_write_button_cb() */

/****************************************************************
*
*
****************************************************************/

void cleanup_write()
{
    unlink(Write_logfilename);
    unlink(Write_outfilename);
} /* end cleanup_write() */

/****************************************************************
*
*
****************************************************************/

static void setup_write(shell)
Widget shell;
{

Arg args[NARGS];
int nargs;
XmString xs;
int i;
Widget write_form, write_rc;

Widget write_snap_files_list_rc;

Widget write_snap_files_buttons_rc;
Widget write_snap_file_sel_button, write_view_snap_button;
Widget write_remove_snap_button, write_clear_snap_button;
Widget write_print_snap_button;

Widget write_parms_rc;
Widget write_labels[XtNumber(Write_labels_strings)];

Widget write_buttons_rc;
Widget write_write_button, write_view_log_button, write_help_button;
Widget write_done_button, write_print_log_button;

/* snapshot file selection box */
Widget write_snap_file_sel;
/* write log shell */
Widget write_log_form, write_log_rc;
Widget write_log_done_button;
/* write out shell */
Widget write_out_form, write_out_rc;
Widget write_out_buttons_rc, write_out_done_button, write_out_print_button;
/* write view snap shell */
Widget write_view_snap_form, write_view_snap_rc;
Widget write_view_snap_done_button;
Widget write_sdds_rc, write_sdds_toggle[XtNumber(Sdds_labels_strings)];

    if (!BurtCreateUniqueFilename(PUBLICWRITEDIRECTORY, ".write.log",
            Write_logfilename))
    {
        fprintf(Log_fp,
            "ERROR: unable to create a unique filename for write log\n");
        exit(1);
    } /* endif */

    if (!BurtCreateUniqueFilename(PUBLICWRITEDIRECTORY, ".nowrite.snap",
            Write_outfilename))
    {
        fprintf(Log_fp,
            "ERROR: unable to create a unique filename for write outfile\n");
        exit(1);
    } /* endif */

    /***************/
    /*             */
    /* write shell */
    /*             */
    /***************/

    Write_shell = XtVaCreateWidget("Restore", 
		    topLevelShellWidgetClass, shell,
		    XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
		    NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 300); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    write_form = XtCreateWidget("write_form", xmFormWidgetClass, Write_shell, 
		    args, nargs);

    /* write popup snapshot file selection box */
    xs = XmStringCreateSimple("Snapshot File Selector");
    nargs = 0;
    XtSetArg(args[nargs], XmNdialogTitle, xs); nargs ++;
    write_snap_file_sel = XmCreateFileSelectionDialog(write_form, 
			    "write_snap_file_sel", args, nargs);
    XtUnmanageChild(XmFileSelectionBoxGetChild(write_snap_file_sel, 
	XmDIALOG_HELP_BUTTON)); /* removing HELP button */
    XmStringFree(xs);

    /* main write rowcolumn */
    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    write_rc = XmCreateRowColumn(write_form, "write_rc", args, nargs);

	/* write snapshot file rowcolumn */
	write_snap_files_list_rc = XmCreateRowColumn(write_rc, 
		"write_snap_files_list_rc", 
		NULL, 0);

	    nargs = 0;
	    XtSetArg(args[nargs], XmNselectionPolicy, XmEXTENDED_SELECT); nargs ++;
	    XtSetArg(args[nargs], XmNheight, 100); nargs ++;
	    Write_snap_scr_lst = XmCreateScrolledList(write_snap_files_list_rc, 
			    "Write_snap_scr_lst", args, nargs);
	    XtManageChild(Write_snap_scr_lst);

	XtManageChild(write_snap_files_list_rc);

	/* write snapshot file buttons rowcolumn */
	nargs = 0;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	write_snap_files_buttons_rc = XmCreateRowColumn(write_rc, 
				    "write_snap_files_buttons_rc", args, nargs);

	    write_snap_file_sel_button =  
		XmCreatePushButton(write_snap_files_buttons_rc, 
		    "write_snap_file_sel_button", 
		    NULL, 0);
	    XtManageChild(write_snap_file_sel_button);

	    write_view_snap_button =  
    XmCreatePushButton(write_snap_files_buttons_rc, "write_view_snap_button", 
		    NULL, 0);
	    XtManageChild(write_view_snap_button);

	    write_remove_snap_button = 
		XmCreatePushButton(write_snap_files_buttons_rc, 
		    "write_remove_snap_button", 
		    NULL, 0);
	    XtManageChild(write_remove_snap_button);

	    write_clear_snap_button = 
		XmCreatePushButton(write_snap_files_buttons_rc, 
		    "write_clear_snap_button", 
		    NULL, 0);
	    XtManageChild(write_clear_snap_button);

	    write_print_snap_button = 
		XmCreatePushButton(write_snap_files_buttons_rc, 
		    "write_print_snap_button", 
		    NULL, 0);
	    XtManageChild(write_print_snap_button);

	XtManageChild(write_snap_files_buttons_rc);

	/* from page 410 AT&T Motif Course Manual */
	nargs = 0;
	XtSetArg(args[nargs], XmNpacking, XmPACK_COLUMN); nargs ++;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	XtSetArg(args[nargs], XmNnumColumns, XtNumber(Write_labels_strings)); nargs ++;
	XtSetArg(args[nargs], XmNadjustLast, False); nargs ++;
	XtSetArg(args[nargs], XmNisAligned, True); nargs ++;
	XtSetArg(args[nargs], XmNentryAlignment, XmALIGNMENT_END); nargs ++;
	write_parms_rc = XmCreateRowColumn(write_rc, "write_parms_rc", 
			    args, nargs);

	    for (i = 0; i < XtNumber(Write_labels_strings); i ++)
	    {
		write_labels[i] = XmCreateLabel(write_parms_rc, 
				    Write_labels_strings[i], NULL, 0);
		Write_toggle[i] = XmCreateToggleButton(write_parms_rc, "",
				    NULL, 0);
	    } /* endfor */
	    XtManageChildren(write_labels, XtNumber(Write_labels_strings));
	    XtManageChildren(Write_toggle, XtNumber(Write_labels_strings));

	XtManageChild(write_parms_rc);

        /* write sdds spec */
        /* from page 414 AT&T Motif Course Manual */
        nargs = 0;
        XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
        write_sdds_rc = XmCreateRadioBox(write_rc, "write_sdds_rc", 
			    args, nargs);
 
            for (i = 0; i < XtNumber(Sdds_labels_strings); i ++)
                write_sdds_toggle[i] = XmCreateToggleButton(write_sdds_rc,
                                    Sdds_labels_strings[i], NULL, 0);
 
            /* setting initial radio box toggle state */
            XmToggleButtonSetState(write_sdds_toggle[NOREQ], True, False);
            Write_sdds_output_spec = NOREQ;
 
            XtManageChildren(write_sdds_toggle, XtNumber(Sdds_labels_strings));
 
        XtManageChild(write_sdds_rc);

	nargs = 0;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	write_buttons_rc = XmCreateRowColumn(write_rc, "write_parms_rc", 
			    args, nargs);

	    write_write_button = 
		XmCreatePushButton(write_buttons_rc, "write_write_button", 
		    NULL, 0);
	    XtManageChild(write_write_button);

	    Write_status_label = 
		XmCreateLabel(write_buttons_rc, "Write_status_label", 
			NULL, 0);

	    set_ok(Write_status_label);
	    XtManageChild(Write_status_label);

	    write_view_log_button = 
		XmCreatePushButton(write_buttons_rc, "write_view_log_button", 
		    NULL, 0);
	    XtManageChild(write_view_log_button);

	    write_print_log_button = 
		XmCreatePushButton(write_buttons_rc, "write_print_log_button", 
		    NULL, 0);
	    XtManageChild(write_print_log_button);

	    write_help_button = 
		XmCreatePushButton(write_buttons_rc, "write_help_button", 
		    NULL, 0);
	    XtManageChild(write_help_button);

	    write_done_button = 
		XmCreatePushButton(write_buttons_rc, "write_done_button", 
		    NULL, 0);
	    XtManageChild(write_done_button);

	XtManageChild(write_buttons_rc);

    XtManageChild(write_rc);

    /* write log shell */
    Write_log_shell = XtVaCreateWidget("Restore Log Viewer", 
		    topLevelShellWidgetClass, Write_shell,
		    XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
		    NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 350); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    write_log_form = XtCreateWidget("write_log_form", xmFormWidgetClass, 
			Write_log_shell, args, nargs);

    /* main write log rowcolumn */
    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    write_log_rc = XmCreateRowColumn(write_log_form, "write_log_rc", 
		    args, nargs);

	nargs = 0;
	XtSetArg(args[nargs], XmNeditMode, XmMULTI_LINE_EDIT); nargs ++;
	XtSetArg(args[nargs], XmNeditable, False); nargs ++;
	XtSetArg(args[nargs], XmNheight, 300); nargs ++;
	Write_log_scroll_txt = 
	    XmCreateScrolledText(write_log_rc, "Write_log_scroll_txt",
		args, nargs);
	XtManageChild(Write_log_scroll_txt);

	write_log_done_button = 
	    XmCreatePushButton(write_log_rc, "write_log_done_button", 
		NULL, 0);
	XtManageChild(write_log_done_button);

    XtManageChild(write_log_rc);
    XtManageChild(write_log_form);
    XtManageChild(Write_log_shell);

    /* write out shell */
    Write_out_shell = XtVaCreateWidget("Read Only Notify", 
		    topLevelShellWidgetClass, Write_shell,
		    XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
		    NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 350); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    write_out_form = 
	XtCreateWidget("write_out_form", xmFormWidgetClass, Write_out_shell, 
	    args, nargs);

    /* main write out rowcolumn */
    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    write_out_rc = XmCreateRowColumn(write_out_form, "write_out_rc", 
		    args, nargs);

	nargs = 0;
	XtSetArg(args[nargs], XmNeditMode, XmMULTI_LINE_EDIT); nargs ++;
	XtSetArg(args[nargs], XmNeditable, False); nargs ++;
	XtSetArg(args[nargs], XmNheight, 300); nargs ++;
	Write_out_scroll_txt = XmCreateScrolledText(write_out_rc, 
	    "Write_out_scroll_txt", args, nargs);
	XtManageChild(Write_out_scroll_txt);

	nargs = 0;
	XtSetArg(args[nargs], XmNorientation, XmHORIZONTAL); nargs ++;
	write_out_buttons_rc = XmCreateRowColumn(write_out_rc, 
				"write_out_buttons_rc", args, nargs);
	    write_out_print_button = XmCreatePushButton(write_out_buttons_rc, 
					"write_out_print_button", NULL, 0);
	    XtManageChild(write_out_print_button);

	    write_out_done_button = XmCreatePushButton(write_out_buttons_rc, 
					"write_out_done_button", NULL, 0);
	    XtManageChild(write_out_done_button);

	XtManageChild(write_out_buttons_rc);

    XtManageChild(write_out_rc);
    XtManageChild(write_out_form);
    XtManageChild(Write_out_shell);

    /* write view snapshot shell */
    Write_view_snap_shell = XtVaCreateWidget("Snapshot File Viewer", 
		    topLevelShellWidgetClass, Write_shell,
		    XmNmappedWhenManaged, False,
		    XmNmwmDecorations, 
			(MWM_DECOR_ALL | MWM_DECOR_MAXIMIZE | MWM_DECOR_MENU),
		    NULL);

    nargs = 0;
    XtSetArg(args[nargs], XmNnoResize, True); nargs ++;
    XtSetArg(args[nargs], XmNheight, 350); nargs ++;
    XtSetArg(args[nargs], XmNwidth, 600); nargs ++;
    write_view_snap_form = XtCreateWidget("write_view_snap_form", 
				xmFormWidgetClass, Write_view_snap_shell,
				args, nargs);

    /* main write rowcolumn */
    nargs = 0;
    XtSetArg(args[nargs], XmNleftAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNtopAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNrightAttachment, XmATTACH_FORM); nargs ++;
    XtSetArg(args[nargs], XmNbottomAttachment, XmATTACH_FORM); nargs ++;
    write_view_snap_rc = XmCreateRowColumn(write_view_snap_form, 
			    "write_view_snap_rc", args, nargs);

	nargs = 0;
	XtSetArg(args[nargs], XmNeditMode, XmMULTI_LINE_EDIT); nargs ++;
	XtSetArg(args[nargs], XmNeditable, False); nargs ++;
	XtSetArg(args[nargs], XmNheight, 300); nargs ++;
	Write_view_snap_scroll_txt = XmCreateScrolledText(write_view_snap_rc, 
					"Write_view_snap_scroll_txt",
					args, nargs);
	XtManageChild(Write_view_snap_scroll_txt);

	write_view_snap_done_button = XmCreatePushButton(write_view_snap_rc, 
					"write_view_snap_done_button", 
					NULL, 0);
	XtManageChild(write_view_snap_done_button);

    XtManageChild(write_view_snap_rc);

    XtManageChild(write_view_snap_form);
    XtManageChild(Write_view_snap_shell);

    XtManageChild(write_form);
    XtManageChild(Write_shell);

    /*************/
    /*           */
    /* Callbacks */
    /*           */
    /*************/

    XtAddCallback(write_write_button, XmNactivateCallback,
        write_write_button_cb, NULL);
    XtAddCallback(write_view_log_button, XmNactivateCallback,
        write_view_log_button_cb, NULL);
    XtAddCallback(write_print_log_button, XmNactivateCallback,
        shared_print_log_button_cb, (char *) Write_logfilename);
    XtAddCallback(write_help_button, XmNactivateCallback,
        shared_temp_help_button_cb, NULL);
    XtAddCallback(write_done_button, XmNactivateCallback,
        shared_done_button_cb, (Widget) Write_shell);
    XtAddCallback(write_remove_snap_button, XmNactivateCallback,
        write_remove_snap_button_cb, NULL);
    XtAddCallback(write_snap_file_sel_button, XmNactivateCallback,
        write_snap_file_sel_button_cb, (Widget) write_snap_file_sel);
    XtAddCallback(write_view_snap_button, XmNactivateCallback,
        write_view_snap_button_cb, NULL);
    XtAddCallback(write_clear_snap_button, XmNactivateCallback,
        write_clear_snap_button_cb, NULL);
    XtAddCallback(write_print_snap_button, XmNactivateCallback,
        write_print_snap_button_cb, NULL);
    XtAddCallback(write_snap_file_sel, XmNcancelCallback,
        XtUnmanageChild, NULL);
    XtAddCallback(write_snap_file_sel, XmNokCallback,
        write_snap_file_sel_ok_cb, NULL);
    /* write log shell */
    XtAddCallback(write_log_done_button, XmNactivateCallback,
        shared_done_button_cb, (Widget) Write_log_shell);
    /* write out shell */
    XtAddCallback(write_out_done_button, XmNactivateCallback,
        shared_done_button_cb, (Widget) Write_out_shell);
    /* write view snapshot shell */
    XtAddCallback(write_view_snap_done_button, XmNactivateCallback,
        shared_done_button_cb, (Widget) Write_view_snap_shell);
    /* write out shell ... for read only notify */
    for (i = 0; i < XtNumber(Sdds_labels_strings); i ++)
        XtAddCallback(write_sdds_toggle[i], XmNvalueChangedCallback,
            write_sdds_toggle_cb, (int) i);
    XtAddCallback(write_out_print_button, XmNactivateCallback,
        shared_print_log_button_cb, (char *) Write_outfilename);

} /* end setup_write() */

/*************************/
/*                       */
/* write shell callbacks */
/*                       */
/*************************/

/****************************************************************
*
*
****************************************************************/

static void write_clear_snap_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    XmListDeleteAllItems(Write_snap_scr_lst);
} /* end write_clear_snap_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void write_print_snap_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

XmStringTable choices;
int n_choices;
int i;
char mbuff[1000];
char *cp;

    /* returns the items themselves, not a copy ... do not free */
    XtVaGetValues(Write_snap_scr_lst,
	XmNselectedItemCount, &n_choices,
	XmNselectedItems, &choices,
	NULL);

    if (n_choices)
    {
	/* loop to discover legal printable files from those selected */
	for (i = 0; i < n_choices; i ++)
	{
	    if (XmStringGetLtoR(choices[i], XmFONTLIST_DEFAULT_TAG, &cp))
	    {
		print_file(cp);
		XtFree(cp);
	    } 
	    else
popup_message("SYS FAULT: unable to convert one of the files to a string.");
	} /* endfor */
    } /* endif */

} /* end write_print_snap_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void write_snap_file_sel_ok_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

char *cp;
XmString xm_str;

    cp = XmTextFieldGetString(XmFileSelectionBoxGetChild(w, XmDIALOG_TEXT));

    xm_str = XmStringCreateSimple(cp);

    if (!XmListItemExists(Write_snap_scr_lst, xm_str))
	/* placing it in position 0 ... meaning at the end */
	XmListAddItem(Write_snap_scr_lst, xm_str, 0); 

    XtFree(cp);
    XmStringFree(xm_str);

} /* end write_snap_file_sel_ok_cb() */

/****************************************************************
*
*
****************************************************************/

static void write_write_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

XmStringTable choices;
XmString xs;
int n_choices;
int i;
int rc;
int buffsize;
char *cp;
char command[CMDSIZE];
char *logbuff;
BOOL ok;

    ok = TRUE;

    /* returns the items themselves, not a copy ... do not free */
    XtVaGetValues(Write_snap_scr_lst,
	XmNitemCount, &n_choices,
	XmNitems, &choices,
	NULL);

    if (n_choices)
    {
	strcpy(command, "burtwb -f");
	for (i = 0; ok && i < n_choices; i ++)
	{
	    if (XmStringGetLtoR(choices[i], XmFONTLIST_DEFAULT_TAG, &cp))
	    {
		if ((strlen(command) +strlen(cp) + 2) < CMDSIZE)
		{
		    strcat(command, " ");
		    strcat(command, cp);
		    XtFree(cp);
		}
		else
		    ok = FALSE;
	    } /* endif */
	} /* endfor */

        if (*Defines)
        {
            if (ok && ((strlen(command) + 2+strlen(Defines)) < CMDSIZE))
            {
    printf("adding Defines >%s< to command\n", Defines);
                strcat(command, " ");
                strcat(command, Defines);
                strcat(command, " ");
            }
            else
                ok = FALSE;
        } /* endif */

        if (*Includes)
        {
            if (ok && ((strlen(command) + 2+strlen(Includes)) < CMDSIZE))
            {
                strcat(command, " ");
                strcat(command, Includes);
                strcat(command, " ");
            }
            else
                ok = FALSE;
        } /* endif */

	if (ok &&
	    ((strlen(command) 
	    + strlen(" -l ")+strlen(Write_logfilename) 
	    + strlen(" -o ")+strlen(Write_outfilename) 
	    + strlen(" -v ")) 
	    < CMDSIZE))
	{
	    unlink(Write_logfilename);
	    strcat(command, " -l ");
	    strcat(command, Write_logfilename);

	    unlink(Write_outfilename);
	    strcat(command, " -o ");
	    strcat(command, Write_outfilename);

	    strcat(command, " -v ");
	} 
	else
	    ok = FALSE;

	if (XmToggleButtonGetState(Write_toggle[0]))
	{
	    if (ok && (strlen(command) + strlen(" -add ")) < CMDSIZE)
		strcat(command, " -add ");
	    else
		ok = FALSE;
	} /* endif */

	if (XmToggleButtonGetState(Write_toggle[1]))
	{
	    if (ok && (strlen(command) + strlen(" -replace ")) < CMDSIZE)
		strcat(command, " -replace ");
	    else
		ok = FALSE;
	} /* endif */

        if (ok)
        {
            if (Write_sdds_output_spec == REQSDDS)
            {
                if (((strlen(command) + strlen(" -sdds"))+1) < CMDSIZE)
                    strcat(command, " -sdds");
                else
                    ok = FALSE;
            }
            else if (Write_sdds_output_spec == REQNOSDDS)
            {
                if (((strlen(command) + strlen(" -nosdds"))+1) < CMDSIZE)
                    strcat(command, " -nosdds");
                else
                    ok = FALSE;
            } /* endif */
        } /* endif */

	if (ok)
	{
	    set_running(Write_status_label);
	    clear_scrolled_text(Write_log_scroll_txt);
	    clear_scrolled_text(Write_out_scroll_txt);
	    
printf(">%s<\n", command);
	    /* calling burtwb */
	    rc = system(command);

	    /* dumping logfile into scrolled text */
	    file_to_scrolled_text(Write_logfilename, Write_log_scroll_txt);

	    if (WIFEXITED(rc))
	    {
		if (WEXITSTATUS(rc))
		    set_not_ok(Write_status_label);
		else
		    set_ok(Write_status_label);

		if (file_exists(Write_outfilename))
		{
		    /* file is there */
		    /* dumping outfile into scrolled text */
		    file_to_scrolled_text(Write_outfilename, 
			Write_out_scroll_txt);
		    XtMapWidget(Write_out_shell);
		} /* endif */
	    }
	    else
	    popup_message("SYS FAULT: system() returned but cmd not exited.");

	}
	else
popup_message("SYS FAULT: could not construct command ... not long enough.");
    } 
    else
	popup_error("ERROR: Need at least one snapshot file.");

} /* end write_write_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void write_remove_snap_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

int *pos_list, pos_cnt;
int fix, i, j;

    pos_list = (int *) NULL;

    if (XmListGetSelectedPos(Write_snap_scr_lst, &pos_list, &pos_cnt))
    {
	/* at least one selected in scrolled list */
	for (i = 0; i < pos_cnt; i ++)
	{
	    /* loop to find all those that have been removed */
	    /* so far that were ABOVE the item being removed */
	    /* now.  Each one we find that was removed ABOVE */
	    /* the one we're removing now decrements the pos */
	    /* of the one we're removing now by one.         */
	    for (j = 0, fix = 0; j < i; j ++)
		if (pos_list[j] < pos_list[i])
		    fix ++;
	    XmListDeletePos(Write_snap_scr_lst, pos_list[i]-fix);
	} /* endfor */
    } /* endif */

    XtFree(pos_list);

} /* end write_remove_snap_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void write_snap_file_sel_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{

    XtManageChild((Widget) client_data);
    XtPopup(XtParent((Widget) client_data), XtGrabNone);

} /* end write_snap_file_sel_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void write_view_log_button_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    XtMapWidget(Write_log_shell);
} /* end write_view_log_button_cb() */

/****************************************************************
*
*
****************************************************************/

static void write_view_snap_button_cb(w, client_data, call_data)
Widget w; 
XtPointer client_data; 
XtPointer call_data;
{

XmStringTable choices;
int n_choices;
int i, j;
int next_file_idx;
int total_bytes;
char *total_buff;
int fd;
char mbuff[1000];
struct filerep *fbuff;
char *filecp;
char *cp;
struct stat stat_buff;
BOOL all_ascii;

    /* clearing the view snapshot scroll text */
    XtVaSetValues(Write_view_snap_scroll_txt,
	XmNvalue, "",
	NULL);

    /* returns the items themselves, not a copy ... do not free */
    XtVaGetValues(Write_snap_scr_lst,
	XmNselectedItemCount, &n_choices,
	XmNselectedItems, &choices,
	NULL);

    if (n_choices)
    {
	if (fbuff = (struct filerep *) calloc(n_choices,sizeof(struct filerep)))
	{
	    /* loop to discover legal printable files from those selected */
	    for (i = 0, next_file_idx = 0; i < n_choices; i ++)
	    {
		if (XmStringGetLtoR(choices[i], XmFONTLIST_DEFAULT_TAG, &cp))
		{
		    if ((fd = open(cp, O_RDONLY)) != -1)
		    {
			if (!fstat(fd, &stat_buff))
			{
			    if (S_ISREG(stat_buff.st_mode))
			    {
				if (fbuff[next_file_idx].buff = 
				    (char *) calloc (stat_buff.st_size,1))
				{
				    if (read(fd, fbuff[next_file_idx].buff, 
					stat_buff.st_size) != -1)
				    {
					/* loop to check file is printable */
					for (j = 0, all_ascii = TRUE; 
					    all_ascii && j < stat_buff.st_size; 
						j ++)
						    all_ascii = 
					isascii(fbuff[next_file_idx].buff[j]);

					if (all_ascii)
					{
					    fbuff[next_file_idx].nbytes 
						= stat_buff.st_size;
					    fbuff[next_file_idx].fname =
						strdup(cp);
					    next_file_idx ++;
					}
					else
					{
					    free(fbuff[next_file_idx].buff);

					    sprintf(mbuff, 
						"ERROR: file %s not viewable.",
						cp);
					    popup_error(mbuff);
					} /* endif */
				    }
				    else
				    {
					free(fbuff[next_file_idx].buff);

					sprintf(mbuff, 
					"SYS FAULT: could not read() file %s.",
					    cp);
					popup_message(mbuff);
				    } /* endif */

				}
				else
				{
				    sprintf(mbuff, 
					"SYS FAULT: failed calloc() %d bytes.", 
					stat_buff.st_size);
				    popup_message(mbuff);
				} /* endif */
			    }
			    else
			    {
				sprintf(mbuff, 
		"NOTE: cannot view file %s because it is not a regular file.", 
				    cp);
			    popup_message(mbuff);
			    } /* endif */
			}
			else
			    popup_message("SYS FAULT: error in fstat().");

			close(fd);
		    }
		    else
		    {
			sprintf(mbuff, "SYS FAULT: cannot open() file %s.", cp);
			popup_message(mbuff);
		    } /* endif */

		    XtFree(cp);

		} 
		else
    popup_message("SYS FAULT: unable to convert one of the files to a string.");
	    } /* endfor */

	    /* now fbuff has all the valid files */
	    for (i = 0, total_bytes = 0; i < next_file_idx; i ++)
		total_bytes += 
		    (fbuff[i].nbytes +strlen(fbuff[i].fname)+1+strlen("\n\n"));

	    /* for ending null char */
	    total_bytes ++;

	    if (total_bytes > 1)
	    {
		/* there are files to view */
		if (total_buff = (char *) calloc(total_bytes, 1))
		{
		    /* loop to append (via copy) each file to total_buff */
		    /* and then free mem from fbuff */
		    for (i = 0, cp = total_buff; i < next_file_idx; i ++)
		    {
			memcpy(cp, fbuff[i].fname, strlen(fbuff[i].fname));
			cp = cp + strlen(fbuff[i].fname);

			*cp = '\n';
			cp ++;

			memcpy(cp, fbuff[i].buff, fbuff[i].nbytes);
			cp = cp + fbuff[i].nbytes;

			memcpy(cp, "\n\n", strlen("\n\n"));
			cp = cp + strlen("\n\n");

			free(fbuff[i].buff);
			free(fbuff[i].fname);
		    } /* endfor */

		    /* although buffer was allocated with calloc(), which */
		    /* presumably allocates a block with all zeros, still */
		    /* need to explicitly place the null char at the end  */
		    /* so as not to get extra garbage characters in the   */
		    /* scroll text ... can't explain why this works ...   */
		    /* just happy that it does.                           */

		    *cp = '\0';

		XtVaSetValues(Write_view_snap_scroll_txt,
			XmNvalue, total_buff,
			NULL);

		    free(total_buff);
		}
		else
		{
		    sprintf(mbuff, "SYS FAULT: failed calloc() %d bytes.", 
			total_bytes);
		    popup_message(mbuff);
		} /* endif */
	    } /* endif */

	    free(fbuff);
	}
	else
	{
	    sprintf(mbuff, 
		"SYS FAULT: failed to calloc() %d bytes.", 
		n_choices*sizeof(struct filerep));
	    popup_message(mbuff);
	} /* endif */
    } /* endif */

    XtMapWidget(Write_view_snap_shell);

} /* end write_view_snap_button_cb() */

/****************************************************************
*
*
****************************************************************/
 
static void write_sdds_toggle_cb(w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
 
    if (XmToggleButtonGetState(w))
        Write_sdds_output_spec = (int) client_data;
 
} /* end write_sdds_toggle_cb() */
 
